Package: NanoMethViz
Type: Package
Title: Visualise methylation data from Oxford Nanopore sequencing
Version: 3.7.0
Authors@R: c(
    person("Shian", "Su", email = "su.s@wehi.edu.au", role = c("cre", "aut")))
Description: NanoMethViz is a toolkit for visualising methylation data from
    Oxford Nanopore sequencing. It can be used to explore methylation patterns
    from reads derived from Oxford Nanopore direct DNA sequencing with
    methylation called by callers including nanopolish, f5c and megalodon. The
    plots in this package allow the visualisation of methylation profiles
    aggregated over experimental groups and across classes of genomic features.
biocViews: Software, LongRead, Visualization, DifferentialMethylation,
        DNAMethylation, Epigenetics, DataImport
URL: https://github.com/shians/NanoMethViz,
        https://shians.github.io/NanoMethViz/
BugReports: https://github.com/Shians/NanoMethViz/issues
Depends: R (>= 4.0.0), methods, ggplot2 (>= 3.4.0)
Imports: cpp11 (>= 0.2.5), readr, cli, S4Vectors, SummarizedExperiment,
        BiocSingular, bsseq, forcats, assertthat, AnnotationDbi, Rcpp,
        dplyr, dbscan, e1071, fs, GenomicRanges, Biostrings, ggrastr,
        glue, graphics, IRanges, limma (>= 3.44.0), patchwork, purrr,
        rlang, R.utils, Rsamtools, scales (>= 1.2.0), stats, stringr,
        tibble, tidyr, utils, withr
Suggests: BiocStyle, DSS, Mus.musculus (>= 1.3.1), Homo.sapiens (>=
        1.3.1), org.Hs.eg.db, TxDb.Hsapiens.UCSC.hg19.knownGene,
        TxDb.Hsapiens.UCSC.hg38.knownGene, org.Mm.eg.db,
        TxDb.Mmusculus.UCSC.mm10.knownGene,
        TxDb.Mmusculus.UCSC.mm39.refGene, knitr, rmarkdown,
        rtracklayer, testthat (>= 3.0.0), covr
LinkingTo: Rcpp
License: Apache License (>= 2.0)
SystemRequirements: C++20
VignetteBuilder: knitr
Encoding: UTF-8
Roxygen: list(markdown = TRUE)
RoxygenNote: 7.3.2
Config/testthat/parallel: true
Config/testthat/edition: 3
git_url: https://git.bioconductor.org/packages/NanoMethViz
git_branch: devel
git_last_commit: 4349b36
git_last_commit_date: 2025-10-29
Repository: Bioconductor 3.23
Date/Publication: 2025-11-04
NeedsCompilation: yes
Packaged: 2025-11-04 23:55:01 UTC; biocbuild
Author: Shian Su [cre, aut]
Maintainer: Shian Su <su.s@wehi.edu.au>

# Generated by roxygen2: do not edit by hand

export("exons<-")
export("methy<-")
export("mod_code<-")
export("samples<-")
export(ModBamFiles)
export(ModBamResult)
export(NanoMethResult)
export(bsseq_to_edger)
export(bsseq_to_log_methy_ratio)
export(cluster_regions)
export(create_tabix_file)
export(exons)
export(exons_to_genes)
export(filter_methy)
export(get_cgi_grcm39)
export(get_cgi_hg19)
export(get_cgi_hg38)
export(get_cgi_mm10)
export(get_cgi_t2t)
export(get_example_exons_mus_musculus)
export(get_exons_grcm39)
export(get_exons_hg19)
export(get_exons_hg38)
export(get_exons_homo_sapiens)
export(get_exons_mm10)
export(get_exons_mus_musculus)
export(get_exons_t2t)
export(load_example_modbamresult)
export(load_example_nanomethresult)
export(methy)
export(methy_col_names)
export(methy_to_bsseq)
export(methy_to_edger)
export(mod_code)
export(modbam_to_tabix)
export(plot_agg_genes)
export(plot_agg_regions)
export(plot_gene)
export(plot_gene_heatmap)
export(plot_grange)
export(plot_grange_heatmap)
export(plot_mds)
export(plot_pca)
export(plot_region)
export(plot_region_heatmap)
export(plot_violin)
export(query_exons_gene_id)
export(query_exons_region)
export(query_exons_symbol)
export(query_methy)
export(region_methy_stats)
export(samples)
export(sigmoid)
exportClasses(ModBamFiles)
exportClasses(ModBamResult)
exportClasses(NanoMethResult)
exportMethods("exons<-")
exportMethods("methy<-")
exportMethods("mod_code<-")
exportMethods("samples<-")
exportMethods(exons)
exportMethods(methy)
exportMethods(mod_code)
exportMethods(plot_gene)
exportMethods(plot_gene_heatmap)
exportMethods(plot_region)
exportMethods(plot_region_heatmap)
exportMethods(samples)
exportMethods(show)
import(Rcpp)
import(assertthat)
import(cpp11)
import(dbscan)
import(dplyr)
import(fs)
import(patchwork)
import(tidyr)
importFrom(BiocSingular,runPCA)
importFrom(Biostrings,width)
importFrom(IRanges,IRanges)
importFrom(R.utils,gunzip)
importFrom(Rsamtools,TabixFile)
importFrom(Rsamtools,scanTabix)
importFrom(S4Vectors,DataFrame)
importFrom(SummarizedExperiment,colData)
importFrom(assertthat,assert_that)
importFrom(assertthat,is.dir)
importFrom(assertthat,is.readable)
importFrom(assertthat,is.string)
importFrom(assertthat,is.writeable)
importFrom(bsseq,BSseq)
importFrom(dplyr,"%>%")
importFrom(dplyr,arrange)
importFrom(dplyr,case_when)
importFrom(dplyr,distinct)
importFrom(dplyr,filter)
importFrom(dplyr,group_by)
importFrom(dplyr,inner_join)
importFrom(dplyr,mutate)
importFrom(dplyr,n)
importFrom(dplyr,rename)
importFrom(dplyr,select)
importFrom(dplyr,summarise)
importFrom(dplyr,ungroup)
importFrom(e1071,sigmoid)
importFrom(forcats,fct_reorder)
importFrom(ggplot2,aes)
importFrom(ggplot2,draw_key_point)
importFrom(ggplot2,geom_rect)
importFrom(ggplot2,geom_segment)
importFrom(ggplot2,geom_text)
importFrom(ggplot2,geom_vline)
importFrom(ggplot2,ggplot)
importFrom(ggplot2,ggplot_build)
importFrom(ggplot2,rel)
importFrom(ggplot2,theme_void)
importFrom(ggplot2,unit)
importFrom(ggplot2,xlim)
importFrom(ggplot2,ylim)
importFrom(ggrastr,rasterise)
importFrom(glue,glue)
importFrom(graphics,hist)
importFrom(limma,plotMDS)
importFrom(methods,.valueClassTest)
importFrom(methods,is)
importFrom(methods,new)
importFrom(purrr,map)
importFrom(purrr,map_lgl)
importFrom(readr,col_character)
importFrom(readr,col_double)
importFrom(readr,col_integer)
importFrom(readr,col_logical)
importFrom(readr,cols)
importFrom(readr,local_edition)
importFrom(readr,read_tsv)
importFrom(rlang,.data)
importFrom(scales,cut_si)
importFrom(scales,label_number)
importFrom(stats,approxfun)
importFrom(stats,kmeans)
importFrom(stats,median)
importFrom(stats,prcomp)
importFrom(stats,quantile)
importFrom(stats,sd)
importFrom(stats,setNames)
importFrom(stringr,str_extract)
importFrom(tibble,add_column)
importFrom(tibble,as_tibble)
importFrom(tibble,rownames_to_column)
importFrom(tibble,tibble)
importFrom(tidyr,unnest)
importFrom(utils,installed.packages)
importFrom(utils,packageVersion)
importFrom(utils,read.delim)
importFrom(utils,read.table)
importFrom(withr,defer)
useDynLib(NanoMethViz)

# NanoMethViz

<!-- badges: start -->
[![Codecov test coverage](https://codecov.io/gh/Shians/NanoMethViz/branch/master/graph/badge.svg)](https://codecov.io/gh/Shians/NanoMethViz?branch=master)
[![R-CMD-check](https://github.com/Shians/NanoMethViz/workflows/R-CMD-check/badge.svg)](https://github.com/Shians/NanoMethViz/actions)
<!-- badges: end -->

NanoMethViz is a toolkit for visualising methylation data from Oxford Nanopore sequencing.

## Installation

You can install NanoMethViz from Bioconductor with:

``` r
if (!requireNamespace("BiocManager", quietly = TRUE))
    install.packages("BiocManager")

BiocManager::install("NanoMethViz")
```

To install the latest developmental version, use:

``` r
if (!requireNamespace("BiocManager", quietly = TRUE))
    install.packages("BiocManager")

BiocManager::install(version='devel')

BiocManager::install("NanoMethViz")
```

## Usage

This package currently works with data from dorado, megalodon, nanopolish and f5c, for information on how to use the
package, please refer to the [Users Guide](https://www.bioconductor.org/packages/release/bioc/vignettes/NanoMethViz/inst/doc/UsersGuide.html).

## Examples

### MDS Plot

The MDS plot is used to visualise differences in the methylation profiles of
multiple samples.

![](man/figures/mds.png)

### Feature Aggregation

The feature aggregation plot can average the methylation profiles across a set
of features.

![](man/figures/agg_genes.png)

### Region Methylation Plot

The region methylation plot can visualise the methylation profile of a region
of interest. As well as provide a heatmap of the methylation along individual
reads.

![](man/figures/peg3_gene.png)


## License

This project is licensed under Apache License, Version 2.0.

StatLm <- ggplot2::ggproto("StatLm", Stat,
    required_aes = c("x", "y"),

    compute_group = function(data, scales, params, n = 20) {
        data <- data[!duplicated(data$x), ]

        if (nrow(data) <= 1) {
            return(data.frame(x = NULL, y = NULL))
        }

        poly_deg <- min(6, nrow(data) - 1)
        rng <- range(data$x, na.rm = TRUE)
        grid <- data.frame(x = seq(rng[1], rng[2], length = n))

        mod <- tryCatch(
            lm(y ~ poly(x, poly_deg), data = data),
            error = function(e) numeric(0)
        )

        if (length(mod) == 0) {
            return(data.frame(x = NULL, y = NULL))
        }

        grid$y <- predict(mod, newdata = grid)

        grid
    }
)

stat_lm <- function(
    mapping = NULL, data = NULL, geom = "line",
    position = "identity", na.rm = FALSE, show.legend = NA,
    inherit.aes = TRUE, n = 50,
    ...
) {
    ggplot2::layer(
        stat = StatLm, data = data, mapping = mapping, geom = geom,
        position = position, show.legend = show.legend, inherit.aes = inherit.aes,
        params = list(n = n, na.rm = na.rm, ...)
    )
}

parse_bam_list <- function(seq, cigar, mod_str, mod_scores, map_pos, strand, mod_code) {
    # Check inputs
    assert_that(is.character(seq))
    assert_that(is.character(cigar))
    assert_that(is.character(mod_str))
    assert_that(is.character(mod_scores))
    assert_that(is.numeric(map_pos))
    assert_that(all(strand %in% c("+", "-")), msg = "Strand must be '+' or '-'")
    assert_that(is.character(mod_code) && nchar(mod_code) == 1)

    # Check lengths
    assert_that(length(seq) == length(cigar))
    assert_that(length(seq) == length(mod_str))
    assert_that(length(seq) == length(mod_scores))
    assert_that(length(seq) == length(map_pos))
    assert_that(length(seq) == length(strand))

    parse_bam_list_cpp(
        seq,
        cigar,
        mod_str,
        mod_scores,
        map_pos,
        as.character(strand),
        mod_code
    )
}

parse_modbam <- function(x, sample, mod_code) {
    if (is.null(x$tag$ML)) {
        mod_scores <- NULL
    } else {
        mod_scores <- x$tag$ML %>%
            purrr::map_chr(~stringr::str_c(., collapse = ","))
    }

    reads_df <- tibble::tibble(
        chr = x$rname,
        strand = x$strand,
        map_pos = x$pos,
        cigar = x$cigar,
        seq = as.character(x$seq),
        mod_string = x$tag$MM,
        mod_scores = mod_scores,
        read_name = x$qname
    ) %>%
        dplyr::filter(seq != "")

    if (nrow(reads_df) == 0) {
        return(NULL)
    }

    out <- reads_df %>%
        mutate(
            modbam_stats = parse_bam_list(
                .data$seq,
                .data$cigar,
                .data$mod_string,
                .data$mod_scores,
                .data$map_pos,
                as.character(.data$strand),
                mod_code
            )

        )

    out %>%
        dplyr::select("read_name", "chr", "strand", "modbam_stats") %>%
        tidyr::unnest("modbam_stats") %>%
        tidyr::drop_na() %>%
        dplyr::mutate(sample = sample, .before = 1) %>%
        dplyr::select("sample", "chr", "pos", "strand", "statistic", "read_name")
}

read_bam <- function(bam_file, query = NULL) {
    modbam_param <- function(query = NULL) {
        if (!is.null(query)) {
            Rsamtools::ScanBamParam(
                flag = Rsamtools::scanBamFlag(isUnmappedQuery = FALSE),
                what = c("qname", "rname", "strand", "pos", "cigar", "seq"),
                tag = c("MM", "ML", "Mm", "Ml"),
                which = query
            )
        } else {
            Rsamtools::ScanBamParam(
                flag = Rsamtools::scanBamFlag(isUnmappedQuery = FALSE),
                what = c("qname", "rname", "strand", "pos", "cigar", "seq"),
                tag = c("MM", "ML", "Mm", "Ml")
            )
        }
    }

    # determine if ML or ml is used in BAM records
    determine_tag_case <- function(records) {
        if (is.null(records$tag$Mm) && is.null(records$tag$Ml)) {
            records$tag$Mm <- NULL
            records$tag$Ml <- NULL
        } else {
            records$tag$MM <- records$tag$Mm
            records$tag$ML <- records$tag$Ml
            records$tag$Mm <- NULL
            records$tag$Ml <- NULL
        }

        records
    }

    # filter any reads without mod tags
    filter_missing_mod_tags <- function(x) {
        tag <- x$tag
        x$tag <- NULL

        missing_tags <- map_lgl(tag$ML, ~length(.) == 0) |
            map_lgl(x$rname, is.na) |
            map_lgl(x$strand, is.na) |
            map_lgl(x$pos, is.na) |
            map_lgl(x$cigar, is.na)

        x <- map(x, ~.[!missing_tags])
        tag <- map(tag, ~.[!missing_tags])

        x$tag <- tag
        x
    }

    Rsamtools::scanBam(
        bam_file,
        param = modbam_param(query = query)
    ) %>%
        map(determine_tag_case) %>%
        map(filter_missing_mod_tags)
}

read_modbam_table <- function(x, chr, start, end, sample, mod_code) {
    bam_file <- Rsamtools::BamFile(x)

    query <- make_granges(chr, start, end)

    if (length(query) == 0) {
        return(list(empty_methy_query_output()))
    }

    reads <- read_bam(bam_file, query = query)

    lapply(reads, parse_modbam, sample = sample, mod_code = mod_code)
}

#' Query exons
#'
#' Query a data.frame, NanoMethResult or ModBamResult for exon annotation.
#'
#' @param x the object to query.
#'
#' @return data.frame of queried exons.
#'
#' @name query_exons
NULL

#' @param chr the chromosome to query.
#' @param start the start of the query region.
#' @param end the end of the query region.
#'
#' @export
#' @describeIn query_exons Query region.
query_exons_region <- function(x, chr, start, end) {
    exons <- get_exons(x)

    genes_df <- exons %>%
        dplyr::group_by(.data$gene_id, .data$chr) %>%
        dplyr::summarise(
            start = min(start),
            end = max(end)
        )

    genes <- genes_df %>%
        dplyr::filter(
            start <= !!end & end >= !!start,
            chr == !!chr
        ) %>%
        dplyr::select("gene_id")

    exons %>%
        dplyr::inner_join(genes, by = "gene_id", multiple = "all")
}

#' @param gene_id the gene_id to query.
#'
#' @export
#' @describeIn query_exons Query gene ID.
query_exons_gene_id <- function(x, gene_id) {
    exons <- get_exons(x)

    out <- exons %>%
        dplyr::filter(gene_id %in% !!gene_id)

    if (nrow(out) == 0) {
        stop(glue::glue("gene {gene_id} not found in exon annotation"))
    }

    out
}

#' @param symbol the gene_id to query.
#'
#' @export
#' @describeIn query_exons Query gene symbol.
query_exons_symbol <- function(x, symbol) {
    exons <- get_exons(x)

    out <- exons %>%
        dplyr::filter(symbol %in% !!symbol)

    if (nrow(out) == 0) {
        stop(glue::glue("gene {symbol} not found in exon annotation"))
    }

    out
}

get_exons <- function(x) {
    if (is(x, "NanoMethResult")) {
        exons(x)
    } else if (is(x, "ModBamResult")) {
        exons(x)
    } else if (is(x, "data.frame")) {
        x
    } else {
        stop("x must be a NanoMethResult, ModBamResult or data.frame")
    }
}

test_that("query_exons_region works with NanoMethResult", {
    nmr <- load_example_nanomethresult()

    # Test basic region query
    result <- query_exons_region(nmr, "chr7", 6703000, 6730000)
    expect_s3_class(result, "tbl_df")
    expect_true(nrow(result) > 0)
    expect_true(all(c("gene_id", "chr", "start", "end", "symbol") %in% colnames(result)))
    expect_true(all(result$chr == "chr7"))

    # Test region with no overlapping genes
    result_empty <- query_exons_region(nmr, "chr7", 1, 100)
    expect_s3_class(result_empty, "tbl_df")
    expect_equal(nrow(result_empty), 0)

    # Test region with partial overlap
    result_partial <- query_exons_region(nmr, "chr7", 6720000, 6725000)
    expect_s3_class(result_partial, "tbl_df")
})

test_that("query_exons_region works with data.frame", {
    nmr <- load_example_nanomethresult()
    exons_df <- exons(nmr)

    # Test with data.frame directly
    result <- query_exons_region(exons_df, "chr7", 6703000, 6730000)
    expect_s3_class(result, "tbl_df")
    expect_true(nrow(result) > 0)
    expect_true(all(result$chr == "chr7"))
})

test_that("query_exons_gene_id works correctly", {
    nmr <- load_example_nanomethresult()

    # Get a valid gene_id from the exons
    valid_gene_id <- exons(nmr)$gene_id[1]

    # Test with valid gene_id
    result <- query_exons_gene_id(nmr, valid_gene_id)
    expect_s3_class(result, "tbl_df")
    expect_true(nrow(result) > 0)
    expect_true(all(result$gene_id == valid_gene_id))

    # Test with multiple gene_ids
    gene_ids <- exons(nmr)$gene_id[1:2]
    result_multiple <- query_exons_gene_id(nmr, gene_ids)
    expect_s3_class(result_multiple, "tbl_df")
    expect_true(nrow(result_multiple) > 0)
    expect_true(all(result_multiple$gene_id %in% gene_ids))

    # Test with data.frame
    exons_df <- exons(nmr)
    result_df <- query_exons_gene_id(exons_df, valid_gene_id)
    expect_s3_class(result_df, "tbl_df")
    expect_true(all(result_df$gene_id == valid_gene_id))
})

test_that("query_exons_gene_id error handling", {
    nmr <- load_example_nanomethresult()

    # Test with non-existent gene_id
    expect_error(
        query_exons_gene_id(nmr, "NonExistentGeneID"),
        "gene NonExistentGeneID not found in exon annotation"
    )

    # Test with vector of non-existent gene_ids
    expect_error(
        query_exons_gene_id(nmr, c("Gene1", "Gene2")),
        "not found in exon annotation"
    )
})

test_that("query_exons_symbol works correctly", {
    nmr <- load_example_nanomethresult()

    # Get a valid symbol from the exons
    valid_symbol <- exons(nmr)$symbol[1]

    # Test with valid symbol
    result <- query_exons_symbol(nmr, valid_symbol)
    expect_s3_class(result, "tbl_df")
    expect_true(nrow(result) > 0)
    expect_true(all(result$symbol == valid_symbol))

    # Test with multiple symbols
    symbols <- unique(exons(nmr)$symbol)[1:2]
    result_multiple <- query_exons_symbol(nmr, symbols)
    expect_s3_class(result_multiple, "tbl_df")
    expect_true(nrow(result_multiple) > 0)
    expect_true(all(result_multiple$symbol %in% symbols))

    # Test with data.frame
    exons_df <- exons(nmr)
    result_df <- query_exons_symbol(exons_df, valid_symbol)
    expect_s3_class(result_df, "tbl_df")
    expect_true(all(result_df$symbol == valid_symbol))
})

test_that("query_exons_symbol error handling", {
    nmr <- load_example_nanomethresult()

    # Test with non-existent symbol
    expect_error(
        query_exons_symbol(nmr, "NonExistentSymbol"),
        "gene NonExistentSymbol not found in exon annotation"
    )

    # Test with vector of non-existent symbols
    expect_error(
        query_exons_symbol(nmr, c("Symbol1", "Symbol2")),
        "not found in exon annotation"
    )
})

test_that("get_exons helper function works correctly", {
    nmr <- load_example_nanomethresult()
    exons_df <- exons(nmr)

    # Test with NanoMethResult
    result_nmr <- NanoMethViz:::get_exons(nmr)
    expect_identical(result_nmr, exons(nmr))

    # Test with data.frame
    result_df <- NanoMethViz:::get_exons(exons_df)
    expect_identical(result_df, exons_df)

    # Test error handling with invalid object
    expect_error(
        NanoMethViz:::get_exons("invalid_object"),
        "x must be a NanoMethResult, ModBamResult or data.frame"
    )

    expect_error(
        NanoMethViz:::get_exons(123),
        "x must be a NanoMethResult, ModBamResult or data.frame"
    )
})

test_that("query_exons functions work with edge cases", {
    nmr <- load_example_nanomethresult()

    # Test query_exons_region with exact boundaries
    first_gene <- exons(nmr) %>%
        dplyr::group_by(gene_id) %>%
        dplyr::summarise(start = min(start), end = max(end), chr = first(chr)) %>%
        dplyr::slice(1)

    result_exact <- query_exons_region(nmr, first_gene$chr, first_gene$start, first_gene$end)
    expect_true(nrow(result_exact) > 0)
    expect_true(first_gene$gene_id %in% result_exact$gene_id)

    # Test query_exons_region with chromosome that doesn't exist
    result_no_chr <- query_exons_region(nmr, "chrNonExistent", 1, 1000)
    expect_equal(nrow(result_no_chr), 0)

    # Test with empty exons data.frame
    empty_exons <- tibble::tibble(
        gene_id = character(0),
        chr = character(0),
        strand = character(0),
        start = integer(0),
        end = integer(0),
        transcript_id = character(0),
        symbol = character(0)
    )

    expect_warning(
        result_empty_df <- query_exons_region(empty_exons, "chr1", 1, 1000),
        "no non-missing arguments"
    )
    expect_equal(nrow(result_empty_df), 0)

    expect_error(
        query_exons_gene_id(empty_exons, "AnyGene"),
        "gene AnyGene not found in exon annotation"
    )

    expect_error(
        query_exons_symbol(empty_exons, "AnySymbol"),
        "gene AnySymbol not found in exon annotation"
    )
})

# nocov start
# code adapted from zzz.R of readr
.onLoad <- function(libname, pkgname) {
    opt <- options()
    new_opt <- list(
        NanoMethViz.site_filter = 3L,
        NanoMethViz.highlight_col = "grey50"
    )

    to_set <- ! names(new_opt) %in% names(opt)

    if (any(to_set)) options(new_opt[to_set])
    invisible()
}
# nocov end

# UCSC Genomes ----

# exons
# the data.frame of exon information containing at least columns gene_id, chr, strand, start, end, transcript_id and symbol.

col_names <- c(
    "bin",
    "chrom",
    "chromStart",
    "chromEnd",
    "name",
    "length",
    "cpgNum",
    "gcNum",
    "perCpg",
    "perGc",
    "obsExp"
)

read_cgi_anno <- function(x) {
    x %>%
        read_tsv(col_names = col_names) %>%
        dplyr::rename(
            gene_id = name,
            chr = chrom,
            start = chromStart,
            end = chromEnd
        ) %>%
        mutate(
            transcript_id = gene_id,
            strand = "*",
            symbol = gene_id
        )
}

download_parse_and_save <- function(genome_name, url) {
    temp_path <- tempfile()
    download.file(url, temp_path)

    anno_name <- paste0("inst/cgi_", genome_name, ".rds")
    saveRDS(read_cgi_anno(temp_path), anno_name, compress = "xz")

    fs::file_delete(temp_path)
}

# mm10 ----
download_parse_and_save(
    "mm10",
    "https://hgdownload.soe.ucsc.edu/goldenPath/mm10/database/cpgIslandExt.txt.gz"
)

# GRCm39 ----
download_parse_and_save(
    "GRCm39",
    "https://hgdownload.soe.ucsc.edu/goldenPath/mm39/database/cpgIslandExt.txt.gz"
)

# hg19 ----
download_parse_and_save(
    "hg19",
    "https://hgdownload.soe.ucsc.edu/goldenPath/hg19/database/cpgIslandExt.txt.gz"
)

# hg38 ----
download_parse_and_save(
    "hg38",
    "https://hgdownload.soe.ucsc.edu/goldenPath/hg38/database/cpgIslandExt.txt.gz"
)

# T2T ----
download_parse_and_save(
    "t2t",
    "https://hgdownload.soe.ucsc.edu/gbdb/hs1/bbi/cpgIslandExt.bb"
)

# T2T Genome ----
temp_path <- tempfile()
download.file("https://hgdownload.soe.ucsc.edu/gbdb/hs1/bbi/cpgIslandExt.bb", temp_path)

anno_t2t <- rtracklayer::import.bb(temp_path) %>%
    as_tibble()

cgi_anno_t2t <- anno_t2t %>%
    dplyr::rename(
        gene_id = name,
        chr = seqnames
    ) %>%
    mutate(
        transcript_id = gene_id,
        strand = "*",
        symbol = gene_id
    )

anno_name <- paste0("inst/cgi_t2t.rds")
saveRDS(exon_anno_t2t_formatted, anno_name, compress = "xz")

fs::file_delete(temp_path)

#' Create filtered methylation file
#'
#' Create a filtered methylation file from an existing one.
#'
#' @param x the path to the methylation file or a NanoMethResult object.
#' @param output_file the output file to write results to (must end in .bgz).
#' @param ... filtering criteria given in dplyr syntax. Use methy_col_names()
#'   to get available column names.
#'
#' @return
#' invisibly returns 'output_file' if x is a file path, otherwise returns
#' NanoMethResult object with methy(x) replaced with filtered value.
#'
#' @export
#'
#' @examples
#' nmr <- load_example_nanomethresult()
#' output_file <- paste0(tempfile(), ".tsv.bgz")
#' filter_methy(nmr, output_file = output_file, chr == "chrX")
#' filter_methy(methy(nmr), output_file = output_file, chr == "chrX")
filter_methy <- function(x, output_file, ...) {
    if (is(x, "NanoMethResult")) {
        return_nmr <- TRUE
        nmr <- x
        input_file <- methy(x)
    } else {
        return_nmr <- FALSE
        input_file <- x
    }

    assertthat::assert_that(fs::file_exists(input_file))
    assertthat::assert_that(
        input_file != output_file,
        msg = "target file name must differ from original methylation file."
    )

    assertthat::assert_that(
        tools::file_ext(output_file) == "bgz",
        msg = "output_file must end with .bgz extension."
    )

    output_tsv <- stringr::str_remove(output_file, ".bgz")

    assert_that(is.dir(fs::path_dir(output_tsv)))
    file.create(path.expand(output_tsv))
    assert_that(is.writeable(output_tsv))

    lines_read <- 0
    lines_kept <- 0

    filter_fn <- function(x) {
        dplyr::filter(x, ...)
    }
    writer_fn <- function(x, i) {
        lines_read <<- lines_read + nrow(x)
        filtered_x <- filter_fn(x)
        lines_kept <<- lines_kept + nrow(filtered_x)
        readr::write_tsv(
            filtered_x,
            file = output_tsv,
            append = TRUE
        )
    }
    readr::local_edition(1) # temporary fix for vroom bad value
    readr::read_tsv_chunked(
        input_file,
        callback = readr::SideEffectChunkCallback$new(writer_fn),
        col_names = methy_col_names(),
        col_types = methy_col_types()
    )

    # assume filter does not change order of rows, therefore data should
    # still be sorted after filtering
    tabix_compress(x = output_tsv, index = TRUE)
    file.remove(output_tsv)

    kept_pct <- scales::percent(lines_kept / lines_read, accuracy = 0.01)
    lines_kept <- scales::comma(lines_kept)
    lines_read <- scales::comma(lines_read)

    message(glue::glue(
        "{lines_kept} of {lines_read} ({kept_pct}) entries kept after filtering"
    ))

    index_name <- paste0(output_file, ".tbi")
    message(glue::glue("results written to '{output_file}' along with index file '{index_name}'"))

    if (return_nmr) {
        methy(nmr) <- output_file
        invisible(nmr)
    } else {
        invisible(output_file)
    }
}

test_that("Filtering methylation files works", {
    nmr <- load_example_nanomethresult()
    output_file <- paste0(tempfile(), ".tsv.bgz")

    expect_no_warning(filter_methy(nmr, output_file = output_file, chr == "chrX"))
    expect_no_warning(filter_methy(nmr, output_file = output_file, chr == "chrX", pos > 1))
    expect_no_warning(filter_methy(methy(nmr), output_file = output_file, chr == "chrX"))

    expect_error(filter_methy("foo", output_file = output_file, chr == "chrX"))
})

#' Plot MDS
#'
#' Plot multi-dimensional scaling plot using algorithm of limma::plotMDS(). It is recommended this be done with the
#' log-methylation-ratio matrix generated by bsseq_to_log_methy_ratio().
#'
#' @param x a numeric matrix of log-methylation-ratio values where rows are genomic regions/features and columns
#'   are samples.
#' @param top the number of top variable features (genes/regions) used to calculate
#'   pairwise distances between samples.
#' @param plot_dims the numeric vector of the two dimensions to be plotted
#'   (default c(1, 2) for first two principal coordinates).
#' @param labels the character vector of labels for data points. By default uses
#'   column names of x, set to NULL to plot unlabeled points.
#' @param groups the character vector or numeric vector of groups that data points
#'   will be coloured by. For character vectors, discrete colours are used. For
#'   numeric vectors, continuous colour scale is used. By default, groups is NULL
#'   and points are not coloured.
#' @param legend_name the name for the colour legend.
#'
#' @return ggplot object of the MDS plot.
#'
#' @details
#' Multi-dimensional scaling (MDS) is a dimensionality reduction technique that
#' represents high-dimensional distances between samples in a lower-dimensional
#' space (typically 2D). This function uses limma's plotMDS algorithm, which
#' selects the most variable features and calculates leading log-fold-change
#' distances between samples. The resulting plot shows samples as points where
#' proximity indicates similarity in methylation profiles. When variance explained
#' is available, it is shown as percentages on the axis labels.
#'
#' @examples
#' nmr <- load_example_nanomethresult()
#' bss <- methy_to_bsseq(nmr)
#' lmr <- bsseq_to_log_methy_ratio(bss)
#' plot_mds(lmr)
#'
#' @importFrom limma plotMDS
#' @importFrom ggplot2 draw_key_point
#' @export
plot_mds <- function(x, top = 500, plot_dims = c(1, 2), labels = colnames(x), groups = NULL, legend_name = "group") {
    if (!is.null(labels)) {
        assertthat::assert_that(ncol(x) == length(labels))
    }

    if (!is.null(groups)) {
        assertthat::assert_that(ncol(x) == length(groups))
    }

    mds_res <- limma::plotMDS(
        x,
        top = top,
        plot = FALSE
    )

    var_exp1 <- round(100 * mds_res$var.explained[plot_dims[1]])
    var_exp2 <- round(100 * mds_res$var.explained[plot_dims[2]])

    if ("eigen.vectors" %in% names(mds_res)) {
        plot_data <- data.frame(
            dim1 = mds_res$eigen.vectors[, plot_dims[1]],
            dim2 = mds_res$eigen.vectors[, plot_dims[2]]
        )
        xlabel <- glue::glue("Leading logFC Dim {plot_dims[1]} ({var_exp1}%)")
        ylabel <- glue::glue("Leading logFC Dim {plot_dims[2]} ({var_exp2}%)")
    } else {
        plot_data <- data.frame(
            dim1 = mds_res$cmdscale.out[, plot_dims[1]],
            dim2 = mds_res$cmdscale.out[, plot_dims[2]]
        )
        xlabel <- glue::glue("Leading logFC Dim {plot_dims[1]}")
        ylabel <- glue::glue("Leading logFC Dim {plot_dims[2]}")
    }

    plot_data$labels <- labels
    plot_data$group <- groups

    p <- ggplot2::ggplot(
        plot_data,
        ggplot2::aes(x = .data$dim1, y = .data$dim2)
    )

    if (is.null(groups)) {
        if (is.null(labels)) {
            # no labels or groups
            p <- p + ggplot2::geom_point()
        } else {
            # no groups, but labels
            p <- p + ggplot2::geom_label(aes(label = labels))
        }
    } else {
        if (is.numeric(groups)) {
            # continuous colour palette ignores labels
            if (!is.null(labels)) {
                message("Ignoring labels as groups is numeric. Set `labels=NULL` to suppress this message.")
            }
            p <- p +
                ggplot2::geom_point(aes(colour = .data$group)) +
                ggplot2::scale_color_continuous(name = legend_name)
        } else {
            # discrete colour palette
            if (is.null(labels)) {
                # no labels, but groups
                p <- p + ggplot2::geom_point(aes(colour = .data$group)) +
                    ggplot2::guides(color = ggplot2::guide_legend(title = legend_name))
            } else {
                # labels and groups
                # key_glyph causes the legend to display points
                p <- p + ggplot2::geom_label(
                    aes(label = labels, colour = .data$group),
                    key_glyph = ggplot2::draw_key_point
                )
            }
        }
    }

    p +
        ggplot2::theme_bw() +
        ggplot2::xlab(xlabel) +
        ggplot2::ylab(ylabel) +
        ggplot2::scale_x_continuous(expand = c(.1, .1))
}

test_that("Plot MDS works", {
    nmr <- load_example_nanomethresult()
    bss <- methy_to_bsseq(nmr)
    lmr <- bsseq_to_log_methy_ratio(bss)

    expect_no_warning(plot_mds(lmr))

    expect_no_warning(plot_mds(lmr, plot_dims = c(2, 3)))
    expect_no_warning(plot_mds(lmr, labels = paste0("samples", 1:6)))
    expect_no_warning(plot_mds(lmr, labels = paste0("samples", 1:6), groups = rep(c("A", "B"), 3)))

    expect_no_warning(plot_mds(lmr, groups = rep(c("A", "B"), 3)))
    expect_no_warning(plot_mds(lmr, groups = rep(c("A", "B"), 3), legend_name = "group_name"))

    expect_message(plot_mds(lmr, groups = 1:6), "Ignoring labels as groups is numeric. Set `labels=NULL` to suppress this message.")
    expect_no_warning(plot_mds(lmr, groups = 1:6, labels = NULL))
})

library(NanoMethViz)
library(bsseq)
library(stringr)

nmr <- load_example_nanomethresult()
bss <- nmr %>% methy_to_bsseq()
samples <- colData(bss)$sample %>%
    str_replace(".*_(\\d)_(bl6|cast)$", "\\2 \\1")
groups <- str_extract(samples, "(bl6|cast)")

ggsave(
    "img/mds.png",
    plot_mds(
        bsseq_to_log_methy_ratio(
            bss,
            regions = exons_to_genes(NanoMethViz::exons(nmr))
        ),
        labels = samples,
        groups = groups
    ) +
        ggtitle("MDS plot"),
    height = 600, width = 800, units = "px", dpi = 150
)


ggsave(
    "img/peg3_gene.png",
    plot_gene(nmr, "Peg3"),
    height = 600, width = 800, units = "px", dpi = 150)

ggsave(
    "img/agg_genes.png",
    plot_agg_regions(
        nmr,
        regions = exons_to_genes(NanoMethViz::exons(nmr))
    ) +
        ggtitle("Aggregated profile over genes"),
    height = 600, width = 800, units = "px", dpi = 150)

#' Get exon annotations
#'
#' Helper functions are provided for obtaining exon annotations from relevant
#' TxDb packages on Bioconductor for the construction of NanoMethResults
#' objects.
#'
#' @name get_exons
#' @rdname get_exons
#'
#' @return tibble (data.frame) object containing exon annotation.
#'
#' @importFrom tibble as_tibble
#' @importFrom dplyr rename
NULL

#' Get exon annotations for Mus musculus (mm10)
#'
#' @return data.frame containing exons
#'
#' @examples
#' m_musculus_exons <- get_exons_mus_musculus()
#'
#' @importFrom tibble as_tibble
#' @importFrom dplyr rename
#' @export
get_exons_mus_musculus <- function() {
    package_check("Mus.musculus", "1.3.1")

    genes <-  AnnotationDbi::keys(Mus.musculus::Mus.musculus, "GENEID")
    exon_data <- suppressMessages(AnnotationDbi::select(
        Mus.musculus::Mus.musculus,
        keys = genes,
        keytype = "GENEID",
        columns = c(
            "GENEID",
            "TXID",
            "EXONCHROM",
            "EXONSTRAND",
            "EXONSTART",
            "EXONEND",
            "SYMBOL"
        )
    ))

    tibble::as_tibble(exon_data) %>%
        dplyr::rename(
            gene_id = "GENEID",
            chr = "EXONCHROM",
            strand = "EXONSTRAND",
            start = "EXONSTART",
            end = "EXONEND",
            transcript_id = "TXID",
            symbol = "SYMBOL"
        )
}

#' @rdname get_exons
#'
#' @examples
#' mm10_exons <- get_exons_mm10()
#'
#' @export
get_exons_mm10 <- function() {
    package_check(
        c("TxDb.Mmusculus.UCSC.mm10.knownGene", "org.Mm.eg.db"),
        c("3.10.0", "3.15.0")
    )

    txdb <- TxDb.Mmusculus.UCSC.mm10.knownGene::TxDb.Mmusculus.UCSC.mm10.knownGene

    genes <-  AnnotationDbi::keys(txdb, "GENEID")
    exon_data <- suppressMessages(AnnotationDbi::select(
        txdb,
        keys = genes,
        keytype = "GENEID",
        columns = c(
            "GENEID",
            "TXID",
            "EXONCHROM",
            "EXONSTRAND",
            "EXONSTART",
            "EXONEND"
        )
    ))

    symbols_data <- suppressMessages(AnnotationDbi::select(
            org.Mm.eg.db::org.Mm.eg.db,
            keys = genes,
            keytype = "ENTREZID",
            columns = c(
                "SYMBOL"
            )
        )) %>%
        dplyr::rename(GENEID = "ENTREZID")

    dplyr::left_join(exon_data, symbols_data, by = "GENEID", multiple = "all") %>%
        tibble::as_tibble() %>%
        dplyr::rename(
            gene_id = "GENEID",
            chr = "EXONCHROM",
            strand = "EXONSTRAND",
            start = "EXONSTART",
            end = "EXONEND",
            transcript_id = "TXID",
            symbol = "SYMBOL"
        )
}

#' @rdname get_exons
#'
#' @examples
#' grcm39_exons <- get_exons_grcm39()
#'
#' @export
get_exons_grcm39 <- function() {
    package_check(
        c("TxDb.Mmusculus.UCSC.mm39.refGene", "org.Mm.eg.db"),
        c("3.10.0", "3.15.0")
    )

    txdb <- TxDb.Mmusculus.UCSC.mm39.refGene::TxDb.Mmusculus.UCSC.mm39.refGene

    genes <-  AnnotationDbi::keys(txdb, "GENEID")
    exon_data <- suppressMessages(AnnotationDbi::select(
        txdb,
        keys = genes,
        keytype = "GENEID",
        columns = c(
            "GENEID",
            "TXID",
            "EXONCHROM",
            "EXONSTRAND",
            "EXONSTART",
            "EXONEND"
        )
    ))

    symbols_data <- suppressMessages(AnnotationDbi::select(
            org.Mm.eg.db::org.Mm.eg.db,
            keys = genes,
            keytype = "ENTREZID",
            columns = c(
                "SYMBOL"
            )
        )) %>%
        dplyr::rename(GENEID = "ENTREZID")

    dplyr::left_join(exon_data, symbols_data, by = "GENEID", multiple = "all") %>%
        tibble::as_tibble() %>%
        dplyr::rename(
            gene_id = "GENEID",
            chr = "EXONCHROM",
            strand = "EXONSTRAND",
            start = "EXONSTART",
            end = "EXONEND",
            transcript_id = "TXID",
            symbol = "SYMBOL"
        )
}

#' Get example exon annotations for mus musculus (mm10)
#'
#' This is a small subset of the exons returned by
#' \code{get_exons_mus_musculus()} for demonstrative purposes. It contains
#' the exons for the genes Brca1, Brca2, Impact, Meg3, Peg3 and Xist.
#'
#' @return data.frame containing exons
#'
#' @examples
#' example_exons <- get_example_exons_mus_musculus()
#'
#' @export
get_example_exons_mus_musculus <- function() {
    package_check("Mus.musculus", "1.3.1")

    genes <-  c("12189", "12190", "16210", "17263", "18616", "213742")
    exon_data <- suppressMessages(AnnotationDbi::select(
        Mus.musculus::Mus.musculus,
        keys = genes,
        keytype = "GENEID",
        columns = c(
            "GENEID",
            "TXID",
            "EXONCHROM",
            "EXONSTRAND",
            "EXONSTART",
            "EXONEND",
            "SYMBOL"
        )
    ))

    tibble::as_tibble(exon_data) %>%
        dplyr::rename(
            gene_id = "GENEID",
            chr = "EXONCHROM",
            strand = "EXONSTRAND",
            start = "EXONSTART",
            end = "EXONEND",
            transcript_id = "TXID",
            symbol = "SYMBOL"
        )
}

#' Get exon annotations for Homo sapiens (hg19)
#'
#' @return data.frame containing exons
#'
#' @examples
#' h_sapiens_exons <- get_exons_homo_sapiens()
#'
#' @importFrom tibble as_tibble
#' @importFrom dplyr rename
#' @export
get_exons_homo_sapiens <- function() {
    package_check("Homo.sapiens", "1.3.1")

    genes <-  AnnotationDbi::keys(Homo.sapiens::Homo.sapiens, "GENEID")
    exon_data <- suppressMessages(AnnotationDbi::select(
        Homo.sapiens::Homo.sapiens,
        keys = genes,
        keytype = "GENEID",
        columns = c(
            "GENEID",
            "TXID",
            "EXONCHROM",
            "EXONSTRAND",
            "EXONSTART",
            "EXONEND",
            "SYMBOL"
        )
    ))

    tibble::as_tibble(exon_data) %>%
        dplyr::rename(
            gene_id = "GENEID",
            chr = "EXONCHROM",
            strand = "EXONSTRAND",
            start = "EXONSTART",
            end = "EXONEND",
            transcript_id = "TXID",
            symbol = "SYMBOL"
        )

}

#' @rdname get_exons
#'
#' @examples
#' hg19_exons <- get_exons_hg19()
#'
#' @export
get_exons_hg19 <- function() {
    package_check(
        c("org.Hs.eg.db", "TxDb.Hsapiens.UCSC.hg19.knownGene"),
        c("3.15.0", "3.2.2"))

    txdb <- TxDb.Hsapiens.UCSC.hg19.knownGene::TxDb.Hsapiens.UCSC.hg19.knownGene
    genes <-  AnnotationDbi::keys(txdb, "GENEID")
    exon_data <- suppressMessages(AnnotationDbi::select(
        txdb,
        keys = genes,
        keytype = "GENEID",
        columns = c(
            "GENEID",
            "TXID",
            "EXONCHROM",
            "EXONSTRAND",
            "EXONSTART",
            "EXONEND"
        )
    ))

    symbols_data <- suppressMessages(AnnotationDbi::select(
            org.Hs.eg.db::org.Hs.eg.db,
            keys = genes,
            keytype = "ENTREZID",
            columns = c(
                "SYMBOL"
            )
        )) %>%
        dplyr::rename(GENEID = "ENTREZID")

    dplyr::left_join(exon_data, symbols_data, by = "GENEID", multiple = "all") %>%
        tibble::as_tibble() %>%
        dplyr::rename(
            gene_id = "GENEID",
            chr = "EXONCHROM",
            strand = "EXONSTRAND",
            start = "EXONSTART",
            end = "EXONEND",
            transcript_id = "TXID",
            symbol = "SYMBOL"
        )
}

#' @rdname get_exons
#'
#' @examples
#' hg38_exons <- get_exons_hg38()
#'
#' @export
get_exons_hg38 <- function() {
    package_check(
        c("org.Hs.eg.db", "TxDb.Hsapiens.UCSC.hg38.knownGene"),
        c("3.15.0", "3.15.0"))

    txdb <- TxDb.Hsapiens.UCSC.hg38.knownGene::TxDb.Hsapiens.UCSC.hg38.knownGene
    genes <-  AnnotationDbi::keys(txdb, "GENEID")
    exon_data <- suppressMessages(AnnotationDbi::select(
        txdb,
        keys = genes,
        keytype = "GENEID",
        columns = c(
            "GENEID",
            "TXID",
            "EXONCHROM",
            "EXONSTRAND",
            "EXONSTART",
            "EXONEND"
        )
    ))

    symbols_data <- suppressMessages(AnnotationDbi::select(
            org.Hs.eg.db::org.Hs.eg.db,
            keys = genes,
            keytype = "ENTREZID",
            columns = c(
                "SYMBOL"
            )
        )) %>%
        dplyr::rename(GENEID = "ENTREZID")

    dplyr::left_join(exon_data, symbols_data, by = "GENEID", multiple = "all") %>%
        tibble::as_tibble() %>%
        dplyr::rename(
            gene_id = "GENEID",
            chr = "EXONCHROM",
            strand = "EXONSTRAND",
            start = "EXONSTART",
            end = "EXONEND",
            transcript_id = "TXID",
            symbol = "SYMBOL"
        )
}

#' @rdname get_exons
#'
#' @examples
#' t2t_exons <- get_exons_t2t()
#'
#' @export
get_exons_t2t <- function() {
    anno_file <- system.file(
        "exons_t2t.rds",
        package = "NanoMethViz",
        mustWork = TRUE
    )

    readRDS(anno_file)
}

#' @importFrom utils installed.packages packageVersion
package_check <- function(packages, req_versions) {
    assertthat::assert_that(length(packages) == length(req_versions))

    any_missing <- FALSE
    for (i in seq_along(packages)) {
        package <- packages[i]
        req_version <- req_versions[i]
        if (!package %in% utils::installed.packages() || packageVersion(package) < req_version) {
            any_missing <- TRUE
            message(glue::glue(
                "required package '{package} (>= {req_version})' is not installed, please install using BiocManager::install(\"{package}\")"
            ))
        }
    }

    if (any_missing) {
        stop("please install required packages before continuing")
    }
}

#' Plot gene aggregate plot
#'
#' @param genes a character vector of gene symbols to include in aggregate plot.
#'   If NULL (default), all genes in exons(x) are used.
#' @inheritParams plot_agg_regions
#'
#' @return a ggplot object containing the aggregate methylation trend of genes.
#'
#' @details
#' This function creates an aggregate methylation profile across multiple genes by
#' scaling all genes to the same relative coordinates (0 to 1) and averaging
#' methylation levels at each relative position. Genes are optionally extended by
#' flanking regions specified by the flank parameter. The resulting plot shows
#' smoothed trends of average methylation probability from gene start to gene end,
#' with optional flanking regions.
#'
#' @export
#'
#' @examples
#' nmr <- load_example_nanomethresult()
#' plot_agg_genes(nmr)
#'
#' # Plot specific genes only
#' plot_agg_genes(nmr, genes = c("Peg3", "Impact"))
#'
#' # Group by sample
#' plot_agg_genes(nmr, group_col = "sample")
#'
plot_agg_genes <- function(
    x,
    genes = NULL,
    binary_threshold = 0.5,
    group_col = NULL,
    flank = 2000,
    stranded = TRUE,
    span = 0.05,
    palette = ggplot2::scale_colour_brewer(palette = "Set1")
) {
    assertthat::assert_that(nrow(exons(x)) > 0, msg = "no exon annotations found in object")

    gene_regions <- exons_to_genes(x)
    if (!is.null(genes)) {
        gene_regions <- gene_regions %>%
            filter(.data$symbol %in% genes)
    }

    plot_agg_regions(
        x,
        regions = gene_regions,
        binary_threshold = binary_threshold,
        group_col = group_col,
        flank = flank,
        stranded = stranded,
        span = span,
        palette = palette
    )
}

plot_agg_tss <- function(
    x,
    genes = NULL,
    binary_threshold = 0.5,
    group_col = NULL,
    flank = 2000,
    stranded = TRUE,
    span = 0.05,
    palette = ggplot2::scale_colour_brewer(palette = "Set1")
) {
    assertthat::assert_that(nrow(exons(x)) > 0, msg = "no exon annotations found in object")

    gene_regions <- exons_to_genes(x)
    if (!is.null(genes)) {
        gene_regions <- gene_regions %>%
            filter(.data$symbol %in% genes)
    }
    tss_regions <- gene_regions %>%
        mutate(
            start = case_when(
                strand == "+" ~ .data$start,
                strand == "-" ~ .data$end,
                TRUE ~ .data$start)
        ) %>%
        mutate(
            start = .data$start - flank,
            end = .data$start + 2 * flank
        )
    kb_marker <- round(flank / 1000, 1)
    labels <- c(
        glue::glue("-{kb_marker}kb"),
        "TSS",
        glue::glue("+{kb_marker}kb")
    )

    p <- plot_agg_regions(
        x,
        regions = tss_regions,
        binary_threshold = binary_threshold,
        group_col = group_col,
        flank = 0,
        stranded = stranded,
        span = span,
        palette = palette
    )
    # hack to delete existing to avoid warning
    p$scales$scales[[which(p$scales$find("x"))]] <- NULL

    p + ggplot2::scale_x_continuous(
        name = glue::glue(""),
        breaks = c(0, 0.5, 1),
        limits = c(0, 1),
        labels = labels
    )
}

plot_agg_tes <- function(
    x,
    genes = NULL,
    binary_threshold = 0.5,
    group_col = NULL,
    flank = 2000,
    stranded = TRUE,
    span = 0.05,
    palette = ggplot2::scale_colour_brewer(palette = "Set1")
) {
    assertthat::assert_that(nrow(exons(x)) > 0, msg = "no exon annotations found in object")

    gene_regions <- exons_to_genes(x)
    if (!is.null(genes)) {
        gene_regions <- gene_regions %>%
            filter(.data$symbol %in% genes)
    }
    tes_regions <- gene_regions %>%
        mutate(
            start = case_when(
                strand == "+" ~ .data$end,
                strand == "-" ~ .data$start,
                TRUE ~ .data$end)
        ) %>%
        mutate(
            start = .data$start - flank,
            end = .data$start + 2 * flank
        )
    kb_marker <- round(flank / 1000, 1)
    labels <- c(
        glue::glue("-{kb_marker}kb"),
        "TES",
        glue::glue("+{kb_marker}kb")
    )

    p <- plot_agg_regions(
        x,
        regions = tes_regions,
        binary_threshold = binary_threshold,
        group_col = group_col,
        flank = 0,
        stranded = stranded,
        span = span,
        palette = palette
    )
    # hack to delete existing to avoid warning
    p$scales$scales[[which(p$scales$find("x"))]] <- NULL

    p + ggplot2::scale_x_continuous(
        name = glue::glue(""),
        breaks = c(0, 0.5, 1),
        limits = c(0, 1),
        labels = labels
    )
}

test_that("Plotting gene aggregates works", {
    nmr <- load_example_nanomethresult()
    expect_no_error(plot_agg_genes(nmr))
    expect_no_error(plot_agg_tss(nmr))
    expect_no_error(plot_agg_tes(nmr))

    # empty exons should throw error
    exons(nmr) <- tibble::tibble(
        gene_id = character(),
        chr = character(),
        strand = character(),
        start = integer(),
        end = integer(),
        transcript_id = character(),
        symbol = character()
    )
    expect_error(plot_agg_genes(nmr))
    expect_error(plot_agg_tss(nmr))
    expect_error(plot_agg_tes(nmr))
})

#' @details The main plotting functions in this package are [plot_gene()] and [plot_region()].
#'
#'   * See \code{vignette("UserGuide", package = "NanoMethViz")} for documentation of how to use this package.
#'
#'
#' @docType package
#'
#' @importFrom dplyr %>%
#' @importFrom methods new .valueClassTest is
#' @importFrom rlang .data
#' @importFrom readr read_tsv local_edition
#' @importFrom ggplot2 aes geom_rect geom_segment geom_text ggplot theme_void
#'   xlim ylim ggplot_build rel unit geom_vline
#' @importFrom dplyr filter group_by ungroup inner_join mutate n select summarise case_when
#' @importFrom tidyr unnest
#' @importFrom glue glue
#' @importFrom assertthat assert_that is.readable is.writeable is.dir is.string
#' @importFrom stringr str_extract
#' @importFrom scales label_number cut_si
#' @importFrom readr cols col_character col_integer col_logical col_double
#' @importFrom tibble tibble as_tibble add_column
#' @importFrom withr defer
#' @importFrom R.utils gunzip
#' @importFrom utils read.delim
#' @import cpp11
#' @import patchwork
#' @import assertthat
#' @import fs
#' @import Rcpp
#' @useDynLib NanoMethViz
#' @keywords internal
"_PACKAGE"

# The following block is used by usethis to automatically manage
# roxygen namespace tags. Modify with care!
## usethis namespace: start
## usethis namespace: end
NULL

#' @importFrom purrr map_lgl
assert_readable <- function(x) {
    readable <- purrr::map_lgl(x, fs::file_exists)

    if (any(!readable)) {
        unreadables <- x[!readable]
        if (length(unreadables) == 1) {
            stop(glue::glue(
                "File path '{unreadables}' does not exist.\n",
                "Please check that the file path is correct and the file exists."
            ))
        } else {
            unreadables <- paste(glue::glue("'{unreadables}'"), collapse = ", ")
            stop(glue::glue(
                "File paths {unreadables} do not exist.\n",
                "Please check that all file paths are correct and the files exist."
            ))
        }
    }
}

#' @importFrom purrr map_lgl
assert_has_index <- function(x) {
    has_index <- purrr::map_lgl(paste0(x, ".bai"), fs::file_exists)

    if (any(!has_index)) {
        no_index <- x[!has_index]
        if (length(no_index) == 1) {
            stop(glue::glue(
                "BAM file '{no_index}' is missing its index file (.bai).\n",
                "To fix this, run: samtools index {no_index}\n",
                "Or ensure the .bai file is in the same directory as the BAM file."
            ))
        } else {
            no_index_list <- paste(glue::glue("'{no_index}'"), collapse = ", ")
            stop(glue::glue(
                "BAM files {no_index_list} are missing their index files (.bai).\n",
                "To fix this, run 'samtools index' on each file:\n",
                "{paste0('samtools index ', no_index, collapse = '\n')}"
            ))
        }
    }
}

#' Assert that file paths are readable
#'
#' This function checks whether all provided file paths exist and are readable.
#' If any file paths do not exist, it throws an informative error message.
#'
#' @param x A character vector of file paths to check for existence
#'
#' @return Nothing if all files exist, otherwise throws an error
#'
#' @keywords internal
assert_valid_genomic_coords <- function(chr, start, end, allow_equal = FALSE) {
    # Check for missing values
    if (any(is.na(chr))) {
        stop("Chromosome cannot be NA. Please provide valid chromosome identifiers.")
    }

    if (any(is.na(start)) || any(is.na(end))) {
        stop("Start and end positions cannot be NA. Please provide valid genomic coordinates.")
    }

    # Convert to numeric if needed
    if (!is.numeric(start)) {
        start_numeric <- suppressWarnings(as.numeric(start))
        if (any(is.na(start_numeric))) {
            stop(glue::glue(
                "Start position must be numeric. Got: {paste(start[is.na(start_numeric)], collapse = ', ')}\n",
                "Please provide numeric start positions."
            ))
        }
        start <- start_numeric
    }

    if (!is.numeric(end)) {
        end_numeric <- suppressWarnings(as.numeric(end))
        if (any(is.na(end_numeric))) {
            stop(glue::glue(
                "End position must be numeric. Got: {paste(end[is.na(end_numeric)], collapse = ', ')}\n",
                "Please provide numeric end positions."
            ))
        }
        end <- end_numeric
    }

    # Check for positive coordinates
    if (any(start < 1)) {
        stop(glue::glue(
            "Start positions must be positive (>= 1). Got: {paste(start[start < 1], collapse = ', ')}\n",
            "Genomic coordinates are 1-based in this package."
        ))
    }

    if (any(end < 1)) {
        stop(glue::glue(
            "End positions must be positive (>= 1). Got: {paste(end[end < 1], collapse = ', ')}\n",
            "Genomic coordinates are 1-based in this package."
        ))
    }

    # Check start <= end
    comparison <- if (allow_equal) start <= end else start < end
    if (any(!comparison)) {
        invalid_idx <- which(!comparison)
        error_msg <- if (allow_equal) {
            "Start positions must be <= end positions."
        } else {
            "Start positions must be < end positions."
        }

        stop(glue::glue(
            "{error_msg}\n",
            "Invalid coordinates found:\n",
            "{paste(glue::glue('  {chr[invalid_idx]}:{start[invalid_idx]}-{end[invalid_idx]}'), collapse = '\n')}\n",
            "Please check your genomic coordinates."
        ))
    }
}

#' Validate sample annotation data frame
#' @param samples sample annotation data.frame
#' @param required_cols required column names
#' @param context description of where this is being used for error messages
#'
#' @keywords internal
assert_valid_samples <- function(samples, required_cols = c("sample", "group"), context = "sample annotation") {
    if (!is.data.frame(samples)) {
        stop(glue::glue(
            "The {context} must be a data.frame.\n",
            "Got object of class: {class(samples)[1]}\n",
            "Please provide a data.frame with columns: {paste(required_cols, collapse = ', ')}"
        ))
    }

    if (nrow(samples) == 0) {
        stop(glue::glue(
            "The {context} data.frame is empty.\n",
            "Please provide sample information with columns: {paste(required_cols, collapse = ', ')}"
        ))
    }

    # Check required columns
    assert_has_columns(samples, required_cols)

    # Check for duplicate sample names
    if ("sample" %in% required_cols && any(duplicated(samples$sample))) {
        duplicates <- samples$sample[duplicated(samples$sample)]
        stop(glue::glue(
            "Found duplicate sample names in {context}: {paste(unique(duplicates), collapse = ', ')}\n",
            "Each sample must have a unique identifier."
        ))
    }

    # Check for empty/NA sample names
    if ("sample" %in% required_cols) {
        if (any(is.na(samples$sample) | samples$sample == "")) {
            stop(glue::glue(
                "Sample names cannot be empty or NA in {context}.\n",
                "Please provide valid sample identifiers."
            ))
        }
    }
}

#' Validate exon annotation data frame
#' @param exons exon annotation data.frame
#'
#' @keywords internal
assert_valid_exons <- function(exons) {
    if (!is.data.frame(exons)) {
        stop(glue::glue(
            "Exon annotation must be a data.frame.\n",
            "Got object of class: {class(exons)[1]}\n",
            "Please provide a data.frame with required exon columns."
        ))
    }

    required_cols <- c("gene_id", "chr", "strand", "start", "end", "transcript_id", "symbol")
    assert_has_columns(exons, required_cols)

    if (nrow(exons) > 0) {
        # Validate genomic coordinates in exons
        tryCatch({
            assert_valid_genomic_coords(exons$chr, exons$start, exons$end, allow_equal = TRUE)
        }, error = function(e) {
            stop(glue::glue(
                "Invalid genomic coordinates found in exon annotation:\n",
                "{e$message}"
            ))
        })

        # Check strand values
        valid_strands <- c("+", "-", "*")
        if (any(!exons$strand %in% valid_strands)) {
            invalid_strands <- unique(exons$strand[!exons$strand %in% valid_strands])
            stop(glue::glue(
                "Invalid strand values in exon annotation: {paste(invalid_strands, collapse = ', ')}\n",
                "Valid strand values are: {paste(valid_strands, collapse = ', ')}"
            ))
        }
    }
}

#' Enhanced validation with better error messages
#' @param data_path path to methylation data file
#' @param samples sample annotation data.frame
#' @param sample_check_limit number of entries to check for sample matching
#'
#' @keywords internal
assert_valid_methy_samples <- function(data_path, samples, sample_check_limit = 1000) {
    tryCatch({
        # Check file format by reading header
        head_values <- read.table(
            gzfile(data_path),
            col.names = methy_col_names(),
            nrows = sample_check_limit,
            comment.char = ""
        )
    }, error = function(e) {
        stop(glue::glue(
            "Failed to read methylation data file '{data_path}'.\n",
            "Error: {e$message}\n",
            "Please check that the file is a valid tabix-indexed methylation file."
        ))
    })

    if (nrow(head_values) == 0) {
        stop(glue::glue(
            "Methylation data file '{data_path}' appears to be empty.\n",
            "Please check that the file contains methylation data."
        ))
    }

    # Check sample matching
    data_samples <- unique(head_values$sample)
    anno_samples <- samples$sample

    matched_samples <- intersect(data_samples, anno_samples)

    if (length(matched_samples) == 0) {
        stop(glue::glue(
            "No sample names from the data file match the sample annotation.\n",
            "Data samples (first {length(data_samples)}): {paste(data_samples, collapse = ', ')}\n",
            "Annotation samples: {paste(anno_samples, collapse = ', ')}\n",
            "Please ensure sample names match between your data and annotation."
        ))
    }

    # Warn about unmatched samples
    unmatched_data <- setdiff(data_samples, anno_samples)
    unmatched_anno <- setdiff(anno_samples, data_samples)

    if (length(unmatched_data) > 0) {
        warning(glue::glue(
            "Found {length(unmatched_data)} samples in data file not present in annotation: {paste(unmatched_data, collapse = ', ')}\n",
            "These samples will be ignored in analysis."
        ))
    }

    if (length(unmatched_anno) > 0) {
        warning(glue::glue(
            "Found {length(unmatched_anno)} samples in annotation not present in data file: {paste(unmatched_anno, collapse = ', ')}\n",
            "These samples will have no data for analysis."
        ))
    }

    message(glue::glue(
        "Successfully matched {length(matched_samples)} samples between data and annotation."
    ))
}

test_that("genomic coordinate validation works", {
    # Test valid coordinates
    expect_silent(assert_valid_genomic_coords("chr1", 1, 100))
    expect_silent(assert_valid_genomic_coords(c("chr1", "chr2"), c(1, 50), c(100, 150)))

    # Test invalid: start > end
    expect_error(
        assert_valid_genomic_coords("chr1", 100, 50),
        "Start positions must be < end positions"
    )

    # Test invalid: negative coordinates
    expect_error(
        assert_valid_genomic_coords("chr1", -5, 100),
        "Start positions must be positive"
    )

    expect_error(
        assert_valid_genomic_coords("chr1", 1, -5),
        "End positions must be positive"
    )

    # Test invalid: NA chromosome
    expect_error(
        assert_valid_genomic_coords(NA, 1, 100),
        "Chromosome cannot be NA"
    )

    # Test invalid: NA coordinates
    expect_error(
        assert_valid_genomic_coords("chr1", NA, 100),
        "Start and end positions cannot be NA"
    )

    # Test string to numeric conversion
    expect_silent(assert_valid_genomic_coords("chr1", "1", "100"))

    # Test invalid string conversion
    expect_error(
        assert_valid_genomic_coords("chr1", "abc", 100),
        "Start position must be numeric"
    )

    # Test allow_equal = TRUE
    expect_silent(assert_valid_genomic_coords("chr1", 100, 100, allow_equal = TRUE))
    expect_error(
        assert_valid_genomic_coords("chr1", 100, 100, allow_equal = FALSE),
        "Start positions must be < end positions"
    )
})

test_that("sample validation works", {
    # Test valid samples
    valid_samples <- data.frame(sample = c("A", "B", "C"), group = c("1", "2", "1"))
    expect_silent(assert_valid_samples(valid_samples))

    # Test duplicate sample names
    dup_samples <- data.frame(sample = c("A", "B", "A"), group = c("1", "2", "1"))
    expect_error(
        assert_valid_samples(dup_samples),
        "Found duplicate sample names"
    )

    # Test empty sample names
    empty_samples <- data.frame(sample = c("A", "", "C"), group = c("1", "2", "1"))
    expect_error(
        assert_valid_samples(empty_samples),
        "Sample names cannot be empty or NA"
    )

    # Test missing required columns
    missing_cols <- data.frame(sample = c("A", "B", "C"))
    expect_error(
        assert_valid_samples(missing_cols),
        "columns missing from"
    )

    # Test non-data.frame input
    expect_error(
        assert_valid_samples("not_a_df"),
        "must be a data.frame"
    )

    # Test empty data.frame
    empty_df <- data.frame(sample = character(0), group = character(0))
    expect_error(
        assert_valid_samples(empty_df),
        "data.frame is empty"
    )
})

test_that("exon validation works", {
    # Test valid exons
    valid_exons <- data.frame(
        gene_id = "GENE1",
        chr = "chr1",
        strand = "+",
        start = 1000,
        end = 2000,
        transcript_id = "T1",
        symbol = "GENE1"
    )
    expect_silent(assert_valid_exons(valid_exons))

    # Test invalid strand
    invalid_strand <- valid_exons
    invalid_strand$strand <- "X"
    expect_error(
        assert_valid_exons(invalid_strand),
        "Invalid strand values"
    )

    # Test invalid coordinates in exons
    invalid_coords <- valid_exons
    invalid_coords$start <- 2000
    invalid_coords$end <- 1000
    expect_error(
        assert_valid_exons(invalid_coords),
        "Invalid genomic coordinates found in exon annotation"
    )

    # Test missing required columns
    missing_cols <- valid_exons[, -which(names(valid_exons) == "symbol")]
    expect_error(
        assert_valid_exons(missing_cols),
        "columns missing from"
    )

    # Test non-data.frame input
    expect_error(
        assert_valid_exons("not_a_df"),
        "must be a data.frame"
    )

    # Test empty exons (should not error)
    empty_exons <- data.frame(
        gene_id = character(0),
        chr = character(0),
        strand = character(0),
        start = integer(0),
        end = integer(0),
        transcript_id = character(0),
        symbol = character(0)
    )
    expect_silent(assert_valid_exons(empty_exons))
})

test_that("error messages provide helpful guidance", {
    # Test that error messages contain helpful suggestions
    expect_error(
        assert_valid_genomic_coords("chr1", 0, 100),
        "Genomic coordinates are 1-based in this package"
    )

    expect_error(
        assert_valid_samples(data.frame(sample = c("A", "A"), group = c("1", "2"))),
        "Each sample must have a unique identifier"
    )
})

## ----setup, include=FALSE-----------------------------------------------------
knitr::opts_chunk$set(echo = TRUE)
knitr::opts_chunk$set(fig.width = 7)
knitr::opts_chunk$set(fig.height = 5)
knitr::opts_chunk$set(dpi = 72)

# preload to avoid loading messages
library(NanoMethViz)
library(dplyr)
exon_tibble <- get_exons_mm10()

## ----eval = FALSE-------------------------------------------------------------
# if (!requireNamespace("BiocManager", quietly = TRUE))
#     install.packages("BiocManager")
# 
# BiocManager::install("NanoMethViz")

## -----------------------------------------------------------------------------
library(NanoMethViz)
library(dplyr)

## -----------------------------------------------------------------------------
# construct with a ModBamFiles object as the methylation data
mbr <- ModBamResult(
    methy = ModBamFiles(
        samples = "sample1",
        system.file(package = "NanoMethViz", "peg3.bam")
    ),
    samples = data.frame(
        sample = "sample1",
        group = 1
    ),
    exons = exon_tibble
)

# use in the same way as you would a NanoMethResult object
plot_gene(mbr, "Peg3", heatmap = TRUE)

## -----------------------------------------------------------------------------
methy_calls <- system.file(package = "NanoMethViz",
    c("sample1_nanopolish.tsv.gz", "sample2_nanopolish.tsv.gz"))

## ----message=F----------------------------------------------------------------
# create a temporary file to store the converted data
methy_tabix <- file.path(tempdir(), "methy_data.bgz")
samples <- c("sample1", "sample2")

# you should see messages when running this yourself
create_tabix_file(methy_calls, methy_tabix, samples)

# don't do this with full datasets, it will take a long time
# we have to use gzfile to tell R that we have a gzip compressed file
methy_data <- read.table(
    gzfile(methy_tabix), col.names = methy_col_names(), nrows = 6)

methy_data

## -----------------------------------------------------------------------------
# methylation data stored in tabix file
methy <- system.file(package = "NanoMethViz", "methy_subset.tsv.bgz")

# tabix is just a special gzipped tab-separated-values file
read.table(gzfile(methy), col.names = methy_col_names(), nrows = 6)

## -----------------------------------------------------------------------------
# helper function extracts exons from TxDb package
exon_tibble <- get_exons_mm10()

head(exon_tibble)

## -----------------------------------------------------------------------------
sample <- c(
  "B6Cast_Prom_1_bl6", "B6Cast_Prom_1_cast",
  "B6Cast_Prom_2_bl6", "B6Cast_Prom_2_cast",
  "B6Cast_Prom_3_bl6", "B6Cast_Prom_3_cast"
)

group <- c("bl6", "cast", "bl6", "cast", "bl6", "cast")

sample_anno <- data.frame(sample, group, stringsAsFactors = FALSE)

sample_anno

## -----------------------------------------------------------------------------
nmr <- NanoMethResult(methy, sample_anno, exon_tibble)

## -----------------------------------------------------------------------------
plot_gene(nmr, "Peg3")

## -----------------------------------------------------------------------------
# loading saved results from previous bsseq analysis
bsseq_dmr <- read.table(
    system.file(package = "NanoMethViz", "dmr_subset.tsv.gz"),
    sep = "\t",
    header = TRUE,
    stringsAsFactors = FALSE
)

## -----------------------------------------------------------------------------
plot_gene(nmr, "Peg3", anno_regions = bsseq_dmr)

## ----message = FALSE----------------------------------------------------------
nmr <- load_example_nanomethresult()
bss <- methy_to_bsseq(nmr)

bss

## -----------------------------------------------------------------------------
gene_regions <- exons_to_genes(NanoMethViz::exons(nmr))
edger_mat <- bsseq_to_edger(bss, gene_regions)

edger_mat

## -----------------------------------------------------------------------------
anno <- rtracklayer::import(system.file(package = "NanoMethViz", "c_elegans.gtf.gz"))

head(anno)

## -----------------------------------------------------------------------------
anno <- anno %>%
    as.data.frame() %>%
    dplyr::rename(
        chr = "seqnames",
        symbol = "gene_name"
    ) %>%
    dplyr::select("gene_id", "chr", "strand", "start", "end", "transcript_id", "symbol")

head(anno)

## ----message = FALSE----------------------------------------------------------
nmr <- load_example_nanomethresult()

plot_gene(nmr, "Peg3")

## -----------------------------------------------------------------------------
new_exons <- NanoMethViz::exons(nmr) %>%
    exons_to_genes() %>%
    mutate(transcript_id = gene_id)

NanoMethViz::exons(nmr) <- new_exons

plot_gene(nmr, "Peg3")

## ----message = FALSE----------------------------------------------------------
# convert to bsseq
bss <- methy_to_bsseq(nmr)
bss

## -----------------------------------------------------------------------------
# create gene annotation from exon annotation
gene_anno <- exons_to_genes(NanoMethViz::exons(nmr))

# create log-methylation-ratio matrix
lmr <- bsseq_to_log_methy_ratio(bss, regions = gene_anno)

## -----------------------------------------------------------------------------
plot_mds(lmr) +
    ggtitle("MDS Plot")

plot_pca(lmr) +
    ggtitle("PCA Plot")

## -----------------------------------------------------------------------------
new_labels <- gsub("B6Cast_Prom_", "", colnames(lmr))
new_labels <- gsub("(\\d)_(.*)", "\\2 \\1", new_labels)
groups <- gsub(" \\d", "", new_labels)

plot_mds(lmr, labels = new_labels, groups = groups) +
    ggtitle("MDS Plot") +
    scale_colour_brewer(palette = "Set1")

## -----------------------------------------------------------------------------
plot_mds(lmr, labels = NULL, groups = groups) +
    ggtitle("MDS Plot") +
    scale_colour_brewer(palette = "Set1")

## ----eval = FALSE-------------------------------------------------------------
# options("NanoMethViz.site_filter" = 5)

## ----eval = FALSE-------------------------------------------------------------
# options("NanoMethViz.highlight_col" = "red")

#' Create BSSeq object from methylation tabix file
#'
#' @param methy the NanoMethResult object or path to the methylation tabix file.
#' @param out_folder the folder to store intermediate files. One file is created
#'   for each sample and contains columns "chr", "pos", "total" and
#'   "methylated".
#' @param verbose TRUE if progress messages are to be printed
#'
#' @return a BSSeq object.
#' @export
#'
#' @examples
#' nmr <- load_example_nanomethresult()
#' bsseq <- methy_to_bsseq(nmr)
methy_to_bsseq <- function(
    methy,
    out_folder = tempdir(),
    verbose = TRUE
) {
    if (is(methy, "NanoMethResult")) {
        methy_path <- NanoMethViz::methy(methy)
    } else {
        methy_path <- methy
        assert_that(fs::file_exists(methy_path))
    }

    timed_log("creating intermediate files...", verbose = verbose)

    files <- convert_methy_to_dss(methy_path, out_folder)

    if (is(methy, "NanoMethResult")) {
        sample_anno <- NanoMethViz::samples(methy)
    } else {
        sample_anno <- tibble::tibble(sample = files$sample)
    }

    if (verbose) {
        timed_log("creating bsseq object...")
    }

    out <- create_bsseq_from_files(files$file_path, sample_anno, verbose = verbose)

    if (verbose) {
        timed_log("done")
    }

    out
}

convert_methy_to_dss <- function(
    methy,
    out_folder
) {
    assert_that(
        fs::file_exists(methy)
    )

    if (!fs::dir_exists(out_folder)) {
        fs::dir_create(out_folder, recurse = TRUE)
    }

    samples <- convert_methy_to_dss_cpp(
        fs::path_expand(methy),
        fs::path_expand(out_folder)
    )

    data.frame(
        sample = samples,
        file_path = path(out_folder, paste0(samples, ".txt"))
    )
}

#' @importFrom S4Vectors DataFrame
#' @importFrom SummarizedExperiment colData
#' @importFrom purrr map
#' @importFrom dplyr select distinct arrange mutate
#' @importFrom bsseq BSseq
create_bsseq_from_files <- function(paths, sample_anno, verbose = TRUE) {
    readr::local_edition(1) # temporary fix for vroom bad value
    read_dss <- purrr::partial(
        readr::read_tsv,
        col_types = readr::cols(
            chr = readr::col_character(),
            pos = readr::col_double(),
            total = readr::col_double(),
            methylated = readr::col_double()
        )
    )

    samples <- sample_anno$sample

    if (verbose) {
        timed_log("reading in parsed data...")
    }
    # read in data
    dat <- purrr::map(paths, read_dss)

    if (verbose) {
        timed_log("constructing matrices...")
    }
    # get unique positions
    combine_distinct_gpos <- function(x, y) {
        rbind(
            dplyr::select(x, "chr", "pos"),
            dplyr::select(y, "chr", "pos")
        ) %>%
            dplyr::distinct()
    }

    unique_pos_df <- purrr::reduce(
            dat,
            combine_distinct_gpos) %>%
        dplyr::arrange(.data$chr, .data$pos) %>%
        dplyr::mutate(id = paste(.data$chr, .data$pos))

    # create methylation matrix
    methylation_mat <- matrix(
        0,
        nrow = nrow(unique_pos_df),
        ncol = length(dat),
        dimnames = list(NULL, samples))

    for (i in seq_along(dat)) {
        row_inds <- paste(dat[[i]]$chr, dat[[i]]$pos) %>%
            factor(levels = unique_pos_df$id)

        methylation_mat[row_inds, i] <- dat[[i]]$methylated
    }

    # create coverage matrix
    cov_mat <- matrix(
        0,
        nrow = nrow(unique_pos_df),
        ncol = length(dat),
        dimnames = list(NULL, samples))

    for (i in seq_along(dat)) {
        row_inds <- paste(dat[[i]]$chr, dat[[i]]$pos) %>%
            factor(levels = unique_pos_df$id)

        cov_mat[row_inds, i] <- dat[[i]]$total
    }

    # create BSseq object
    result <- bsseq::BSseq(
        chr = unique_pos_df$chr,
        pos = unique_pos_df$pos,
        M = methylation_mat,
        Cov = cov_mat,
        sampleNames = samples
    )

    SummarizedExperiment::colData(result) <- S4Vectors::DataFrame(sample_anno)

    rownames(SummarizedExperiment::colData(result)) <- samples

    result
}

test_that("Importing nanopolish works", {
    # setup
    methy_calls <- system.file(package = "NanoMethViz",
        c("sample1_nanopolish.tsv.gz", "sample2_nanopolish.tsv.gz"), mustWork = FALSE)
    temp_file <- paste0(tempfile(), ".tsv.bgz")
    withr::defer(file.remove(temp_file))

    # test
    expect_message(create_tabix_file(methy_calls, temp_file))
    expect_s4_class(methy_to_bsseq(temp_file), "BSseq")
})

test_that("Importing megalodon works", {
    # setup
    methy_calls <- system.file(package = "NanoMethViz",
       "megalodon_calls.txt.gz", mustWork = FALSE)
    temp_file <- paste0(tempfile(), ".tsv.bgz")
    withr::defer(file.remove(temp_file))

    # test
    expect_message(create_tabix_file(methy_calls, temp_file))
})

test_that("plot_pca works correctly", {
  # create test data
    nmr <- load_example_nanomethresult()
    bss <- methy_to_bsseq(nmr)
    x <- bsseq_to_log_methy_ratio(bss)
    groups <- samples(nmr)$group
    labels <- colnames(x)

    expect_no_error(plot_pca(x))
    expect_no_error(plot_pca(x, labels = labels, groups = groups))
    expect_error(plot_pca(x, groups = c(groups, "foo")))
    expect_error(plot_pca(x, labels = c(labels, "foo")))
})

plot_clustered_reads <- function(x, chr, start, end, min_pts = 5, title = glue::glue("{chr}:{start}-{end}")) {
    methy_data <- query_methy(x, chr, start, end) %>%
        dplyr::filter(.data$pos > start & .data$pos < end) %>%
        dplyr::inner_join(samples(x), by = "sample")

    cluster_res <- cluster_reads(x, chr, start, end, min_pts = min_pts)

    append_read_group <- function(x, k) {
        x$read_group <- paste0(k, stacked_interval_inds(x))
        x
    }

    read_data <- methy_data %>%
        dplyr::group_by(.data$read_name) %>%
        dplyr::summarise(start = min(.data$pos), end = max(.data$pos))

    group_data <- methy_data %>%
        dplyr::group_by(.data$read_name) %>%
        dplyr::summarise(group = unique(.data$group))

    read_data <- dplyr::inner_join(
        read_data,
        group_data,
        by = "read_name",
        multiple = "all",
        suffix = c("_read", "_group")
    )

    read_data <- read_data %>%
        dplyr::group_by(.data$group) %>%
        dplyr::group_modify(append_read_group) %>%
        dplyr::ungroup()

    methy_data <- dplyr::inner_join(
        methy_data,
        read_data,
        by = dplyr::join_by("read_name", "group")
    )

    p1 <- plot_methylation_data(
        methy_data,
        sample_anno = samples(x),
        read_anno = cluster_res,
        chr = chr,
        start = start,
        end = end,
        group_col = "cluster_id",
        title = title,
        points = TRUE
    ) +
        ggplot2::coord_cartesian(xlim = c(start, end), expand = FALSE)

    heatmap_data <- dplyr::inner_join(
        methy_data,
        cluster_res,
        by = dplyr::join_by("read_name", "strand", "start", "end")
    )

    p2 <- plot_methy_data_heatmap(
        methy_data = heatmap_data,
        pos_style = "to_scale",
        subsample = 30,
        group_col = "cluster_id"
    ) +
        ggplot2::coord_cartesian(xlim = c(start, end), expand = TRUE)


    p1 / p2
}

test_that("cluster_reads assertions", {
    x <- load_example_modbamresult()
    chr <- "chr7"
    start <- 6713892
    end <- 6720421
    min_pts <- 5

    # Assertion tests
    expect_error(cluster_reads(x, chr, start, end, "min_pts"))
    expect_error(cluster_reads(x, chr, "start", end, min_pts))
    expect_error(cluster_reads(x, chr, start, end, -2))

    # Successful assertion test
    expect_no_error(cluster_reads(x, chr, start, end, min_pts))
    expect_no_warning(plot_clustered_reads(x, chr, start, end))
    expect_error(plot_clustered_reads(x, chr, start, end, min_pts = 1))
})

#' ModBamFiles class
#'
#' This is a class for holding information about modbam files. It is a
#' data.frame containing information about samples and paths to modbam files.
#'
#' @export
setClass(
    "ModBamFiles",
    contains = "data.frame"
)

#' Constructor for a ModBamFiles object
#'
#' This function creates a ModBamFiles object containing information about the
#' samples and file paths. This constructor checks that the files are readable
#' and have an index.
#'
#' @param samples a character vector with the names of the samples.
#' @param paths a character vector with the file paths for the BAM files.
#'
#' @return A ModBamFiles object with the sample and path information.
#'
#' @importFrom tibble tibble
#' @export
ModBamFiles <- function(samples, paths) {
    # validate inputs
    if (length(samples) != length(paths)) {
        stop(glue::glue(
            "Length of samples ({length(samples)}) must equal length of paths ({length(paths)}).\n",
            "Please provide matching samples and paths."
        ))
    }

    if (length(samples) == 0) {
        stop("At least one sample and path must be provided.")
    }

    if (any(is.na(samples) | samples == "")) {
        stop("Sample names cannot be empty or NA. Please provide valid sample identifiers.")
    }

    if (any(duplicated(samples))) {
        duplicates <- samples[duplicated(samples)]
        stop(glue::glue(
            "Found duplicate sample names: {paste(unique(duplicates), collapse = ', ')}\n",
            "Each sample must have a unique identifier."
        ))
    }

    assert_readable(paths)
    assert_has_index(paths)

    x <- tibble::tibble(
        sample = samples,
        path = paths
    )

    new("ModBamFiles", x)
}

#' @docType methods
#' @rdname ModBamFiles
#'
#' @param object a ModBamFiles object.
#' @export
setMethod("show", signature("ModBamFiles"), function(object) {
    print(glue::glue(
        "A ModBamFiles object containing {sample_n} samples:",
        sample_n = nrow(object)
    ))
    print(object)
})

#' Modbam methylation results
#'
#' A ModBamResult object stores modbam data used for NanoMethViz
#' visualisation. It contains stores a ModBamFiles object, sample information
#' and optional exon information. The object is constructed using the
#' ModBamResult() constructor function described in "Usage".
#'
#' @slot methy a ModBamFiles data.frame specifying the samples and paths to bam
#'   files.
#' @slot samples the data.frame of sample annotation containing at least columns
#'   sample and group.
#' @slot exons the data.frame of exon information containing at least columns
#'   gene_id, chr, strand, start, end, transcript_id and symbol.
#' @slot mod_code the modification code of interest.
#'
#' @return a NanoMethResult object to be used with plotting functions
#'
#' @export
setClass(
    "ModBamResult",
    representation(
        methy = "ModBamFiles",
        samples = "data.frame",
        exons = "data.frame",
        mod_code = "character"
    )
)

#' @describeIn ModBamResult-class modbam information getter.
#'
#' @param object the ModBamResult object.
#'
#' @return a ModBamFiles data.frame.
#'
#' @export
setMethod(
    "methy",
    signature(object = "ModBamResult"),
    definition = function(object) {
        object@methy
    }
)

#' @describeIn ModBamResult-class modbam information setter.
#'
#' @param object the ModBamResult object.
#' @param value the path to the methylation data.
#'
#' @export
setMethod(
    "methy<-",
    signature(object = "ModBamResult", value = "ModBamFiles"),
    definition = function(object, value) {
        object@methy <- value
        object
    }
)

#' @describeIn ModBamResult-class sample annotation getter.
#'
#' @param object the ModBamResult object.
#'
#' @return the sample annotation.
#'
#' @export
setMethod(
    "samples",
    signature(object = "ModBamResult"),
    definition = function(object) {
        object@samples
    }
)

#' @describeIn ModBamResult-class sample annotation setter.
#'
#' @param object the ModBamResult object.
#' @param value the data.frame of sample annotation containing at least columns
#'   sample and group.
#'
#' @export
setMethod(
    "samples<-",
    signature(object = "ModBamResult", value = "data.frame"),
    definition = function(object, value) {
        assert_valid_samples(value, context = "ModBamResult samples setter")

        # Convert group to factor
        value$group <- as.factor(value$group)

        object@samples <- tibble::as_tibble(value)
        object
    }
)

#' @describeIn ModBamResult-class exon annotation getter.
#'
#' @param object the ModBamResult object.
#'
#' @return the exon annotation.
#'
#' @export
setMethod(
    "exons",
    signature(object = "ModBamResult"),
    definition = function(object) {
        object@exons
    }
)

#' @describeIn ModBamResult-class exon annotation setter.
#'
#' @param object the ModBamResult object.
#' @param value the exon annotation.
#'
#' @export
setMethod(
    "exons<-",
    signature(object = "ModBamResult", value = "data.frame"),
    definition = function(object, value) {
        assert_valid_exons(value)

        object@exons <- tibble::as_tibble(value)
        object
    }
)

#' @describeIn ModBamResult-class mod code getter.
#'
#' @param object the ModBamResult object.
#'
#' @return the mod code.
#'
#' @export
setMethod(
    "mod_code",
    signature(object = "ModBamResult"),
    definition = function(object) {
        object@mod_code
    }
)

#' @describeIn ModBamResult-class mod code setter.
#'
#' @param object the ModBamResult object.
#' @param value the mod code.
#'
#' @export
setMethod(
    "mod_code<-",
    signature(object = "ModBamResult", value = "character"),
    definition = function(object, value) {
        # validate mod_code
        if (is.na(value) || value == "") {
            stop("Modification code cannot be empty or NA. Common codes are 'm' for 5mC, 'h' for 5hmC.")
        }

        value <- as.character(value)
        if (nchar(value) != 1) {
            stop(glue::glue(
                "Modification code must be a single character. Got: '{value}'\n",
                "Common codes are 'm' for 5mC, 'h' for 5hmC, 'a' for 6mA."
            ))
        }

        object@mod_code <- value
        object
    }
)

#' @describeIn ModBamResult-class Constructor
#'
#' @param methy a ModBamFiles object.
#' @param samples the data.frame of sample annotation containing at least
#'   columns sample and group.
#' @param exons (optional) the data.frame of exon information containing at
#'   least columns gene_id, chr, strand, start, end, transcript_id and symbol.
#' @param mod_code a character with the mod code of interest. Defaults to "m"
#'  for 5mC. See details for other options.
#'
#' @details
#' The possible tags for mod_code can be found at
#' \url{https://samtools.github.io/hts-specs/SAMtags.pdf} under the
#' 'Base modifications' section.
#'
#' @export
ModBamResult <- function(methy, samples, exons = NULL, mod_code = "m") {
    # validate inputs
    if (is.null(exons)) {
        exons <- tibble::tibble(
            gene_id = character(),
            chr = character(),
            strand = character(),
            start = integer(),
            end = integer(),
            transcript_id = character(),
            symbol = character()
        )
    }

    # validate ModBamFiles object
    if (!is(methy, "ModBamFiles")) {
        stop(glue::glue(
            "The 'methy' argument must be a ModBamFiles object.\n",
            "Got object of class: {class(methy)[1]}\n",
            "Create a ModBamFiles object using: ModBamFiles(samples = ..., paths = ...)"
        ))
    }

    # validate samples and exons
    assert_valid_samples(samples, context = "ModBamResult constructor")
    assert_valid_exons(exons)

    # validate mod_code
    if (is.na(mod_code) || mod_code == "") {
        stop("Modification code cannot be empty or NA. Common codes are 'm' for 5mC, 'h' for 5hmC.")
    }

    mod_code <- as.character(mod_code)
    if (nchar(mod_code) != 1) {
        stop(glue::glue(
            "Modification code must be a single character. Got: '{mod_code}'\n",
            "Common codes are 'm' for 5mC, 'h' for 5hmC, 'a' for 6mA."
        ))
    }

    # Check that samples in annotation match ModBamFiles samples
    modbam_samples <- methy$sample
    anno_samples <- samples$sample

    matched_samples <- intersect(modbam_samples, anno_samples)
    if (length(matched_samples) == 0) {
        stop(glue::glue(
            "No sample names match between ModBamFiles and sample annotation.\n",
            "ModBamFiles samples: {paste(modbam_samples, collapse = ', ')}\n",
            "Annotation samples: {paste(anno_samples, collapse = ', ')}\n",
            "Please ensure sample names match between your BAM files and annotation."
        ))
    }

    # warn about unmatched samples
    unmatched_bam <- setdiff(modbam_samples, anno_samples)
    unmatched_anno <- setdiff(anno_samples, modbam_samples)

    if (length(unmatched_bam) > 0) {
        warning(glue::glue(
            "Found {length(unmatched_bam)} samples in ModBamFiles not in annotation: {paste(unmatched_bam, collapse = ', ')}\n",
            "These samples will be ignored in analysis."
        ))
    }

    if (length(unmatched_anno) > 0) {
        warning(glue::glue(
            "Found {length(unmatched_anno)} samples in annotation not in ModBamFiles: {paste(unmatched_anno, collapse = ', ')}\n",
            "These samples will have no data for analysis."
        ))
    }

    # convert group to factor
    samples$group <- as.factor(samples$group)

    message(glue::glue(
        "Successfully created ModBamResult with {length(matched_samples)} matched samples."
    ))

    methods::new(
        "ModBamResult",
        methy = methy,
        samples = tibble::as_tibble(samples),
        exons = tibble::as_tibble(exons),
        mod_code = mod_code
    )
}

test_that("Modbam to tabix conversion works", {
    out_file <- paste0(tempfile(), ".tsv.bgz")
    mbr <- ModBamResult(
        methy = ModBamFiles(
            samples = "sample1",
            paths = system.file("peg3.bam", package = "NanoMethViz", mustWork = FALSE)
        ),
        samples = data.frame(
            sample = "sample1",
            group = "group1"
        )
    )

    expect_no_error(modbam_to_tabix(mbr, out_file))
    expect_true(file_exists(out_file))

    tabix_data <- expect_no_error(read_tsv(out_file, col_names = methy_col_names()))
    expect_equal(nrow(tabix_data), 10371)
    expect_equal(ncol(tabix_data), 6)
    expect_equal(unique(tabix_data$sample), "sample1")

    fs::file_delete(out_file)
})


test_that("Modbam to tabix error checking works", {
    out_file <- paste0(tempfile())
    mbr <- ModBamResult(
        methy = ModBamFiles(
            samples = "sample1",
            paths = system.file("peg3.bam", package = "NanoMethViz", mustWork = FALSE)
        ),
        samples = data.frame(
            sample = "sample1",
            group = "group1"
        )
    )

    expect_error(modbam_to_tabix(mbr, out_file), "output_file must end with .bgz extension.")
    expect_false(file_exists(out_file))

    out_folder <- paste0(tempfile(), ".tsv.bgz")
    fs::dir_create(out_folder)

    expect_error(modbam_to_tabix(mbr, out_folder), "output_file exists and is not a file")
})

test_that("ModBamResults getters and setters work", {
    # setup
    mbr <- load_example_modbamresult()

    # test
    expect_s4_class(NanoMethViz::methy(mbr), "ModBamFiles")
    expect_s3_class(NanoMethViz::exons(mbr), "data.frame")
    expect_s3_class(NanoMethViz::samples(mbr), "data.frame")

    expect_no_warning(
        ModBamResult(
            NanoMethViz::methy(mbr),
            NanoMethViz::samples(mbr)
        )
    )

    expect_no_warning(methy(mbr) <- methy(mbr))
    expect_no_warning(samples(mbr) <- samples(mbr))
    expect_no_warning(exons(mbr) <- exons(mbr))
    expect_error(methy(mbr) <- "invalid_path")
    expect_error(exons(mbr) <- exons(mbr)[, -"strand"])

    expect_error(
        ModBamFiles(
            paths = system.file(package = "NanoMethViz", "missing.bam", mustWork = FALSE),
            samples = "sample1"
        ),
        regex = "File path .+ does not exist"
    )

    expect_error(
        ModBamFiles(
            paths = system.file(package = "NanoMethViz", "no_index.bam", mustWork = FALSE),
            samples = "sample1"
        ),
        regex = ".+ is missing its index file"
    )
})

#' Plot GRanges
#'
#' @param x the NanoMethResult object.
#' @param grange the GRanges object with one entry.
#' @inheritParams plot_region
#' @inherit plot_region return
#'
#' @examples
#' nmr <- load_example_nanomethresult()
#' plot_grange(nmr, GenomicRanges::GRanges("chr7:6703892-6730431"))
#'
#' @export
plot_grange <- function(
    x,
    grange,
    anno_regions = NULL,
    binary_threshold = NULL,
    avg_method = c("mean", "median"),
    spaghetti = FALSE,
    heatmap = TRUE,
    heatmap_subsample = 50,
    gene_anno = TRUE,
    smoothing_window = 2000,
    window_prop = 0,
    palette = ggplot2::scale_colour_brewer(palette = "Set1"),
    line_size = 1,
        span = NULL
) {
    if (!missing("span")) {
        warning("the 'span' argument has been deprecated, please use 'smoothing_window' instead")
    }
    avg_method <- match.arg(avg_method)

    assert_that(
        is(grange, "GRanges"),
        length(grange) == 1
    )

    chr <- as.character(GenomicRanges::seqnames(grange))
    start <- GenomicRanges::start(grange)
    end <- GenomicRanges::end(grange)

    plot_region(
        x,
        chr = chr,
        start = start,
        end = end,
        anno_regions = anno_regions,
        binary_threshold = binary_threshold,
        avg_method = avg_method,
        spaghetti = spaghetti,
        heatmap = heatmap,
        heatmap_subsample = heatmap_subsample,
        gene_anno = gene_anno,
        smoothing_window = smoothing_window,
        window_prop = window_prop,
        palette = palette,
        line_size = line_size
    )
}

#' Plot GRanges heatmap
#'
#' @param x the NanoMethResult object.
#' @param grange the GRanges object with one entry.
#' @inheritParams plot_region_heatmap
#'
#' @return a ggplot plot containing the heatmap.
#'
#' @examples
#' nmr <- load_example_nanomethresult()
#' plot_grange_heatmap(nmr, GenomicRanges::GRanges("chr7:6703892-6730431"))
#'
#' @export
plot_grange_heatmap <- function(
    x,
    grange,
    pos_style = c("to_scale", "compact"),
    window_prop = 0,
    subsample = 50
) {
    assert_that(
        is(grange, "GRanges"),
        length(grange) == 1
    )

    chr <- as.character(GenomicRanges::seqnames(grange))
    start <- GenomicRanges::start(grange)
    end <- GenomicRanges::end(grange)

    plot_region_heatmap(
        x,
        chr = chr,
        start = start,
        end = end,
        pos_style = pos_style,
        window_prop = window_prop,
        subsample = subsample
    )
}

test_that("Plotting region methylation heatmap works", {
    # setup
    nmr <- load_example_nanomethresult()

    # test
    expect_no_warning(p <- plot_region_heatmap(nmr, "chr7", 6703892, 6730431))
    expect_s3_class(p, "ggplot")

    expect_no_warning(p <- plot_region_heatmap(nmr, "chr7", 6703892, 6730431, pos_style = "compact"))
    expect_s3_class(p, "ggplot")
})

test_that("plot_region_heatmap works with different parameters", {
    nmr <- load_example_nanomethresult()

    # test with different window_prop values
    expect_no_warning(p <- plot_region_heatmap(nmr, "chr7", 6703892, 6730431, window_prop = 0.1))
    expect_s3_class(p, "ggplot")

    # test with vector window_prop
    expect_no_warning(p <- plot_region_heatmap(nmr, "chr7", 6703892, 6730431, window_prop = c(0.05, 0.15)))
    expect_s3_class(p, "ggplot")

    # test with different subsample value
    expect_no_warning(p <- plot_region_heatmap(nmr, "chr7", 6703892, 6730431, subsample = 30))
    expect_s3_class(p, "ggplot")

    # test with factor chromosome
    chr_factor <- as.factor("chr7")
    expect_no_warning(p <- plot_region_heatmap(nmr, chr_factor, 6703892, 6730431, pos_style = "to_scale"))
    expect_s3_class(p, "ggplot")
})

test_that("Plotting works with GRanges", {
    nmr <- load_example_nanomethresult()
    gr <- GenomicRanges::GRanges(
        seqnames = "chr7",
        ranges = IRanges::IRanges(start = 6703892, end = 6730431),
        strand = "*"
    )

    expect_no_warning(p <- plot_grange_heatmap(nmr, gr))
    expect_s3_class(p, "ggplot")

    expect_no_warning(p <- plot_grange_heatmap(nmr, gr, pos_style = "compact"))
    expect_s3_class(p, "ggplot")
})

test_that("Plotting region methylation works without exons", {
    # setup
    nmr <- load_example_nanomethresult()
    nmr@exons <- tibble::tibble(
        gene_id = character(0),
        chr = character(0),
        strand = character(0),
        start = integer(0),
        end = integer(0),
        transcript_id = character(0),
        symbol = character(0))

    expect_no_warning(p <- plot_region_heatmap(nmr, "chr7", 6703892, 6730431))
})

test_that("plot_region_heatmap window calculation works correctly", {
    nmr <- load_example_nanomethresult()

    # test that window calculations work correctly by testing edge regions
    # small region to test window calculation
    expect_no_warning(p <- plot_region_heatmap(nmr, "chr7", 6720000, 6720100, window_prop = 0.5))
    expect_s3_class(p, "ggplot")

    # test zero window
    expect_no_warning(p <- plot_region_heatmap(nmr, "chr7", 6720000, 6720100, window_prop = 0))
    expect_s3_class(p, "ggplot")
})

#' @rdname plot_gene_heatmap
#'
#' @param window_prop the size of flanking region to plot. Can be a vector of two
#'   values for left and right window size. Values indicate proportion of gene
#'   length.
#' @param pos_style the style for plotting the base positions along the x-axis.
#'   Defaults to "to_scale", plotting (potentially) overlapping squares
#'   along the genomic position to scale. The "compact" options plots only the
#'   positions with measured modification.
#' @param subsample the number of read of packed read rows to subsample to.
#'
#' @return a ggplot plot containing the heatmap.
#'
#' @details
#' This function creates a heatmap visualization of methylation data for a specific gene.
#' Each row in the heatmap represents one or more packed reads, where colored segments
#' indicate methylation probability at each genomic position.
#'
#' @examples
#' nmr <- load_example_nanomethresult()
#' plot_gene_heatmap(nmr, "Peg3")
#'
#' @export
setMethod(
    "plot_gene_heatmap",
    signature(x = "NanoMethResult", gene = "character"),
    function(
        x,
        gene,
        window_prop = 0.3,
        pos_style = c("to_scale", "compact"),
        subsample = 50
    ) {
        pos_style <- match.arg(pos_style)

        plot_gene_heatmap_impl(
            x = x,
            gene = gene,
            window_prop = window_prop,
            pos_style = pos_style,
            subsample = subsample
        )
    }
)

#' @rdname plot_gene_heatmap
#'
#' @export
setMethod(
    "plot_gene_heatmap",
    signature(x = "ModBamResult", gene = "character"),
    function(
        x,
        gene,
        window_prop = 0.3,
        pos_style = c("to_scale", "compact"),
        subsample = 50
    ) {
        pos_style <- match.arg(pos_style)

        plot_gene_heatmap_impl(
            x = x,
            gene = gene,
            window_prop = window_prop,
            pos_style = pos_style,
            subsample = subsample
        )
    }
)

plot_gene_heatmap_impl <- function(
    x,
    gene,
    window_prop,
    pos_style,
    subsample
) {
    # validate genes
    if (is.na(gene) || gene == "") {
        stop("Gene symbol cannot be empty or NA. Please provide a valid gene symbol.")
    }

    if (nrow(exons(x)) == 0) {
        stop(glue::glue(
            "No exon annotation found in the data object.\\n",
            "Gene '{gene}' cannot be plotted without exon information.\\n",
            "Please add exon annotation using: exons(your_object) <- your_exons"
        ))
    }

    if (length(window_prop) == 1) {
        # convert to two sided window
        window_prop <- c(window_prop, window_prop)
    }

    # query_gene_methy
    if (!gene %in% exons(x)$symbol) {
        available_genes <- unique(exons(x)$symbol)
        similar_genes <- available_genes[grepl(paste0("^", substr(gene, 1, 3)), available_genes, ignore.case = TRUE)]

        error_msg <- glue::glue(
            "Gene '{gene}' not found in exon annotation.\\n",
            "Please check the gene symbol spelling."
        )

        if (length(similar_genes) > 0 && length(similar_genes) <= 10) {
            error_msg <- paste0(
                error_msg, "\\n",
                glue::glue("Similar genes found: {paste(similar_genes, collapse = ', ')}")
            )
        } else if (length(available_genes) <= 20) {
            error_msg <- paste0(
                error_msg, "\\n",
                glue::glue("Available genes: {paste(available_genes, collapse = ', ')}")
            )
        } else {
            error_msg <- paste0(
                error_msg, "\\n",
                glue::glue("Use unique(exons(your_object)$symbol) to see all {length(available_genes)} available genes.")
            )
        }

        stop(error_msg)
    }

    pos_range <- gene_pos_range(x, gene)

    chr <- exons(x) %>%
        dplyr::filter(.data$symbol == gene) %>%
        dplyr::slice(1) %>%
        dplyr::pull(chr)

    plot_region_heatmap_impl(
        x = x,
        chr = chr,
        start = pos_range[1],
        end = pos_range[2],
        window_prop = window_prop,
        pos_style = pos_style,
        subsample = subsample
    )
}

#' Plot PCA
#'
#' Plot principal component analysis plot using BiocSingular::runPCA(). It is recommended this be done with
#' the log-methylation-ratio matrix generated by bsseq_to_log_methy_ratio().
#'
#' @param x a numeric matrix where rows are genomic regions/features and columns
#'   are samples. Should be a log-methylation-ratio matrix for best results.
#' @param plot_dims the numeric vector of the two dimensions to be plotted
#'   (default c(1, 2) for PC1 vs PC2).
#' @param labels the character vector of labels for data points. By default uses
#'   column names of x, set to NULL to plot unlabeled points.
#' @param groups the character vector or numeric vector of groups that data points
#'   will be coloured by. For character vectors, discrete colours are used. For
#'   numeric vectors, continuous colour scale is used. By default, groups is NULL
#'   and points are not coloured.
#' @param legend_name the name for the colour legend.
#'
#' @return ggplot object of the PCA plot.
#'
#' @details
#' Principal Component Analysis (PCA) is a dimensionality reduction technique that
#' identifies the directions of maximum variance in high-dimensional data. Unlike
#' MDS, PCA axis labels show the principal component numbers rather than variance
#' explained percentages. The first principal component (PC1) captures the most
#' variance in the data, PC2 captures the second most, and so on. Samples that
#' cluster together have similar methylation profiles across the genomic regions
#' included in the analysis.
#'
#' @examples
#' nmr <- load_example_nanomethresult()
#' bss <- methy_to_bsseq(nmr)
#' lmr <- bsseq_to_log_methy_ratio(bss)
#' plot_pca(lmr)
#'
#' @importFrom BiocSingular runPCA
#'
#' @export
plot_pca <- function(x, plot_dims = c(1, 2), labels = colnames(x), groups = NULL, legend_name = "group") {
    if (!is.null(labels)) {
        assertthat::assert_that(ncol(x) == length(labels))
    }

    if (!is.null(groups)) {
        assertthat::assert_that(ncol(x) == length(groups))
    }

    pca_res <- BiocSingular::runPCA(t(x), rank = max(plot_dims))

    plot_data <- data.frame(
        dim1 = pca_res$x[, plot_dims[1]],
        dim2 = pca_res$x[, plot_dims[2]]
    )

    plot_data$labels <- labels
    plot_data$group <- groups

    xlabel <- glue::glue("PCA Dim {plot_dims[1]}")
    ylabel <- glue::glue("PCA Dim {plot_dims[2]}")

    p <- ggplot2::ggplot(
        plot_data,
        ggplot2::aes(x = .data$dim1, y = .data$dim2)
    )

    if (is.null(groups)) {
        if (is.null(labels)) {
            # no labels or groups
            p <- p + ggplot2::geom_point()
        } else {
            # no groups, but labels
            p <- p + ggplot2::geom_label(aes(label = labels))
        }
    } else {
        if (is.numeric(groups)) {
            # continuous colour palette ignores labels
            if (!is.null(labels)) {
                message("Ignoring labels as groups is numeric. Set `labels=NULL` to suppress this message.")
            }
            p <- p +
                ggplot2::geom_point(aes(colour = .data$group)) +
                ggplot2::scale_color_continuous(name = legend_name)
        } else {
            # discrete colour palette
            if (is.null(labels)) {
                # no labels, but groups
                p <- p + ggplot2::geom_point(aes(colour = .data$group)) +
                    ggplot2::guides(color = ggplot2::guide_legend(title = legend_name))
            } else {
                # labels and groups
                # key_glyph causes the legend to display points
                p <- p + ggplot2::geom_label(
                    aes(label = labels, colour = .data$group),
                    key_glyph = ggplot2::draw_key_point
                )
            }
        }
    }

    p +
        ggplot2::theme_bw() +
        ggplot2::xlab(xlabel) +
        ggplot2::ylab(ylabel) +
        ggplot2::scale_x_continuous(expand = c(.1, .1))
}

StatLowess <- ggplot2::ggproto("StatLowess", ggplot2::Stat,
    required_aes = c("x", "y"),

    compute_group = function(
        data,
        scales,
        params,
        span = 2/3
    ) {
        lowess_fit <- lowess(data$x, data$y, f = span)
        as.data.frame(lowess_fit)
    }
)

stat_lowess <- function(
    mapping = NULL,
    data = NULL,
    geom = "line",
    position = "identity",
    na.rm = TRUE,
    show.legend = NA,
    inherit.aes = TRUE,
    span = 2/3,
    ...
) {
    ggplot2::layer(
        stat = StatLowess,
        data = data,
        mapping = mapping,
        geom = geom,
        position = position,
        show.legend = show.legend,
        inherit.aes = inherit.aes,
        params = list(
            span = span,
            na.rm = na.rm,
            ...
        )
    )
}

# karyogram <- function(nmr, chr_info) {
#     assert_has_columns(chr_info, c("chr", "length"))
#
#     n_chr <- nrow(chr_info)
#     plots <- list()
#
#     widest <- max(chr_info$length)
#
#     for (i in 1:n_chr) {
#         m <- query_methy(nmr, chr_info$chr[i], 1, chr_info$length[i])
#
#         bins <- round(seq(1, chr_info$length[i], length.out = 2^10))
#
#         bin_means <- binMeans(m$statistic > 0, x = m$pos, bx = bins)
#
#         df <- tibble(
#             pos = bins[-length(bins)],
#             methy_prop = as.numeric(bin_means),
#             count = attr(bin_means, "count")
#         )
#
#         plots[[i]] <- ggplot(df, aes(x = pos, y = 1, fill = methy_prop)) +
#             geom_tile() +
#             scico::scale_colour_scico(palette = 'imola', direction = -1) +
#             theme_void() +
#             theme(plot.margin = margin(0, 2, 0, 0)) +
#             ggtitle(chr_info$chr[i]) +
#             xlim(1, widest)
#     }
#
#     patchwork::wrap_plots(plots, ncol = 1, guides = "collect")
# }
#

#' Load an example NanoMethResult object
#'
#' Load an example NanoMethResult object for demonstration of plotting
#' functions. Run `load_example_nanomethresults` without the function call to
#' see how the object is constructed.
#'
#' @return a NanoMethResults object
#'
#' @export
#'
#' @examples
#' nmr <- load_example_nanomethresult()
load_example_nanomethresult <- function() {
    methy <- system.file(package = "NanoMethViz", "methy_subset.tsv.bgz", mustWork = FALSE)

    sample <- c(
        "B6Cast_Prom_1_bl6",
        "B6Cast_Prom_1_cast",
        "B6Cast_Prom_2_bl6",
        "B6Cast_Prom_2_cast",
        "B6Cast_Prom_3_bl6",
        "B6Cast_Prom_3_cast"
    )
    group <- c(
        "bl6",
        "cast",
        "bl6",
        "cast",
        "bl6",
        "cast"
    )
    sample_anno <- data.frame(sample, group, stringsAsFactors = FALSE)

    exon_tibble <- get_example_exons_mus_musculus()

    NanoMethResult(methy, sample_anno, exon_tibble)
}

#' Load an example ModBamResult object
#'
#' Load an example ModBamResult object for demonstration of plotting
#' functions. Run `load_example_modbamresult` without the function call to
#' see how the object is constructed.
#'
#' @return a ModBamResult object
#'
#' @export
#'
#' @examples
#' mbr <- load_example_modbamresult()
load_example_modbamresult <- function() {
    ModBamResult(
        methy = ModBamFiles(
            paths = system.file(package = "NanoMethViz", "peg3.bam", mustWork = FALSE),
            samples = "sample1"
        ),
        samples = tibble::tibble(
            sample = "sample1",
            group = "group1"
        ),
        exons = get_example_exons_mus_musculus()
    )
}

# create a list where the nth element contains the nth values of the original
# vectors
#' @importFrom stats setNames
vec_zip <- function(..., .names = NULL) {
    x <- do.call(data.frame, list(..., stringsAsFactors = FALSE))
    stats::setNames(split(x, seq_len(nrow(x))), .names)
}

extract_file_names <- function(x) {
    fs::path_ext_remove(fs::path_file(x))
}

logit <- function(p) {
    log(p / (1-p))
}

assert_has_columns <- function(x, cols) {
    if (!all(cols %in% colnames(x))) {
        stop(glue::glue(
            "columns missing from {input}: {missing_cols}",
            input = deparse(substitute(x)),
            missing_cols = paste(setdiff(cols, colnames(x)), collapse = ", ")
        ))
    }
}

timed_log <- function(..., verbose = TRUE) {
    time_stamp <- paste0("[", format(Sys.time(), "%Y-%m-%d %H:%M:%S"), "] ")
    if (verbose) {
        message(time_stamp, ...)
    }
}

gene_pos_range <- function(nmr, gene) {
    NanoMethViz::exons(nmr) %>%
        dplyr::filter(.data$symbol == !!gene) %>%
        dplyr::select("start", "end") %>%
        unlist() %>%
        range()
}

get_tabix_sequences <- function(file) {
    f <- gzfile(file, "rb")

    # read magic
    readChar(f, 4)

    # get n_ref and skip other fields
    n_ref <- readBin(f, "integer")
    for (i in seq_len(7)) readBin(f, "integer")

    # read sequences
    seqs <- character(n_ref)
    for (i in seq_len(n_ref)) {
        seqs[i] <- readBin(f, "character")
    }

    close(f)
    seqs
}

stack_plots <- function(spaghetti, heat) {
    if (is(spaghetti, "patchwork")) {
        spaghetti[[1]] + heat + spaghetti[[2]] +
            patchwork::plot_layout(
                nrow = 3,
                heights = c(spaghetti$patches$layout$heights[1], 1, spaghetti$patches$layout$heights[2])
            )
    } else {
        spaghetti + heat + patchwork::plot_layout(nrow = 2)
    }
}

map_rows <- function(x, .f, ...) {
    if (!is.data.frame(x) || nrow(x) == 0) {
        stop("'x' must be a non-empty data.frame")
    }

    # Apply the function .f to each row of x using lapply() and store the results in a list.
    out <- lapply(seq_len(nrow(x)), function(i) .f(x[i, ], ...))

    out
}

neg_strand_offset <- function(x, offset) {
    assert_has_columns(x, c("strand", "pos"))

    x %>%
        dplyr::mutate(pos = case_when(
            strand == "-" ~ pos + offset,
            TRUE ~ pos
        ))
}

missingness <- function(x) {
    mean(is.na(x))
}

mat_row_map <- function(x, f) {
    apply(x, 1, f)
}

mat_col_map <- function(x, f) {
    apply(x, 2, f)
}

df_to_matrix <- function(x, rownames = 1) {
    if (!is.null(rownames)) {
        rnames <- x[, rownames, drop = TRUE]
        x <- x[, -rownames]
    }

    x <- as.matrix(x)
    rownames(x) <- rnames
    x
}

is_coordinate <- function(x) {
    is.numeric(x) && x == round(x) && x >= 0
}

assertthat::on_failure(is_coordinate) <- function(call, env) {
  glue::glue("'{deparse(call$x)}' ({eval(call$x)}) is not a valid coordinate")
}

same_length <- function(...) {
    nulls <- purrr::map_lgl(list(...), is.null)
    lengths <- purrr::map_dbl(list(...), length)
    !any(nulls) && all(purrr::map_lgl(lengths, ~ . == lengths[1]))
}

#' @importFrom IRanges IRanges
make_granges <- function(chr, start, end) {
    assertthat::assert_that(same_length(chr, start, end))
    assertthat::assert_that(all(start < end))
    GenomicRanges::GRanges(
        seqnames = chr,
        ranges = IRanges::IRanges(start, end)
    )
}

get_bam_total_reads <- function(path) {
    sum(Rsamtools::idxstatsBam(path)[, c("mapped", "unmapped")])
}

test_that("Plotting feature clustering", {
    nmr <- load_example_nanomethresult()
    gene_anno <- exons_to_genes(NanoMethViz::exons(nmr))

    expect_no_warning(cluster_regions(nmr, gene_anno, centers = 2, grid_method = "uniform"))
    expect_no_warning(cluster_regions(nmr, gene_anno, centers = 3, grid_method = "uniform"))
    expect_no_warning(cluster_regions(nmr, gene_anno, centers = 2, grid_method = "density"))
})

test_that("make_granges returns correct GRanges object", {
    chr <- c("chr1", "chr2", "chr3")
    start <- c(100, 200, 300)
    end <- c(200, 300, 400)

    # Call the make_granges function
    granges <- make_granges(chr, start, end)

    # Test assertions
    expect_s4_class(granges, "GRanges")
    expect_equal(length(granges), length(chr))
    expect_equal(names(granges), NULL)  # Expect names to be NULL

    # Check each range individually
    for (i in seq_along(granges)) {
        expect_equal(as.character(GenomicRanges::seqnames(granges))[i], chr[i])
        expect_equal(GenomicRanges::start(granges)[i], start[i])
        expect_equal(GenomicRanges::end(granges)[i], end[i])
    }
})

test_that("make_granges throws error for unequal input lengths", {
    chr <- c("chr1", "chr2", "chr3")
    start <- c(100, 200)
    end <- c(200, 300, 400)

    # Expect an error to be thrown for unequal input lengths
    expect_error(make_granges(chr, start, end))
})

test_that("Plotting gene works", {
    # setup
    nmr <- load_example_nanomethresult()
    mbr <- load_example_modbamresult()

    mbr_lower <- load_example_modbamresult()
    methy(mbr_lower) <- ModBamFiles(
        paths = system.file(package = "NanoMethViz", "peg3_lower_case.bam", mustWork = FALSE),
        samples = "sample1"
    )

    dmr_region <- data.frame(
        chr = "chr7",
        start = 6727500,
        end = 6732000
    )

    data_list <- list(nmr, mbr, mbr_lower)

    params <- expand.grid(
        heatmap = c(TRUE, FALSE),
        spaghetti = c(TRUE, FALSE),
        gene_anno = c(TRUE, FALSE)
    )

    # test plot_gene() ----
    for (x in data_list) {
        p <- expect_no_warning(plot_gene(x, "Peg3"))
        expect_s3_class(p, "ggplot")

        p <- expect_no_warning(plot_gene(x, "Peg3", anno_regions = dmr_region))
        expect_s3_class(p, "ggplot")

        p <- expect_no_warning(plot_gene(x, "Peg3", heatmap = TRUE))
        expect_s3_class(p, "ggplot")

        p <- expect_no_warning(plot_gene(x, "Peg3", heatmap = TRUE, heatmap_subsample = 5))
        expect_s3_class(p, "ggplot")

        for (bt in c(0, 0.5, 1)) {
            p <- expect_no_warning(
                plot_gene(
                    x, "Peg3",
                    binary_threshold = bt
                )
            )
            expect_s3_class(p, "ggplot")
        }
        for (i in 1:nrow(params)) {
            p <- expect_no_warning(
                plot_gene(
                    x, "Peg3",
                    heatmap = params$heatmap[i],
                    spaghetti = params$spaghetti[i],
                    gene_anno = params$gene_anno[i]
                )
            )

            expect_s3_class(p, "ggplot")
        }

        expect_error(plot_gene(x, "missing_gene", heatmap = TRUE))
    }

    # test plot_region() ----
    for (x in data_list) {
        p <- expect_no_warning(plot_region(x, "chr7", 6703892, 6730431))
        expect_s3_class(p, "ggplot")

        p <- expect_no_warning(plot_region(x, factor("chr7"), 6703892, 6730431))
        expect_s3_class(p, "ggplot")

        p <- expect_no_warning(plot_region(nmr, "chr7", 6703892, 6730431, heatmap = TRUE, heatmap_subsample = 5))
        expect_s3_class(p, "ggplot")

        for (i in 1:nrow(params)) {
            p <- expect_no_warning(
                plot_region(
                    x, "chr7", 6703892, 6730431,
                    heatmap = params$heatmap[i],
                    spaghetti = params$spaghetti[i],
                    gene_anno = params$gene_anno[i]
                )
            )

            expect_s3_class(p, "ggplot")
        }
        expect_error(plot_region(x, chr = "chr7", start = 0, end = 0), "Start positions must be")
        expect_warning(plot_region(x, chr = "chr7", start = 1, end = 10), "No methylation data found in region")
    }

    # test plot_region() with unknown chromosome ----
    expect_warning(expect_error(plot_region(nmr, "unknown_chr", 6703892, 6730431)))
    expect_error(plot_region(mbr, "unknown_chr", 6703892, 6730431))

    # test plot_grange() ----
    grange <- GenomicRanges::GRanges("chr7:6703892-6730431")
    for (x in data_list) {
        for (i in 1:nrow(params)) {
            p <- expect_no_warning(
                plot_grange(
                    x, grange,
                    heatmap = params$heatmap[i],
                    spaghetti = params$spaghetti[i]
                )
            )

            expect_s3_class(p, "ggplot")
        }
    }
})

test_that("Plotting gene methylation heatmap works", {
    # setup
    nmr <- load_example_nanomethresult()

    # test
    p <- expect_no_warning(plot_gene_heatmap(nmr, "Peg3"))
    expect_s3_class(p, "ggplot")

    # test for a bug whereby samples not present in data cause function to hang
    nmr_extra_sample <- load_example_nanomethresult()
    samples(nmr_extra_sample) <- bind_rows(
        samples(nmr_extra_sample),
        c(sample = "foo", group = "bar")
    )
    p <- expect_no_warning(plot_gene_heatmap(nmr_extra_sample, "Peg3"))
    expect_s3_class(p, "ggplot")
})

test_that("plot_gene_heatmap works with different parameters", {
    nmr <- load_example_nanomethresult()

    # test with different window_prop values
    p <- expect_no_warning(plot_gene_heatmap(nmr, "Peg3", window_prop = 0.5))
    expect_s3_class(p, "ggplot")

    # test with vector window_prop
    p <- expect_no_warning(plot_gene_heatmap(nmr, "Peg3", window_prop = c(0.2, 0.4)))
    expect_s3_class(p, "ggplot")

    # test with compact pos_style
    p <- expect_no_warning(plot_gene_heatmap(nmr, "Peg3", pos_style = "compact"))
    expect_s3_class(p, "ggplot")

    # test with different subsample value
    p <- expect_no_warning(plot_gene_heatmap(nmr, "Peg3", subsample = 25))
    expect_s3_class(p, "ggplot")
})

test_that("plot_gene_heatmap error handling", {
    nmr <- load_example_nanomethresult()

    # test with empty exons
    nmr_no_exons <- nmr
    nmr_no_exons@exons <- tibble::tibble(
        gene_id = character(0),
        chr = character(0),
        strand = character(0),
        start = integer(0),
        end = integer(0),
        transcript_id = character(0),
        symbol = character(0)
    )

    expect_error(
        plot_gene_heatmap(nmr_no_exons, "Peg3"),
        "No exon annotation found in the data object"
    )

    # test with gene not in annotation
    expect_error(
        plot_gene_heatmap(nmr, "NonExistentGene"),
        "Gene 'NonExistentGene' not found in exon annotation"
    )
})

test_that("NanoMethResults getters and setters work", {
    # setup
    nmr <- load_example_nanomethresult()

    # test
    expect_true(fs::file_exists(NanoMethViz::methy(nmr)))
    expect_type(NanoMethViz::methy(nmr), "character")
    expect_s3_class(NanoMethViz::exons(nmr), "data.frame")
    expect_s3_class(NanoMethViz::samples(nmr), "data.frame")

    expect_no_warning(
        NanoMethResult(
            NanoMethViz::methy(nmr),
            NanoMethViz::samples(nmr)
        )
    )

    expect_no_warning(methy(nmr) <- methy(nmr))
    expect_no_warning(samples(nmr) <- samples(nmr))
    expect_no_warning(exons(nmr) <- exons(nmr))

    expect_error(methy(nmr) <- "invalid_path")
    expect_error(exons(nmr) <- exons(nmr)[, -"strand"])
})

test_that("region_methy_stats calculates methylation statistics for each region", {
    # Create sample data
    nmr <- load_example_nanomethresult()
    regions <- exons_to_genes(exons(nmr))

    # Call the function
    result <- region_methy_stats(nmr, regions, threshold = 0.6)

    # Assert the result has the expected structure and columns
    expect_s3_class(result, "data.frame")
    expect_identical(colnames(result), c("gene_id", "chr", "strand", "symbol", "start", "end", "mean_methy_prob", "prevalence"))

    # Assert the number of rows in the result matches the number of input regions
    expect_equal(nrow(result), nrow(regions))

    # Assert that the calculated statistics are within the expected range
    expect_true(all(result$mean_methy_prob >= 0 & result$mean_methy_prob <= 1))
    expect_true(all(result$prevalence >= 0 & result$prevalence <= 1))
})

test_that("Aggregate plotting works", {
    # setup
    nmr <- load_example_nanomethresult()
    mbr <- NanoMethViz:::load_example_modbamresult()
    gene_anno <- exons_to_genes(NanoMethViz::exons(nmr))

    # test
    for (x in list(nmr, mbr)) {
        expect_no_warning(plot_agg_regions(x, gene_anno))
        expect_no_warning(plot_agg_regions(x, gene_anno, "sample"))
        expect_no_warning(plot_agg_regions(x, gene_anno, "group"))
    }
})

test_that("Aggregate plotting error checking works", {
    # setup
    nmr <- load_example_nanomethresult()
    missing_chr <- NanoMethViz::exons(nmr)[, -2]
    missing_chr_list <- list(
        NanoMethViz::exons(nmr)[, -2],
        NanoMethViz::exons(nmr),
        NanoMethViz::exons(nmr)[, -2]
    )

    # test
    expect_error(plot_agg_regions(nmr, gene_anno, "foo"))
})

test_that("Plotting gene works", {
    # setup
    nmr <- load_example_nanomethresult()

    # test
    p_gene <- expect_no_warning(plot_gene(nmr, "Peg3"))
    p_gene2 <- expect_no_warning(plot_gene(nmr, "Peg3", spaghetti = TRUE))
    expect_s3_class(p_gene, "patchwork")
    expect_s3_class(p_gene, "ggplot")

    expect_s3_class(p_gene2, "patchwork")
    expect_s3_class(p_gene2, "ggplot")
})

expect_nrow <- function(object, n) {
    # 1. Capture object and label
    act <- testthat::quasi_label(rlang::enquo(object), arg = "object")

    # 2. Call expect()
    act$n <- nrow(act$val)
    testthat::expect(
        act$n == n,
        sprintf("%s has nrow %i, not nrow %i.", act$lab, act$n, n)
    )

    # 3. Invisibly return the value
    invisible(act$val)
}

expect_ncol <- function(object, n) {
    # 1. Capture object and label
    act <- testthat::quasi_label(rlang::enquo(object), arg = "object")

    # 2. Call expect()
    act$n <- ncol(act$val)
    testthat::expect(
        act$n == n,
        sprintf("%s has ncol %i, not ncol %i.", act$lab, act$n, n)
    )

    # 3. Invisibly return the value
    invisible(act$val)
}

test_that("methy_to_bsseq works", {
    # setup
    nmr <- load_example_nanomethresult()
    bss <- methy_to_bsseq(nmr)

    # test
    expect_no_error(methy_to_bsseq(methy(nmr)))
    expect_s4_class(methy_to_bsseq(nmr), "BSseq")
    expect_equal(ncol(bss), 6)
    expect_equal(
        nrow(SummarizedExperiment::colData(bss)),
        nrow(NanoMethViz::samples(nmr))
    )
    expect_equal(
        ncol(SummarizedExperiment::colData(bss)),
        ncol(NanoMethViz::samples(nmr))
    )
    expect_equal(
        colnames(SummarizedExperiment::colData(bss)),
        colnames(NanoMethViz::samples(nmr))
    )
})

test_that("bsseq_to_* works", {
    nmr <- load_example_nanomethresult()
    bss <- methy_to_bsseq(NanoMethViz::methy(nmr))

    # test
    edger_counts <- expect_no_warning(bsseq_to_edger(bss))
    expect_ncol(edger_counts, 12)
    expect_nrow(edger_counts, 4778)
    lmr <- expect_no_warning(bsseq_to_log_methy_ratio(bss))
    expect_ncol(lmr, 6)
    expect_nrow(lmr, 4778)

    expect_equal(ncol(edger_counts), 2 * ncol(bss))
    expect_equal(ncol(lmr), ncol(bss))
})

#' Nanopore Methylation Result
#'
#' A NanoMethResult object stores data used for NanoMethViz visualisation. It
#' contains stores a path to the methylation data, sample information and
#' optional exon information. The object is constructed using the
#' NanoMethResult() constructor function described in "Usage".
#'
#' @slot methy the path to the methylation tabix file.
#' @slot samples the data.frame of sample annotation containing at least columns
#'   sample and group.
#' @slot exons the data.frame of exon information containing at least columns
#'   gene_id, chr, strand, start, end, transcript_id and symbol.
#'
#' @return a NanoMethResult object to be used with plotting functions
#'
#' @examples
#' methy <- system.file(package = "NanoMethViz", "methy_subset.tsv.bgz", mustWork = FALSE)
#' sample <- c(
#'     "B6Cast_Prom_1_bl6",
#'     "B6Cast_Prom_1_cast",
#'     "B6Cast_Prom_2_bl6",
#'     "B6Cast_Prom_2_cast",
#'     "B6Cast_Prom_3_bl6",
#'     "B6Cast_Prom_3_cast"
#' )
#' group <- c(
#'     "bl6",
#'     "cast",
#'     "bl6",
#'     "cast",
#'     "bl6",
#'     "cast"
#' )
#' sample_anno <- data.frame(sample, group, stringsAsFactors = FALSE)
#' exon_tibble <- get_example_exons_mus_musculus()
#' NanoMethResult(methy, sample_anno, exon_tibble)
#'
#' @export
setClass(
    "NanoMethResult",
    representation(
        methy = "character",
        samples = "data.frame",
        exons = "data.frame"
    )
)

#' @describeIn NanoMethResult-class Constructor
#'
#' @param methy the path to the methylation tabix file.
#' @param samples the data.frame of sample annotation containing at least
#'   columns sample and group.
#' @param exons (optional) the data.frame of exon information containing at
#'   least columns gene_id, chr, strand, start, end, transcript_id and symbol.
#'
#' @export
NanoMethResult <- function(methy, samples, exons = NULL) {
    # validate input
    if (is.null(exons)) {
        exons <- tibble::tibble(
            gene_id = character(),
            chr = character(),
            strand = character(),
            start = integer(),
            end = integer(),
            transcript_id = character(),
            symbol = character()
        )
    }

    # validate inputs
    assert_readable(methy)
    assert_valid_samples(samples, context = "NanoMethResult constructor")
    assert_valid_exons(exons)

    # convert group to factor
    samples$group <- as.factor(samples$group)

    # ensure samples match
    assert_valid_methy_samples(methy, samples)

    methods::new(
        "NanoMethResult",
        methy = methy,
        samples = tibble::as_tibble(samples),
        exons = tibble::as_tibble(exons)
    )
}

# methy ----

#' @describeIn NanoMethResult-class methylation data path getter.
#'
#' @param object the NanoMethResult object.
#'
#' @return the path to the methylation data.
#'
#' @examples
#' x <- load_example_nanomethresult()
#' methy(x)
#'
#' @export
setMethod("methy", signature("NanoMethResult"), function(object) {
    object@methy
})

#' @describeIn NanoMethResult-class methylation data path setter.
#'
#' @param object the NanoMethResult object.
#' @param value the path to the methylation data.
#'
#' @export
setMethod("methy<-", signature("NanoMethResult"), function(object, value) {
    assert_readable(value)

    object@methy <- value
    object
})

# sample ----
#' @describeIn NanoMethResult-class sample annotation getter.
#'
#' @param object the NanoMethResult object.
#'
#' @return the sample annotation.
#'
#' @export
setMethod("samples", signature("NanoMethResult"), function(object) {
    object@samples
})

#' @describeIn NanoMethResult-class sample annotation setter.
#'
#' @param object the NanoMethResult object.
#' @param value the data.frame of sample annotation containing at least columns
#'   sample and group.
#'
#' @export
setMethod("samples<-", signature("NanoMethResult", "data.frame"), function(object, value) {
    assert_valid_samples(value, context = "NanoMethResult samples setter")

    # Convert group to factor
    value$group <- as.factor(value$group)

    object@samples <- tibble::as_tibble(value)
    object
})

# exons ----
#' @describeIn NanoMethResult-class exon annotation getter.
#'
#' @param object the NanoMethResult object.
#'
#' @return the exon annotation.
#'
#' @export
setMethod("exons", signature("NanoMethResult"), function(object) {
    object@exons
})

#' @describeIn NanoMethResult-class exon annotation setter.
#'
#' @param object the NanoMethResult object.
#' @param value the exon annotation.
#'
#' @export
setMethod("exons<-", signature("NanoMethResult", "data.frame"), function(object, value) {
    assert_valid_exons(value)

    object@exons <- tibble::as_tibble(value)
    object
})

#' Sort methylation file
#' @keywords internal
#'
#' @param x the path to the methylation file to sort
#'
#' @return invisibly returns path of sorted file
sort_methy_file <- function(x) {
    assert_readable(x)

    if (.Platform$OS.type == "windows") {
        methy_df <- readr::read_tsv(
            x,
            col_names = methy_col_names(),
            col_types = methy_col_types()
        )
        methy_df <- dplyr::arrange(methy_df, .data$chr, .data$pos)
        readr::write_tsv(methy_df, x, col_names = FALSE, progress = FALSE)
    } else {
        cmd <- glue::glue("sort --compress-program=gzip -S 4G -k2,3V {x} -o {x}")
        system(cmd)
    }

    invisible(x)
}

tabix_compress <- function(x, index = TRUE) {
    assert_readable(x)

    f <- Rsamtools::bgzip(x, overwrite = TRUE)
    if (index) {
        tabix_index(f)
    }

    f
}

tabix_index <- function(x) {
    assert_readable(x)

    Rsamtools::indexTabix(x, seq = 2, start = 3, end = 3)
}

#' Convert methylation file to tabix format
#' @keywords internal
#'
#' @param x the path to the sorted methylation file
#'
#' @return invisibly returns the path to the tabix file
raw_methy_to_tabix <- function(x) {
    assert_readable(x)

    bgz_name <- tabix_compress(x)
    tabix_index(bgz_name)

    invisible(bgz_name)
}

#' Plot aggregate regions
#'
#' @param x the NanoMethResult or ModBamResult object.
#' @param regions a table of regions containing at least columns chr, strand,
#'   start and end. Any additional columns can be used for grouping.
#' @param binary_threshold the modification probability such that calls with
#'   modification probability above the threshold are considered methylated, and
#'   those with probability equal or below are considered unmethylated.
#' @param group_col the column name to group aggregated trends by. This column can
#'   be found in either the regions table or samples(x). When NULL (default), all
#'   data is aggregated together. Common values include "sample" to show individual
#'   samples or "group" to show sample groups.
#' @param flank the number of flanking bases to add to each side of each region.
#' @param stranded if TRUE, negative strand features will have their coordinates
#'   flipped to reflect biological features like transcription start sites
#'   (e.g., for genes, coordinates run from TSS to TES regardless of strand).
#' @param span the span parameter for loess smoothing of the trend lines.
#' @param palette the ggplot colour palette used for groups.
#'
#' @return a ggplot object containing the aggregate methylation trend.
#'
#' @examples
#' nmr <- load_example_nanomethresult()
#' gene_anno <- exons_to_genes(NanoMethViz::exons(nmr))
#' plot_agg_regions(nmr, gene_anno)
#' plot_agg_regions(nmr, gene_anno, group_col = "sample")
#' plot_agg_regions(nmr, gene_anno, group_col = "group")
#'
#' @export
plot_agg_regions <- function(
    x,
    regions,
    binary_threshold = 0.5,
    group_col = NULL,
    flank = 2000,
    stranded = TRUE,
    span = 0.05,
    palette = ggplot2::scale_colour_brewer(palette = "Set1")
) {
    if (!is.null(group_col)) {
        avail_columns <- c(colnames(samples(x)), colnames(regions))
        assertthat::assert_that(
            group_col %in% avail_columns,
            msg = glue::glue("'{group_col}' could not be found in columns of 'regions' or samples(x)")
        )
    }

    # grouped regions crashes downstream operations
    regions <- ungroup(regions)

    # query methylation data
    methy_data <- plot_agg_regions.get_methy_data(x, regions, flank)

    # process each methy_data list element to obtain methylation proportions at relative positions
    methy_data <- plot_agg_regions.process_methy_data(methy_data, stranded, flank, binary_threshold)

    # unnest methy_data and add sample annotation
    methy_data <- plot_agg_regions.unnest_with_anno(methy_data, NanoMethViz::samples(x))

    # take the average methylation proportion over equal sized positional bins
    methy_data <- plot_agg_regions.avg_over_bins(methy_data, group_col = group_col)

    # change aes spec depending on if group_col is available
    # hacky fix to the deprecation of aes_string which handled a NULL group_col
    # value
    if (!is.null(group_col)) {
        aes_spec <- ggplot2::aes(
                x = .data$binned_pos,
                y = .data$methy_prop,
                group = .data[[group_col]],
                col = .data[[group_col]])
    } else {
        aes_spec <- ggplot2::aes(
                x = .data$binned_pos,
                y = .data$methy_prop)
    }

    # set up plot
    p <- ggplot2::ggplot() +
        ggplot2::ylim(c(0, 1)) +
        ggplot2::theme_bw() +
        ggplot2::stat_smooth(
            aes_spec,
            method = "loess",
            formula = "y ~ x",
            span = span,
            na.rm = TRUE,
            se = FALSE,
            data = methy_data) +
        palette

    # if flank is 0, then then only annotated start and end, else annotate flanks as well
    if (flank == 0) {
        labels <- c("start", "end")
        breaks <- c(0, 1)
        limits <- c(0, 1)
    } else {
        breaks <- c(-.33, 0, 1, 1.33)
        limits <- c(-0.33, 1.33)
        kb_marker <- round(flank / 1000, 1)
        labels <- c(glue::glue("-{kb_marker}kb"), "start", "end", glue::glue("+{kb_marker}kb"))
        p <- p +
            ggplot2::geom_vline(xintercept = 0, linetype = "dashed", color = "grey80") +
            ggplot2::geom_vline(xintercept = 1, linetype = "dashed", color = "grey80")
    }

    region_widths <- regions$end - regions$start
    width_avg <- round(mean(region_widths))
    width_std_dev <- round(sd(region_widths))

    p + ggplot2::coord_cartesian(clip = "off") +
        ggplot2::scale_x_continuous(
            name = glue::glue("Relative Position (Avg.Width = {width_avg}, Std.Dev = {width_std_dev})"),
            breaks = breaks,
            limits = limits,
            labels = labels) +
        ggplot2::ylab("Average Methylation Proportion")
}

plot_agg_regions.get_methy_data <- function(x, regions, flank) {
    query_row_methy <- function(i, methy, regions, flank) {
        # query a larger region such that smoothing doesn't
        # misbehave around ends
        flank <- flank * 1.1
        query_methy(
            methy,
            regions$chr[i],
            regions$start[i] - flank,
            regions$end[i] + flank,
            force = TRUE)
    }

    methy_data <- purrr::map(
        seq_len(nrow(regions)),
        query_row_methy,
        methy = x,
        regions = regions,
        flank = flank
    )

    regions %>%
        dplyr::mutate(methy_data = methy_data)
}

plot_agg_regions.process_methy_data <- function(methy_data, stranded, flank, binary_threshold) {

    remove_empty_methy_data <- function(x) {
        x %>%
            dplyr::filter(purrr::map_lgl(.data$methy_data, ~nrow(.x) > 0))
    }

    rescale_positions <- function(x, stranded, flank) {
        # for each region, scale the position values to be between 0 and 1
        for (i in seq_len(nrow(x))) {
            # determine if the positions fall inside the region
            m_data <- x$methy_data[[i]]$pos
            within <- m_data >= x$start[i] & m_data <= x$end[i]
            upstream <- m_data < x$start[i]
            downstream <- m_data > x$end[i]

            # rescale the positions depending on the region
            rel_pos <- numeric(length(m_data))
            rel_pos[within] <- (m_data[within] - x$start[i]) / (x$end[i] - x$start[i])

            # flanks each take up 1/3 of the plot space
            rel_pos[upstream] <- (m_data[upstream] - x$start[i]) / flank / 3
            rel_pos[downstream] <- 1 + ((m_data[downstream] - x$end[i]) / flank / 3)

            # flip positions for negative strand if stranded is TRUE
            if (stranded && length(x$strand[i]) > 0 && x$strand[i] == "-") {
                rel_pos <- 1 - rel_pos
            }

            # store rescaled positions
            x$methy_data[[i]]$rel_pos <- rel_pos
        }

        x
    }

    binarise_statistics <- function(x, binary_threshold) {
        binarise_statistc_in_df <- function(x, binary_threshold) {
            x %>%
                dplyr::mutate(is_methylated = e1071::sigmoid(.data$statistic) > binary_threshold)
        }

        x %>%
            dplyr::mutate(
                methy_data = purrr::map(
                    .data$methy_data,
                    binarise_statistc_in_df, binary_threshold = binary_threshold
                )
            )
    }

    average_binarised_methy_per_pos <- function(x) {
        avg_methy_by_rel_pos <- function(x) {
            x %>%
                dplyr::group_by(.data$sample, .data$chr, .data$strand, .data$rel_pos) %>%
                dplyr::summarise(
                    methy_prop = mean(.data$is_methylated),
                    .groups = "drop")
        }

        x %>%
            dplyr::mutate(
                methy_data = purrr::map(.data$methy_data, avg_methy_by_rel_pos)
            )
    }

    methy_data %>%
        remove_empty_methy_data() %>%
        rescale_positions(stranded, flank) %>%
        binarise_statistics(binary_threshold) %>%
        average_binarised_methy_per_pos()
}


plot_agg_regions.unnest_with_anno <- function(x, samples_anno) {
    x %>%
        dplyr::select(!dplyr::any_of(c("chr", "strand", "start", "end"))) %>%
        tidyr::unnest("methy_data") %>%
        dplyr::inner_join(samples_anno, by = "sample", multiple = "all")
}

plot_agg_regions.avg_over_bins <- function(x, group_col, grid_size = 2^10) {
    min <- -1.1 / 3
    max <- 1 + 1.1 / 3
    binned_pos <- seq(min, max, length.out = grid_size + 1)
    binned_pos_df <- data.frame(
        binned_pos = binned_pos[-1],
        interval = cut(binned_pos, breaks = binned_pos)[-1]
    )

    x %>%
        dplyr::mutate(interval = cut(.data$rel_pos, breaks = binned_pos)) %>%
        dplyr::inner_join(binned_pos_df, by = "interval", multiple = "all") %>%
        dplyr::group_by(dplyr::across(dplyr::all_of(group_col)), binned_pos) %>%
        dplyr::summarise(methy_prop = mean(.data$methy_prop), .groups = "drop")
}

#' Convert exon annotation to genes
#'
#' @param x the exon level annotation containing columns "gene_id", "chr",
#'   "strand" and "symbol".
#'
#' @return the gene level annotation where each gene is taken to span the
#'   earliest start position and latest end position of its exons.
#' @export
#'
#' @examples
#' nmr <- load_example_nanomethresult()
#' exons_to_genes(NanoMethViz::exons(nmr))
exons_to_genes <- function(x) {
    if (is(x, "NanoMethResult") | is(x, "ModBamResult")) {
        x <- NanoMethViz::exons(x)
    }

    x %>%
        dplyr::group_by(.data$gene_id, .data$chr, .data$strand, .data$symbol) %>%
        dplyr::summarise(start = min(.data$start), end = max(.data$end), .groups = "drop") %>%
        dplyr::ungroup()
}

#' Get CpG islands annotations
#'
#' Helper functions are provided for obtaining CpG islands annotations from UCSC table browser.
#'
#' @keywords internal
#'
#' @return tibble (data.frame) object containing CpG islands annotation.
#'
get_cgi <- function(genome) {
    available_genomes <- c("hg19", "hg38", "t2t", "mm10", "grcm39")
    assertthat::assert_that(
        genome %in% available_genomes,
        msg = sprintf("genome must be one of %s", paste(available_genomes, collapse = ", "))
    )

    rds_path <- switch(
        genome,
        hg19 = system.file("cgi_hg19.rds", package = "NanoMethViz", mustWork = FALSE),
        hg38 = system.file("cgi_hg38.rds", package = "NanoMethViz", mustWork = FALSE),
        t2t = system.file("cgi_t2t.rds", package = "NanoMethViz", mustWork = FALSE),
        mm10 = system.file("cgi_mm10.rds", package = "NanoMethViz", mustWork = FALSE),
        grcm39 = system.file("cgi_GRCm39.rds", package = "NanoMethViz", mustWork = FALSE)
    )

    readRDS(rds_path)
}

#' @rdname get_exons
#'
#' @examples
#' cgi_mm10 <- get_cgi_mm10()
#'
#' @export
get_cgi_mm10 <- function() {
    get_cgi("mm10")
}

#' @rdname get_exons
#'
#' @examples
#' cgi_GRCm39 <- get_cgi_grcm39()
#'
#' @export
get_cgi_grcm39 <- function() {
    get_cgi("grcm39")
}

#' @rdname get_exons
#'
#' @examples
#' cgi_t2t <- get_cgi_t2t()
#'
#' @export
get_cgi_t2t <- function() {
    get_cgi("t2t")
}

#' @rdname get_exons
#'
#' @examples
#' cgi_hg19 <- get_cgi_hg19()
#'
#' @export
get_cgi_hg19 <- function() {
    get_cgi("hg19")
}

#' @rdname get_exons
#'
#' @examples
#' cgi_hg38 <- get_cgi_hg38()
#'
#' @export
get_cgi_hg38 <- function() {
    get_cgi("hg38")
}

#' Query methylation data
#'
#' @param x the NanoMethResults object or a path to the methylation data
#'   (tabix-bgzipped).
#' @param chr the vector of chromosomes
#' @param start the vector of start positions
#' @param end the vector of end positions
#' @param simplify whether returned results should be row-concatenated
#' @param force whether to force empty output when query region 'chr' does not
#'   appear in data. Without 'force', an empty result indicates that the
#'   requested 'chr' appears in the data but no data overlaps with requested
#'   region, and an invalid 'chr' will cause an error.
#' @param truncate when querying from ModBamFiles, whether or not to truncate
#'   returned results to only those within the specified region. Otherwise
#'   methylation data for entire reads overlapping the region will be returned.
#' @param site_filter the minimum amount of coverage to report a site. This
#'   filters the queried data such that any site with less than the filter is
#'   not returned. The default is 1, which means that all sites are returned.
#'   This option can be set globally using the `options(site_filter = ...)`
#'   which will affect all plotting functions in NanoMethviz.
#'
#' @return A table containing the data within the queried regions. If simplify
#'   is TRUE (default) then returns all data in a single table, otherwise returns
#'   a list of tables where each table is the data from one region.
#'
#' @details
#' The argument `site_filter` can be set globally using the `options(site_filter
#' = ...)` command. The same data entry may appear multiple times in the output
#' if it overlaps multiple regions.
#'
#' @examples
#' nmr <- load_example_nanomethresult()
#' query_methy(methy(nmr), "chr7", 6703892, 6730431)
#'
#' @importFrom Rsamtools TabixFile scanTabix
#'
#' @export
query_methy <- function(
    x,
    chr,
    start,
    end,
    simplify = TRUE,
    force = FALSE,
    truncate = TRUE,
    site_filter = getOption("NanoMethViz.site_filter", 3L)
) {
    # validate input
    chr <- as.character(chr)
    start <- as.numeric(start)
    end <- as.numeric(end)

    # validate genomic coordinates
    assert_valid_genomic_coords(chr, start, end)

    # validate site_filter
    if (!is.numeric(site_filter) || site_filter < 0) {
        stop(glue::glue(
            "site_filter must be a non-negative number. Got: {site_filter}\n",
            "This parameter filters sites with coverage below the threshold. Set using `options(site_filter = ...)`."
        ))
    }

    if (is(x, "NanoMethResult")) {
        x <- methy(x)
    }

    if (is(x, "ModBamResult")) {
        mod_code <- mod_code(x)
    }

    assert_that(
        same_length(chr, start, end),
        msg = "vectors 'chr', 'start' and 'end' must be the same length"
    )

    assert_that(site_filter > 0)

    input_type <- guess_input_type(x)
    if (input_type == "tabix") {
        out <- query_methy_tabix(x, chr, start, end, force = force)

    } else if (input_type == "modbam") {
        out <- query_methy_modbam(x, chr, start, end, mod_code)
        if (truncate) {
            truncate_fn <- function(x, start, end) {
                x %>%
                    dplyr::filter(
                        .data$pos >= start,
                        .data$pos <= end
                    )
            }

            # truncate each query by its start and end positions
            out <- purrr::pmap(list(out, start, end), truncate_fn)
        }
    } else {
        stop("'x' is not a recognised file type")
    }

    # compute modification probability column
    out <- map(out, function(x) {
        x %>%
            dplyr::mutate(mod_prob = sigmoid(.data$statistic))
    })

    # filter output for coverage
    out <- purrr::map(out, function(x) {
        pos_count_filter <- x %>%
            dplyr::count(.data$chr, .data$pos) %>%
            dplyr::rename("coverage" = "n") %>%
            dplyr::filter(.data$coverage >= site_filter) %>%
            dplyr::select(-"coverage")

        x %>%
            dplyr::inner_join(
                pos_count_filter,
                by = c("chr", "pos")
            )
    })

    if (simplify) {
        out <- dplyr::bind_rows(out)
    }

    out
}

query_methy_df <- function(x, regions, simplify = TRUE, force = FALSE) {
    assert_has_columns(regions, c("chr", "start", "end"))
    query_methy(x, regions$chr, regions$start, regions$end, simplify, force)
}

query_methy_gr <- function(x, regions, simplify = TRUE, force = FALSE) {
    assert_that(is(regions, "GRanges"))
    query_methy(
        x,
        as.character(GenomicRanges::seqnames(regions)),
        as.numeric(GenomicRanges::start(regions)),
        as.numeric(GenomicRanges::end(regions)),
        simplify,
        force
    )
}

query_methy_gene <- function(x, gene, window_prop = 0, simplify = TRUE) {
    if (!gene %in% exons(x)$symbol) {
        stop(glue::glue("gene '{gene}' not found in NanoMethViz::exons(x)"))
    }

    if (length(window_prop) == 1) {
        # convert to two sided window
        window_prop <- c(window_prop, window_prop)
    }

    pos_range <- gene_pos_range(x, gene)

    gene_width <- pos_range[2] - pos_range[1]
    window_left <- gene_width * window_prop[1]
    window_right <- gene_width * window_prop[2]

    chr <- exons(x) %>%
        dplyr::filter(.data$symbol == gene) %>%
        dplyr::slice(1) %>%
        dplyr::pull(chr)

    query_methy(
        x,
        chr = chr,
        start = pos_range[1] - window_left,
        end = pos_range[2] + window_right,
        simplify = simplify
    )
}

#' @importFrom utils read.table
query_methy_tabix <- function(x, chr, start, end, force) {
    get_missing_seqs <- function(x) {
        # query tabix index for missing sequences
        tabix_seqs <- get_tabix_sequences(paste0(x, ".tbi"))

        which(!chr %in% tabix_seqs)
    }

    # set up tabix file
    tabix_file <- Rsamtools::TabixFile(x)

    # set up output
    out <- list()
    for (i in seq_along(chr)) {
        nm <- paste0(chr[i], ":", start[i], "-", end[i])
        out[[nm]] <- empty_methy_query_output()
    }

    miss <- get_missing_seqs(x)
    # if missing sequences, warn and remove from query
    if (length(miss) != 0) {
        miss_seqs <- unique(chr[miss])
        warning(
            "requested sequences missing from tabix file and will be excluded from query:",
            paste(miss_seqs, collapse = ", ")
        )

        # remove queries with missing sequences
        chr <- chr[-miss]
        start <- start[-miss]
        end <- end[-miss]
    }

    # if no sequences left, return empty output
    if (length(chr) == 0) {
        if (!force) {
            stop("no chromosome matches between query and tabix file, please check chromosome format matches between query and methylation file.")
        }
    }

    # make query into a granges object
    query <- make_granges(chr, start, end)
    if (length(query) > 0) {
        query_result <- Rsamtools::scanTabix(tabix_file, param = query)
        # helper function to parse tabix output
        parse_tabix <- function(x) {
            if (length(x) == 0) {
                return(empty_methy_query_output())
            }

            # help readr function recognise input as data text rather than path
            if (length(x) == 1) {
                x <- paste0(x, "\n")
            }

            read_methy_lines(x)
        }

        methy_data <- lapply(
            query_result,
            parse_tabix
        )

        for (i in seq_along(methy_data)) {
            out[[names(query_result)[i]]] <- methy_data[[i]]
        }
    }

    out
}

query_methy_modbam <- function(x, chr, start, end, mod_code) {
    assertthat::assert_that(
        is(x, "ModBamResult") ||
        is(x, "ModBamFiles")
    )
    if (is(x, "ModBamResult")) {
        x <- methy(x)
    }

    assert_readable(x$path)

    # query each file
    x <- data.frame(
        sample = x$sample,
        path = x$path
    )

    # get data for each region from each file
    out <- x %>%
        dplyr::mutate(
            mod_table = map_rows(x, function(x) {
                read_modbam_table(
                    x$path,
                    chr = chr,
                    start = start,
                    end = end,
                    sample = x$sample,
                    mod_code = mod_code)
            })
        )

    # reduce list nesting by one level
    tables <- do.call(c, out$mod_table)

    # assign bind together tables from the same regions
    nms <- names(tables)

    if (is.null(nms)) {
        warning(glue::glue("no data found in {chr}:{start}-{end}"))
    }

    # extract and bind data from each region from each file
    extract_and_bind_data <- function(nm, tables) {
        out <- do.call(rbind, tables[names(tables) == nm])
        if (is.null(out)) {
            out <- empty_methy_query_output()
        }
        out
    }

    out <- purrr::map(
        unique(nms),
        extract_and_bind_data,
        tables = tables
    )

    out
}

can_open_tabix <- function(x) {
    assert_readable(x)
    out <- TRUE

    tryCatch(
        Rsamtools::TabixFile(x),
        warning = function(x) { out <<- FALSE },
        error = function(x) { out <<- FALSE }
    )

    return(out)
}

empty_methy_query_output <- function() {
    tibble::tibble(
        "sample" = character(),
        "chr" = character(),
        "pos" = integer(),
        "strand" = character(),
        "statistic" = numeric(),
        "read_name" = character()
    )
}

guess_input_type <- function(x) {
    if (is(x, "ModBamResult")) {
        return("modbam")
    }

    if (is(x, "character")) {
        if (all(map_lgl(x, can_open_tabix))) {
            return("tabix")
        }
    }

    return("uknown")
}

read_methy_lines <- function(x) {
    read.delim(
        textConnection(x),
        header = FALSE,
        col.names = methy_col_names()
    ) %>%
        tibble::tibble()
}

test_that("Querying methylation works", {
    # setup
    nmr <- load_example_nanomethresult()
    mbr <- load_example_modbamresult()

    # test
    # each query is half of Peg3
    queries <- tibble(
        chr = c("chr7", "chr7"),
        start = c(6703892, 6717162),
        end = c(6717161, 6730431)
    )

    # test basic operation
    methy_data_reg <- expect_no_warning(query_methy(nmr, "chr7", 6703892, 6730431))
    expect_equal(
        colnames(methy_data_reg),
        c("sample", "chr", "pos", "strand", "statistic", "read_name", "mod_prob")
    )
    expect_no_warning(query_methy(nmr, queries$chr, queries$start, queries$end))
    expect_no_warning(query_methy(nmr, queries$chr, queries$start, queries$end, simplify = FALSE))

    # test working on direct methy
    query_methy(methy(nmr), queries$chr[1], queries$start[1], queries$end[2])

    # test working on methods
    methy_data_fct <- expect_no_warning(query_methy(nmr, factor(queries$chr[1]), queries$start[1], queries$end[2]))
    expect_equal(methy_data_reg, methy_data_fct)

    # test working on gene
    expect_no_warning(methy_data_gene <- query_methy_gene(nmr, "Peg3"))
    expect_equal(methy_data_fct, methy_data_gene)

    # test warnings and errors
    expect_warning(expect_error(query_methy(nmr, "Missing", 1, 1000)))
    expect_error(query_methy_gene(nmr, "Missing"))

    # test working on table queries
    expect_no_warning(query_methy_df(methy(nmr), queries))

    # test working on GRanges
    regions_gr <- GenomicRanges::GRanges(queries)
    methy_data_gr <- expect_no_warning(query_methy_gr(nmr, regions_gr))
    expect_equal(methy_data_reg, methy_data_gr)

    # test working on modbam
    methy_data_modbam <- expect_no_warning(query_methy(mbr, queries$chr[1], queries$start[1], queries$end[2]))
    expect_equal(colnames(methy_data_reg), colnames(methy_data_modbam))

    queries_warn <- tibble(
        chr = c("chr7", "chr7", "chr7", "foo"),
        start = c(6703892, 6717162, 10000, 1),
        end = c(6717161, 6730431, 20000, 2)
    )

    # test ordering of unsimplified output
    expect_warning(
        methy_data_list <- query_methy(nmr, queries_warn$chr, queries_warn$start, queries_warn$end, simplify = FALSE),
        "requested sequences missing from tabix file and will be excluded from query:foo"
    )
    expect_true(length(methy_data_list) == 4)

    # test to make sure that reading from text connection works for methylation data
    methy_data_with_inf <- readLines(system.file("methy_data_with_inf.tsv", package = "NanoMethViz", mustWork = FALSE))
    methy_data <- expect_no_warning(read_methy_lines(methy_data_with_inf))
    expect_equal(nrow(methy_data), length(methy_data_with_inf))

    # test when query regions are empty
    queries_with_empty <- tibble(
        chr = c("chr7", "chr7", "chr7"),
        start = c(6703892, 6717162, 1),
        end = c(6717161, 6730431, 10)
    )
    expect_s3_class(query_methy_df(mbr, queries_with_empty), "data.frame")
})

#' Plot gene annotations with exons and introns
#'
#' @description
#' Creates a ggplot visualization of gene annotations showing exons and intron regions
#' within a specified genomic range.
#'
#' @param exons_df A data frame containing exon information with columns:
#'   \itemize{
#'     \item gene_id: Gene identifier
#'     \item transcript_id: Transcript identifier
#'     \item symbol: Gene symbol
#'     \item strand: Strand direction ("+", "-", or "*")
#'     \item start: Exon start position
#'     \item end: Exon end position
#'   }
#' @param plot_start Numeric, the start position of the plotting region
#' @param plot_end Numeric, the end position of the plotting region
#'
#' @return A ggplot object showing gene annotations. The plot has an attribute
#'   "plot_height" indicating the height needed to display all transcripts.
#'   Returns an empty plot with height 0 if no exons are in the region.
#'
#' @details
#' The function filters exons to the specified region, organizes them by transcript,
#' and visualizes them with connecting lines for introns. Gene symbols are placed
#' near the middle of each transcript. Arrows indicate the strand direction when available.
#'
#' @importFrom dplyr mutate group_by summarise filter inner_join arrange ungroup
#' @importFrom tidyr unnest
#' @importFrom tibble tibble
#' @importFrom ggplot2 ggplot theme_void ylim
#'
#' @noRd
plot_gene_annotation <- function(exons_df, plot_start, plot_end) {
    if (nrow(exons_df) == 0) {
        p <- ggplot2::ggplot() + ggplot2::theme_void()
        attr(p, "plot_height") <- 0
        return(p)
    }

    # remove transcripts outside plot area
    exons_df <- .anno_filter_regions(exons_df, plot_start, plot_end)

    exons_df <- exons_df %>%
        dplyr::mutate(
            uid = factor(paste(.data$gene_id, .data$transcript_id, sep = ".")),
            y_offset = 2.5 * as.integer(.data$uid) - 1
        )

    exons_count <- exons_df %>%
        dplyr::group_by(.data$uid) %>%
        dplyr::summarise(exons = dplyr::n())

    gap <- exons_df %>%
        dplyr::inner_join(exons_count, by = c("uid"), multiple = "all") %>%
        dplyr::filter(.data$exons > 1) %>%
        dplyr::group_by("transcript_id") %>%
        dplyr::arrange(.data$start) %>%
        dplyr::ungroup()

    if (nrow(gap) > 0) {
        gap <- gap %>%
            dplyr::group_by(.data$uid, .data$strand, .data$y_offset) %>%
            dplyr::summarise(
                gap_start = list(.data$end[-length(.data$end)]),
                gap_end = list(.data$start[-1])
            ) %>%
            tidyr::unnest(cols = c("gap_start", "gap_end"))
    } else {
        gap <- tibble::tibble(
            uid = character(),
            y_offset = numeric(),
            strand = character(),
            gap_start = numeric(),
            gap_end = numeric()
        )
    }

    gap_pos <- .anno_get_gaps(gap, "+")
    gap_neg <- .anno_get_gaps(gap, "-")
    gap_none <- .anno_get_gaps(gap, "*")

    region_width <- plot_end - plot_start
    gene_labels <- exons_df %>%
        dplyr::group_by(.data$gene_id, .data$symbol, .data$transcript_id, .data$y_offset, .data$strand) %>%
        dplyr::summarise(
            gene_middle = (min(.data$start) + max(.data$end)) / 2,
            label_pos = .data$gene_middle,
            label_pos = pmin(.data$label_pos, plot_end - region_width * 0.05),
            label_pos = pmax(.data$label_pos, plot_start +  region_width * 0.05)
        )

    exons_df <- .anno_truncate_region(exons_df, plot_start, plot_end, "*")

    gap_pos <- .anno_truncate_region(gap_pos, plot_start, plot_end, "+")
    gap_neg <- .anno_truncate_region(gap_neg, plot_start, plot_end, "-")
    gap_none <- .anno_truncate_region(gap_none, plot_start, plot_end, "*")
    gene_labels <- gene_labels %>%
        dplyr::filter(
            dplyr::between(.data$label_pos, plot_start, plot_end)
        )

    if (length(exons_df$y_offset) > 0) {
        plot_height <- 2 + max(exons_df$y_offset)
    } else {
        plot_height <- 0
    }

    p <- ggplot2::ggplot() +
        ggplot2::theme_void() +
        .anno_connector_arrows(gap_pos) +
        .anno_connector_arrows(gap_neg) +
        .anno_connector_lines(gap_none) +
        .anno_exons(exons_df) +
        .anno_gene_labels(gene_labels) +
        ggplot2::ylim(0, plot_height)

    attr(p, "plot_height") <- plot_height

    p
}

# helper functions ----
.anno_exons <- function(exons_df) {
    ggplot2::geom_rect(
        ggplot2::aes(
            xmin = .data$start,
            xmax = .data$end,
            ymin = .data$y_offset + 0.05,
            ymax = .data$y_offset + 0.55
        ),
        data = exons_df,
        fill = "#696969",
        colour = "black"
    )
}

.anno_connector_arrows <- function(gaps) {
    list(
        # first half with arrow
        ggplot2::geom_segment(
            ggplot2::aes(
                x = .data$start,
                xend = (.data$start + .data$end)/2,
                y = .data$y_offset + 0.275,
                yend = .data$y_offset + 0.275
            ),
            lineend = "butt",
            linejoin = "mitre",
            data = gaps,
            arrow = grid::arrow(
                type = "open",
                length = ggplot2::unit(5, "points")
            )
        ),
        # second half without arrow
        ggplot2::geom_segment(
            ggplot2::aes(
                x = (.data$start + .data$end)/2,
                xend = .data$end,
                y = .data$y_offset + 0.275,
                yend = .data$y_offset + 0.275
            ),
            lineend = "butt",
            linejoin = "mitre",
            data = gaps
        )
    )
}

.anno_connector_lines <- function(gaps) {
    ggplot2::geom_segment(
        ggplot2::aes(
            x = .data$end,
            xend = .data$start,
            y = .data$y_offset + 0.275,
            yend = .data$y_offset + 0.275
        ),
        data = gaps
    )
}

.anno_gene_labels <- function(gene_labels) {
    gene_labels$symbol <- as.character(gene_labels$symbol)
    gene_labels$symbol[gene_labels$strand == "+"] <- paste(
        gene_labels$symbol[gene_labels$strand == "+"],
        ">"
    )

    gene_labels$symbol[gene_labels$strand == "-"] <- paste(
        "<",
        gene_labels$symbol[gene_labels$strand == "-"]
    )

    ggplot2::geom_text(
        ggplot2::aes(x = .data$label_pos, y = .data$y_offset + 1.2, label = .data$symbol),
        data = gene_labels,
        hjust = "center",
        size = ggplot2::rel(2.5)
    )
}

.anno_filter_regions <- function(exons_df, plot_start, plot_end) {
    transcripts <- exons_df %>%
        dplyr::summarise(
            .by = "transcript_id",
            start = min(.data$start),
            end = max(.data$end)
        )

    transcripts <- transcripts %>%
        dplyr::filter(
            .data$start <= plot_end & .data$end >= plot_start
        )

    exons_df %>%
        dplyr::filter(
            .data$transcript_id %in% transcripts$transcript_id
        )
}

.anno_truncate_region <- function(x, plot_start, plot_end, strand) {
    if (strand == "-") {
        x <- x %>%
            dplyr::filter(.data$end <= plot_end, .data$start >= plot_start)
        x$end[x$end < plot_start] <- plot_start
        x$start[x$start > plot_end] <- plot_end
    } else {
        x <- x %>%
            dplyr::filter(.data$start <= plot_end, .data$end >= plot_start)
        x$start[x$start < plot_start] <- plot_start
        x$end[x$end > plot_end] <- plot_end
    }

    x
}

.anno_get_gaps <- function(gaps, strand = c("+", "-", "*")) {
    strand <- match.arg(strand)

    gaps <- gaps %>%
        split(gaps$strand)

    if (!is.null(gaps[[strand]])) {
        out <- gaps[[strand]]

        if (strand == "-") {
            temp <- out$gap_start
            out$gap_start <- out$gap_end
            out$gap_end <- temp
        }
    } else {
        out <- tibble::tibble(
            uid = character(0),
            y_offset = numeric(0),
            strand = character(0),
            gap_start = integer(0),
            gap_end = integer(0)
        )
    }

    dplyr::rename(
        out,
        start = "gap_start",
        end = "gap_end"
    )
}

#' Cluster regions by K-means
#'
#' Cluster regions by k-means based on their methylation profiles. In order to
#' cluster using k-means the methylation profile of each region is interpolated
#' and sampled at fixed points. The first 10 principal components are used for
#' the k-means clustering. The clustering is best behaved in regions of similar
#' width and CpG density.
#'
#' @param x the NanoMethResult object.
#' @param regions a table of regions containing at least columns chr, strand,
#'   start and end.
#' @param centers number of centers for k-means, identical to the number of
#'   output clusters.
#' @param grid_method the method for generating the sampling grid. The default
#'   option "density" attempts to create a grid with similar density as the
#'   data, "uniform" creates a grid of uniform density.
#'
#' @return the table of regions given by the 'regions' argument with the column
#'   'cluster' added.
#'
#' @importFrom stats quantile approxfun prcomp kmeans sd
#' @importFrom graphics hist
#' @export
#'
#' @examples
#' nmr <- load_example_nanomethresult()
#' gene_anno <- exons_to_genes(NanoMethViz::exons(nmr))
#' # uniform grid due to low number of input features
#' gene_anno_clustered <- cluster_regions(nmr, gene_anno, centers = 2, grid_method = "uniform")
#' plot_agg_regions(nmr, gene_anno_clustered, group_col = "cluster")
cluster_regions <- function(x, regions, centers = 2, grid_method = c("density", "uniform")) {
    grid_method <- match.arg(grid_method)

    methy_df <- tibble(
        regions,
        methy = query_methy(x, regions$chr, regions$start, regions$end, simplify = FALSE)
    )

    # rescale position values
    for (i in seq_len(nrow(methy_df))) {
        start <- methy_df$start[i]
        end <- methy_df$end[i]
        methy_df$methy[[i]] <- if (methy_df$strand[i] == "-") {
            # re-sort by position due to flip
            methy_df$methy[[i]] %>%
                mutate(pos = 1 - ((.data$pos - start) / (end - start))) %>%
                arrange(.data$pos)
        } else {
            # include * case
            methy_df$methy[[i]] %>%
                mutate(pos = (.data$pos - start) / (end - start))
        }
    }

    # take grid size as twice the 90% percentile
    grid_length <- 2 * floor(quantile(purrr::map_int(methy_df$methy, ~length(unique(.x$pos))), 0.9))

    if (grid_method == "uniform") {
        grid <- seq(0, 1, length.out = grid_length)
    } else if (grid_method == "density") {
        unique_pos <- unlist(purrr::map(methy_df$methy, ~unique(.x$pos)))
        hist_out <- hist(unique_pos, plot = FALSE, breaks = grid_length)
        epdf <- approxfun(hist_out$mids, hist_out$density / sum(hist_out$density), rule = 2)
        epdf_vals <- epdf(seq(0, 1, length.out = grid_length))

        ecdf_vals <- cumsum(epdf_vals) / max(cumsum(epdf_vals))
        inv_cdf <- suppressWarnings(
            approxfun(ecdf_vals, seq(0, 1, length.out = grid_length), rule = 2, ties = mean)
        )

        grid <- inv_cdf(seq(0, 1, length.out = grid_length))
    }

    sample_methy_grid <- function(x, grid) {
        summaried_methy <- x %>%
            group_by(.data$pos) %>%
            summarise(methy_prob = median(e1071::sigmoid(.data$statistic))) %>%
            ungroup()

        smoothed_f <- approxfun(
            summaried_methy$pos,
            summaried_methy$methy_prob,
            rule = 2
        )

        smoothed_f(grid)
    }

    methy_matrix <- map(methy_df$methy, sample_methy_grid, grid = grid) %>%
        unlist() %>%
        matrix(nrow = length(methy_df$methy), byrow = TRUE)

    km_pca <- kmeans(methy_matrix, centers = centers)

    row_means_split <- function(x, f) {
        out <- list()
        unique_f <- unique(f)

        for (i in seq_along(unique_f)) {
            out[[i]] <- colMeans(x[which(f == i), , drop = FALSE])
        }

        out
    }

    cluster_means <- row_means_split(methy_matrix, km_pca$cluster)

    out <- regions %>%
        mutate(cluster = km_pca$cluster)

    out$cluster_dev <- numeric(nrow(out))
    for (i in seq_len(nrow(out))) {
        out$cluster_dev[i] <- sum((methy_matrix[i, ] - cluster_means[[out$cluster[i]]])^2) / ncol(methy_matrix)
    }

    out %>%
        mutate(cluster = factor(.data$cluster))
}

#' Column names for methylation data
#'
#' @return column names for methylation data
#' @export
#'
#' @examples
#' methy_col_names()
methy_col_names <- function() {
    c(
        "sample",
        "chr",
        "pos",
        "strand",
        "statistic",
        "read_name"
    )
}

methy_col_types <- function() {
    readr::cols_only(
        sample = readr::col_factor(),
        chr = readr::col_factor(),
        pos = readr::col_integer(),
        strand = readr::col_factor(levels = c("+", "-", "*")),
        statistic = readr::col_double(),
        read_name = readr::col_character()
    )
}

f5c_col_types <- function() {
    readr::cols_only(
        chromosome = readr::col_character(),
        start = readr::col_integer(),
        read_name = readr::col_character(),
        log_lik_ratio = readr::col_double(),
        num_cpgs = readr::col_double(),
        sequence = readr::col_character()
    )
}

nanopolish_col_types <- function() {
    readr::cols_only(
        chromosome = readr::col_character(),
        start = readr::col_integer(),
        strand = readr::col_factor(levels = c("+", "-", "*")),
        read_name = readr::col_character(),
        log_lik_ratio = readr::col_double(),
        num_motifs = readr::col_double(),
        sequence = readr::col_character()
    )
}

megalodon_col_types_old <- function() {
    readr::cols_only(
        read_id = readr::col_character(),
        chrm = readr::col_character(),
        strand = readr::col_integer(),
        pos = readr::col_integer(),
        mod_log_prob = readr::col_double()
    )
}

megalodon_col_types <- function() {
    readr::cols_only(
        read_id = readr::col_character(),
        chrm = readr::col_character(),
        strand = readr::col_character(),
        pos = readr::col_integer(),
        mod_log_prob = readr::col_double()
    )
}

bedmethy_col_names <- function() {
    c(
        "chrom",
        "start",
        "end",
        "modified_base_code",
        "score",
        "strand",
        "start_incl",
        "end_incl",
        "color",
        "n_valid_cov",
        "fraction_modified",
        "n_mod",
        "n_canonical",
        "n_other_mod",
        "n_delete",
        "n_fail",
        "n_diff",
        "n_nocall"
    )
}

modkit_col_types <- function() {
    readr::cols_only(
        read_id = readr::col_character(),
        forward_read_position = readr::col_double(),
        ref_position = readr::col_double(),
        chrom = readr::col_character(),
        mod_strand = readr::col_character(),
        ref_strand = readr::col_character(),
        ref_mod_strand = readr::col_character(),
        fw_soft_clipped_start = readr::col_double(),
        fw_soft_clipped_end = readr::col_double(),
        read_length = readr::col_double(),
        mod_qual = readr::col_double(),
        mod_code = readr::col_character(),
        base_qual = readr::col_double(),
        ref_kmer = readr::col_character(),
        query_kmer = readr::col_character(),
        canonical_base = readr::col_character(),
        modified_primary_base = readr::col_character(),
        inferred = readr::col_logical(),
        flag = readr::col_double()
    )
}

#' Cluster reads based on methylation
#'
#' @param x a ModBamResult object.
#' @param chr the chromosome name where to find the region.
#' @param start the start position of the region.
#' @param end the end position of the region.
#' @param min_pts the minimum number of points needed to form a cluster (default = 10).
#'
#' @return A tibble with information about each read's cluster assignment and read statistics.
#'
#' @import tidyr
#' @import dplyr
#' @import dbscan
#' @importFrom tibble rownames_to_column
cluster_reads <- function(x, chr, start, end, min_pts = 5) {
    assertthat::assert_that(
        is(x, "ModBamResult"),
        assertthat::is.string(chr) || (is.factor(chr) && assertthat::is.scalar(chr)),
        assertthat::is.number(start) && assertthat::is.number(end),
        assertthat::is.number(min_pts) && min_pts >= 1
    )

    # query data
    methy_data <- query_methy(x, chr, start, end)

    if (nrow(methy_data) == 0) {
        stop(glue::glue(
            "No reads containing methylation data found in region {chr}:{start}-{end}.\n",
            "This could be due to:\n",
            "  - No data in this genomic region\n",
            "  - Incorrect chromosome naming\n",
            "  - Region outside of data coverage\n",
            "Please check your genomic coordinates and data."
        ))
    }

    methy_data <- methy_data %>%
        dplyr::filter(.data$pos >= start & .data$pos < end)

    read_stats <- get_read_stats(methy_data)

    # identify the read names whose span is at least 90% the length of maximum span
    # filter methylation data for only those reads that meet the above condition of span
    max_span <- max(read_stats$span)
    keep_reads <- read_stats$read_name[read_stats$span > 0.9 * max_span]
    methy_data <- methy_data %>%
        dplyr::filter(.data$read_name %in% keep_reads)

    # convert methylation data into a matrix with one row for each read name
    mod_mat <- methy_data %>%
        dplyr::select("read_name", "pos", "mod_prob") %>%
        dplyr::arrange(.data$pos) %>%
        tidyr::pivot_wider(names_from = "pos", values_from = "mod_prob") %>%
        df_to_matrix()

    # pre-check before filtering
    if (nrow(mod_mat) < min_pts) {
        stop(glue::glue(
            "Insufficient reads for clustering: found {nrow(mod_mat)} reads but need at least {min_pts}.\n",
            "Try reducing 'min_pts' parameter or expanding the genomic region."
        ))
    }

    # remove positions with high missingness (>60%) then reads with high missingness (>30%)
    mod_mat_filled <- mod_mat[order(rownames(mod_mat)), ]
    col_missingness <- mat_col_map(mod_mat_filled, missingness)
    mod_mat_filled <- mod_mat_filled[, col_missingness < 0.6]
    row_missingness <- mat_row_map(mod_mat_filled, missingness)
    mod_mat_filled <- mod_mat_filled[row_missingness < 0.3, ]

    # fill in missing values with mean methylation probability across that read
    for (i in seq_len(nrow(mod_mat_filled))) {
        mod_mat_filled[i, is.na(mod_mat_filled[i, ])] <- mean(mod_mat_filled[i, ], na.rm = TRUE)
    }

    # post-check before filtering
    if (nrow(mod_mat_filled) < min_pts) {
        stop(glue::glue(
            "Insufficient reads after filtering: {nrow(mod_mat_filled)} reads remaining but need at least {min_pts}.\n",
            "Try reducing 'min_pts' parameter or adjusting filtering criteria."
        ))
    }

    # cluster reads using HDBSCAN algorithm with specified minimum number of points
    dbsc <- dbscan::hdbscan(mod_mat_filled, minPts = min_pts)
    clust_df <- data.frame(read_name = rownames(mod_mat_filled), cluster_id = dbsc$cluster)

    # merge and process results of cluster analysis and read statistics
    clust_df %>%
        dplyr::inner_join(read_stats, by = "read_name") %>%
        dplyr::arrange(.data$cluster_id) %>%
        dplyr::mutate(
            cluster_id = as.factor(.data$cluster_id),
            start = as.integer(.data$start),
            end = as.integer(.data$end),
            span = as.integer(.data$span)
        )
}

# summarize read statistics (start, end, strand) based on same read name
get_read_stats <- function(methy_data) {
    methy_data %>%
        group_by(.data$read_name) %>%
        summarise(
            start = min(.data$pos),
            end = max(.data$pos),
            mean = mean(.data$mod_prob, na.rm = TRUE),
            span = .data$end - .data$start,
            strand = unique(.data$strand)
        ) %>%
        arrange(.data$strand)
}

plot_gene_impl <- function(
        x,
        gene,
        window_prop = 0.3,
        anno_regions = NULL,
        binary_threshold = NULL,
        avg_method = c("mean", "median"),
        spaghetti = FALSE,
        heatmap = TRUE,
        heatmap_subsample = 50,
        smoothing_window = 2000,
        gene_anno = TRUE,
        palette = ggplot2::scale_colour_brewer(palette = "Set1"),
        line_size = 1,
        mod_scale = c(0, 1),
        span = NULL
) {
    if (!missing("span")) {
        warning("the 'span' argument has been deprecated, please use 'smoothing_window' instead")
    }

    avg_method <- match.arg(avg_method)

    # validate genes
    if (is.na(gene) || gene == "") {
        stop("Gene symbol cannot be empty or NA. Please provide a valid gene symbol.")
    }

    if (nrow(exons(x)) == 0) {
        stop(glue::glue(
            "No exon annotation found in the data object.\n",
            "Gene '{gene}' cannot be plotted without exon information.\n",
            "Please add exon annotation using: exons(your_object) <- your_exons"
        ))
    }

    if (length(window_prop) == 1) {
        # convert to two sided window_prop
        window_prop <- c(window_prop, window_prop)
    }

    exons_anno <- query_exons_symbol(x, symbol = gene)

    if (nrow(exons_anno) == 0) {
        available_genes <- unique(exons(x)$symbol)
        similar_genes <- available_genes[grepl(paste0("^", substr(gene, 1, 3)), available_genes, ignore.case = TRUE)]

        error_msg <- glue::glue(
            "Gene '{gene}' not found in exon annotation.\n",
            "Please check the gene symbol spelling."
        )

        if (length(similar_genes) > 0 && length(similar_genes) <= 10) {
            error_msg <- paste0(
                error_msg, "\n",
                glue::glue("Similar genes found: {paste(similar_genes, collapse = ', ')}")
            )
        } else if (length(available_genes) <= 20) {
            error_msg <- paste0(
                error_msg, "\n",
                glue::glue("Available genes: {paste(available_genes, collapse = ', ')}")
            )
        } else {
            error_msg <- paste0(
                error_msg, "\n",
                glue::glue("Use unique(exons(your_object)$symbol) to see all {length(available_genes)} available genes.")
            )
        }

        stop(error_msg)
    }

    feature_chr <- unique(exons_anno$chr)
    feature_start <- min(exons_anno$start)
    feature_end <- max(exons_anno$end)

    plot_region(
        x = x,
        chr = feature_chr,
        start = feature_start,
        end = feature_end,
        anno_regions = anno_regions,
        binary_threshold = binary_threshold,
        avg_method = avg_method,
        spaghetti = spaghetti,
        heatmap = heatmap,
        heatmap_subsample = heatmap_subsample,
        smoothing_window = smoothing_window,
        gene_anno = gene_anno,
        window_prop = window_prop,
        palette = palette,
        line_size = line_size,
        mod_scale = mod_scale
    )
}

#' @rdname plot_gene
#'
#' @inheritParams plot_region
#'
#' @details
#' This function plots the methylation data for a given gene. The main trendline plot shows the average methylation
#' probability across the gene. The heatmap plot shows the methylation probability for each read across the gene. The
#' gene annotation plot shows the exons of the gene. In the heatmap, each row represents one or more non-overlapping reads
#' where the coloured segments represent the methylation probability at each position. Data along a read is connected by
#' a grey line. The gene annotation plot shows the isoforms and exons of the gene, with arrows indicating the direction
#' of transcription.
#'
#' Since V3.0.0 NanoMethViz has changed the smoothing strategy from a loess smoothing to a weighted moving average. This
#' is because the loess smoothing was too computationally expensive for large datasets and had a span parameter that was
#' difficult to tune. The new smoothing strategy is controlled by the smoothing_window argument.
#'
#' @examples
#' nmr <- load_example_nanomethresult()
#' plot_gene(nmr, "Peg3")
#'
#' @export
setMethod("plot_gene", signature(x = "NanoMethResult", gene = "character"),
    plot_gene_impl
)

#' @describeIn plot_gene S4 method for ModBamResult
setMethod("plot_gene", signature(x = "ModBamResult", gene = "character"),
    plot_gene_impl
)

#' @importFrom stats median
#' @importFrom forcats fct_reorder
lollipop <- function(x, chr, start, end, binary = FALSE) {
    methy_data <- query_methy(x, chr, start, end)

    numeric_median <- function(x) stats::median(as.numeric(x))

    df <- methy_data %>%
        dplyr::group_by(.data$sample, .data$pos) %>%
        dplyr::summarise(val = mean(e1071::sigmoid(.data$statistic))) %>%
        dplyr::mutate(
            pos = as.factor(scales::comma(.data$pos, accuracy = 1)),
            group = str_extract(.data$sample, "bl6|cast")
        ) %>%
        dplyr::ungroup() %>%
        dplyr::mutate(
            group = factor(.data$group),
            sample = forcats::fct_reorder(.data$sample, .data$group, numeric_median),
            state = ifelse(.data$val > 0.5, "methylated", "unmethylated")
        )

    p <- ggplot2::ggplot(df, aes(x = .data$pos, y = .data$sample)) +
        ggplot2::geom_hline(yintercept = df$sample, size = 2) +
        ggplot2::theme(
            axis.text.x = ggplot2::element_blank(),
            axis.ticks.x = ggplot2::element_blank(),
            axis.title.x = ggplot2::element_blank())

    if (binary) {
        p <- p +
            ggplot2::geom_point(aes(fill = .data$state), pch = 21, size = 7) +
            ggplot2::scale_fill_manual(values = c("black", "white"))

    } else {
        p <- p +
            ggplot2::geom_point(aes(fill = .data$val), pch = 21, size = 7) +
            ggplot2::scale_fill_steps(low = "white", high = "blueviolet", n.breaks = 6)
    }

    p
}

#' Plot violin for regions
#'
#' This function plots a violin plot of the methylation proportion for each region in the regions table. The methylation proportion is calculated as the mean of the modification probability wihtin each region and the violin represents the . The regions are then grouped and coloured by the group_col column in the regions table or samples(x).
#'
#' @param x the NanoMethResult object.
#' @param regions a table of regions containing at least columns chr, strand,
#'   start and end. Any additional columns can be used for grouping.
#' @param binary_threshold the modification probability such that calls with
#'   modification probability above the threshold are considered methylated, and
#'   those with probability equal or below are considered unmethylated.
#' @param group_col the column to group aggregated trends by. This column can
#'   be in from the regions table or samples(x).
#' @param palette the ggplot colour palette used for groups.
#'
#' @return a ggplot object containing the methylation violin plot.
#'
#' @examples
#' nmr <- load_example_nanomethresult()
#' gene_anno <- exons_to_genes(NanoMethViz::exons(nmr))
#' plot_violin(nmr, gene_anno)
#' plot_violin(nmr, gene_anno, group_col = "sample")
#'
#' @export
plot_violin <- function(
    x,
    regions,
    binary_threshold = 0.5,
    group_col = "group",
    palette = ggplot2::scale_colour_brewer(palette = "Set1")
) {
    if (!is.null(group_col)) {
        avail_columns <- c(colnames(samples(x)), colnames(regions))
        assertthat::assert_that(
            group_col %in% avail_columns,
            msg = glue::glue("'{group_col}' could not be found in columns of 'regions' or samples(x)")
        )
    }

    # grouped regions crashes downstream operations
    regions <- ungroup(regions)

    regions$methy_data <- purrr::map(
        seq_len(nrow(regions)),
        ~query_methy(
            x,
            regions$chr[.x],
            regions$start[.x],
            regions$end[.x],
            force = TRUE
        )
    )

    # remove regions with no data
    regions <- regions %>%
        dplyr::filter(purrr::map_lgl(.data$methy_data, function(x) nrow(x) != 0))

    region_data <- regions %>%
        dplyr::select(!dplyr::any_of(c("chr", "strand", "start", "end"))) %>%
        tidyr::unnest("methy_data") %>%
        dplyr::summarise(
            methy_prop = mean(.data$mod_prob > binary_threshold),
            .by = c("gene_id", "symbol", "sample", "chr", "strand")) %>%
        dplyr::inner_join(samples(x), by = "sample", multiple = "all")

    if (!is.null(group_col)) {
        aes_spec <- ggplot2::aes(
                x = .data[[group_col]],
                y = .data$methy_prop,
                col = .data$group)
    } else {
        aes_spec <- ggplot2::aes(
                x = .data[[group_col]],
                y = .data$methy_prop)
    }

    ggplot2::ggplot(region_data, aes_spec) +
        ggplot2::geom_violin(draw_quantiles = 0.5) +
        ggplot2::scale_y_continuous(limits = c(0, 1)) +
        ggplot2::theme_bw() +
        palette
}

# expand multiple motifs
expand_motifs <- function(x) {
    x_mult <- x[x$num_cpgs > 1, ]
    x <- x[x$num_cpgs == 1, ]
    x$num_cpgs <- NULL

    mult_expand <- with(x_mult, {
        start_offsets <- stringr::str_locate_all(
            stringr::str_sub(sequence, start = 6),
            "CG"
        ) %>%
        purrr::map(
            function(x) {
                x[, 1] - 1
            }
        ) %>%
        unlist()

        tidyr::uncount(x_mult, .data$num_cpgs) %>%
            dplyr::mutate(start = start_offsets + .data$start) %>%
            dplyr::select(!"num_cpgs")
    })

    rbind(x, mult_expand) %>%
        dplyr::select(-sequence)
}

reformat_f5c <- function(x, sample) {
    x <- x %>%
        expand_motifs() %>%
        add_column(sample = sample, .before = 1)

    x %>%
        dplyr::transmute(
            sample = factor(.data$sample),
            chr = factor(.data$chromosome),
            pos = as.integer(.data$start) + 1,
            strand = factor("*", levels = c("+", "-", "*")),
            statistic = .data$log_lik_ratio,
            read_name = .data$read_name
        )
}

reformat_nanopolish <- function(x, sample) {
    x <- x %>%
        dplyr::rename(num_cpgs = "num_motifs") %>%
        expand_motifs() %>%
        add_column(sample = sample, .before = 1)

    x %>%
        dplyr::transmute(
            sample = factor(.data$sample),
            chr = factor(.data$chromosome),
            pos = as.integer(.data$start) + 1,
            strand = .data$strand,
            statistic = .data$log_lik_ratio,
            read_name = .data$read_name
        )
}

reformat_megalodon <- function(x, sample) {
    x %>%
        dplyr::rename(
            chr = "chrm",
            statistic = "mod_log_prob",
            read_name = "read_id") %>%
        add_column(sample = sample, .before = 1) %>%
        dplyr::mutate(
            sample = as.factor(.data$sample),
            chr = factor(.data$chr),
            statistic = logit(exp(.data$statistic)),
            pos = as.integer(.data$pos) + 1,
            strand = factor(.data$strand, levels = c("+", "-", "*"))) %>%
        dplyr::select(methy_col_names())
}

reformat_modkit <- function(x, sample) {
    x %>%
        dplyr::filter(ref_position > 0) %>% # remove unmapped positions
        add_column(sample = sample, .before = 1) %>%
        dplyr::rename(
            chr = "chrom",
            pos = "ref_position",
            strand = "mod_strand",
            statistic = "mod_qual",
            read_name = "read_id"
        ) %>%
        dplyr::mutate(
            sample = as.factor(.data$sample),
            chr = factor(.data$chr),
            pos = as.integer(.data$pos),
            strand = factor(.data$strand, levels = c("+", "-", "*")),
            statistic = logit(.data$statistic)
        ) %>%
        dplyr::select(
            "sample",
            "chr",
            "pos",
            "strand",
            "statistic",
            "read_name"
        )
}

guess_methy_source <- function(methy_file) {
    assert_readable(methy_file)

    readr::local_edition(1) # temporary fix for vroom bad value
    first_line <- readr::read_lines(methy_file, n_max = 1)

    # Be robust to BOM and whitespace
    header <- stringr::str_replace(first_line, "^\ufeff", "")
    header <- stringr::str_trim(header)

    # Detect delimiter (prefer tab, fallback to comma, then whitespace)
    delim <- if (stringr::str_detect(header, "\t")) "\t"
        else if (stringr::str_detect(header, ",")) ","
        else "\\s+"

    cols <- stringr::str_split(header, delim)[[1]]
    cols <- stringr::str_trim(cols)
    cols_lower <- stringr::str_to_lower(cols)

    has_all <- function(required) all(required %in% cols_lower)

    # Identify by minimal, distinctive column sets; allow extras and any order
    if (has_all(c("read_id", "chrm", "strand", "pos", "mod_log_prob"))) {
        return("megalodon")
    }

    if (has_all(c("read_id", "ref_position", "chrom", "mod_strand", "mod_qual"))) {
        return("modkit")
    }

    if (has_all(c("chromosome", "strand", "start", "read_name", "log_lik_ratio", "num_motifs", "sequence"))) {
        return("nanopolish")
    }

    if (has_all(c("chromosome", "start", "read_name", "log_lik_ratio", "num_cpgs", "sequence"))) {
        return("f5c")
    }

    stop("Format not recognised.")
}

#' Convert methylation calls to NanoMethViz format
#' @keywords internal
#'
#' @param input_files the files to convert
#' @param output_file the output file to write results to (must end in .bgz)
#' @param samples the names of samples corresponding to each file
#' @param verbose TRUE if progress messages are to be printed
#'
#' @return invisibly returns the output file path, creates a tabix file (.bgz)
#'   and its index (.bgz.tbi)
convert_methy_format <- function(
    input_files,
    output_file,
    samples = fs::path_ext_remove(fs::path_file(input_files)),
    verbose = TRUE
) {
    for (f in input_files) {
        assert_readable(f)
    }

    assert_that(
        is.character(output_file)
    )

    assert_that(is.dir(fs::path_dir(output_file)))
    file.create(path.expand(output_file))
    assert_that(is.writeable(output_file))

    for (element in vec_zip(file = input_files, sample = samples)) {
        if (verbose) {
            message(glue::glue("processing {element$file}..."))
        }
        methy_source <- guess_methy_source(element$file)
        if (verbose) {
            message(glue::glue("guessing file is produced by {methy_source}..."))
        }

        col_types <- switch(
            methy_source,
            "nanopolish" = nanopolish_col_types(),
            "f5c" = f5c_col_types(),
            "megalodon" = megalodon_col_types(),
            "modkit" = modkit_col_types()
        )

        reformatter <- switch(
            methy_source,
            "nanopolish" = reformat_nanopolish,
            "f5c" = reformat_f5c,
            "megalodon" = reformat_megalodon,
            "modkit" = reformat_modkit
        )

        writer_fn <- function(x, i) {
            readr::write_tsv(
                reformatter(x, sample = element$sample),
                file = output_file,
                append = TRUE
            )
        }
        readr::local_edition(1) # temporary fix for vroom bad value
        readr::read_tsv_chunked(
            element$file,
            col_types = col_types,
            readr::SideEffectChunkCallback$new(writer_fn)
        )
    }

    invisible(output_file)
}

# Skeleton for merge function for two methylation files, currently Unix only
merge_methy_files <- function(inputs, output) {
    assertthat::assert_that(
        all(fs::is_file(inputs)),
        all(fs::file_exists(inputs))
    )

    assertthat::assert_that(
        stringr::str_detect(output, "[.]bgz$"),
        msg = "'output' must end in .bgz"
    )

    output <- stringr::str_remove(output, ".bgz")

    temp_files <- purrr::map_chr(seq_along(inputs), ~tempfile())

    temp_merged <- tempfile()

    purrr::walk2(
        inputs,
        temp_files,
        ~R.utils::gunzip(.x, destname = .y, remove = FALSE)
    )

    files_str <- paste(temp_files, collapse = " ")
    cmd <- glue::glue("sort -m -k2,3V -o {temp_merged} {files_str}")
    system(cmd)

    fs::file_copy(temp_merged, output, overwrite = TRUE)
    if (fs::file_exists(paste0(output, ".bgz.tbi"))) {
        fs::file_delete(paste0(output, ".bgz.tbi"))
    }

    tabix_compress(output)
    fs::file_delete(output)
}

plot_methy_data_heatmap <- function(
        methy_data,
        pos_style,
        subsample,
        group_col = "group",
        mod_scale = c(0, 1)
) {
    # subsample reads down to a certain number of read groups
    subsample_groups <- function(x, key, subsample) {
        if (nrow(x) > subsample) {
            x <- dplyr::sample_n(x, subsample)
        }
        x
    }

    methy_data <- methy_data %>%
        dplyr::nest_by(group = .data[[group_col]], read_group = .data$read_group) %>%
        dplyr::group_by(.data$group) %>%
        dplyr::group_modify(subsample_groups, subsample = subsample)
    methy_data <- tidyr::unnest(methy_data, "data")

    read_data <- methy_data %>%
        dplyr::group_by(.data$read_name) %>%
        dplyr::summarise(start = min(.data$pos), end = max(.data$pos)) %>%
        dplyr::inner_join(
            dplyr::select(methy_data, "read_name", "read_group", "group"),
            by = "read_name"
        )

    if (pos_style == "compact") {
        # only plots sites with measured modification, evenly spaced
        p <- ggplot2::ggplot(
            methy_data,
            aes(
                x = factor(.data$pos),
                y = .data$read_group,
                fill = .data$mod_prob
            )
        ) +
            ggplot2::geom_raster() +
            heatmap_fill_scale +
            ggplot2::theme(
                axis.ticks.x = ggplot2::element_blank(),
                axis.text.x = ggplot2::element_blank(),
                axis.ticks.y = ggplot2::element_blank(),
                axis.title.y = ggplot2::element_blank(),
                axis.text.y = ggplot2::element_blank(),
                panel.grid.minor = ggplot2::element_blank(),
                panel.grid.major = ggplot2::element_blank()
            ) +
            ggplot2::xlab("Site")
    } else if (pos_style == "to_scale") {
        # plots all sites in range, evenly spaced with square geoms
        # data will overlap
        p <- ggplot2::ggplot(methy_data, aes(y = .data$read_group)) +
            ggplot2::geom_errorbarh(
                ggplot2::aes(xmin = .data$start, xmax = .data$end),
                data = read_data,
                alpha = 1,
                color = "darkgray",
                linewidth = 0.8,
                height = 0
            ) +
            ggplot2::geom_point(aes(x = .data$pos, col = .data$mod_prob), alpha = 1, shape = 15) +
            ggplot2::scale_x_continuous(
                labels = scales::label_number(scale_cut = scales::cut_si("b")),
                expand = ggplot2::expansion(0, 0)
            ) +
            heatmap_col_scale +
            theme_methy_heatmap +
            ggplot2::xlab("Position")
    }

    p + ggplot2::facet_wrap(~group, scales = "free_y", ncol = 1, strip.position = "right")
}

#' Convert BSseq object to edgeR methylation matrix
#'
#' @param bsseq the BSseq object.
#' @param regions the regions to calculate log-methylation ratios over. If left NULL, ratios will be calculated per
#'   site.
#'
#' @return a matrix compatible with the edgeR differential methylation pipeline
#' @export
#'
#' @examples
#' methy <- system.file("methy_subset.tsv.bgz", package = "NanoMethViz", mustWork = FALSE)
#' bsseq <- methy_to_bsseq(methy)
#' edger_mat <- bsseq_to_edger(bsseq)
bsseq_to_edger <- function(bsseq, regions = NULL) {
    edger_col_names <- bsseq_to_edger.get_edger_col_names(bsseq)

    if (!is.null(regions)) {
        assertthat::assert_that(is(regions, "data.frame"))
        regions <- GenomicRanges::GRanges(regions)
        edger_row_names <- regions$gene_id
    } else {
        edger_row_names <- bsseq_to_edger.get_edger_row_names(bsseq)
    }

    # construct matrix
    methylated <- bsseq_to_edger.get_me_mat(bsseq, regions)
    unmethylated <- bsseq_to_edger.get_un_mat(bsseq, regions)

    edger_mat <- matrix(
        0,
        ncol = 2 * ncol(methylated),
        nrow = nrow(methylated),
        dimnames = list(edger_row_names, edger_col_names)
    )

    for (i in 0:(ncol(methylated) - 1)) {
        edger_mat[, 2 * i + 1] <- methylated[, i + 1]
        edger_mat[, 2 * i + 2] <- unmethylated[, i + 1]
    }

    edger_mat
}

#' Convert NanoMethResult object to edgeR methylation matrix
#'
#' @inheritParams methy_to_bsseq
#' @param regions the regions to calculate log-methylation ratios over. If left
#' NULL, ratios will be calculated per site.
#'
#' @return a matrix compatible with the edgeR differential methylation pipeline
#' @export
#'
#' @examples
#' nmr <- load_example_nanomethresult()
#' edger_mat <- methy_to_edger(nmr)
#'
methy_to_edger <- function(methy, regions = NULL, out_folder = tempdir(), verbose = TRUE) {
    bsseq <- methy_to_bsseq(methy = methy, out_folder = out_folder, verbose = verbose)
    bsseq_to_edger(bsseq, regions = regions)
}

#' Convert BSseq object to log-methylation-ratio matrix
#'
#' Creates a log-methylation-ratio matrix from a BSseq object that is useful for
#' dimensionality reduction plots.
#'
#' @param bsseq the BSseq object.
#' @param regions the regions to calculate log-methylation ratios over. If left NULL, ratios will be calculated per
#'   site.
#' @param prior_count the prior count added to avoid taking log of 0.
#' @param drop_na whether to drop rows with all NA values.
#'
#' @return a matrix containing log-methylation-ratios.
#' @export
#'
#' @examples
#' nmr <- load_example_nanomethresult()
#' bsseq <- methy_to_bsseq(nmr)
#' regions <- exons_to_genes(NanoMethViz::exons(nmr))
#' log_m_ratio <- bsseq_to_log_methy_ratio(bsseq, regions)

bsseq_to_log_methy_ratio <- function(bsseq, regions = NULL, prior_count = 2, drop_na = TRUE) {
    assertthat::assert_that(prior_count >= 0)

    if (prior_count < 1) {
        warning("prior_count of 1 or higher is recommended")
    }

    if (!is.null(regions)) {
        assertthat::assert_that(is(regions, "data.frame"))
        regions <- GenomicRanges::GRanges(regions)
        row_names <- regions$gene_id
    } else {
        row_names <- bsseq_to_edger.get_edger_row_names(bsseq)
    }

    col_names <- SummarizedExperiment::colData(bsseq)$sample
    methylated <- bsseq_to_edger.get_me_mat(bsseq, regions)
    unmethylated <- bsseq_to_edger.get_un_mat(bsseq, regions)

    log_mat <- log2(methylated + prior_count) - log2(unmethylated + prior_count)

    dimnames(log_mat) <- list(row_names, col_names)

    if (drop_na) {
        # drop rows that are all NA
        log_mat <- log_mat[!apply(is.na(log_mat), 1, all), ]
    }

    log_mat
}

# helper functions ----
bsseq_to_edger.get_me_mat <- function(bsseq, regions) {
    if (!is.null(regions)) {
        bsseq::getCoverage(bsseq, regions, type = "M", what = "perRegionTotal")
    } else {
        bsseq::getCoverage(bsseq, type = "M")
    }
}

bsseq_to_edger.get_un_mat <- function(bsseq, regions) {
    if (!is.null(regions)) {
        cov <- bsseq::getCoverage(bsseq, regions, type = "Cov", what = "perRegionTotal")
    } else {
        cov <- bsseq::getCoverage(bsseq, type = "Cov")
    }
    me_mat <- bsseq_to_edger.get_me_mat(bsseq, regions)
    cov - me_mat
}

bsseq_to_edger.get_edger_col_names <- function(bsseq) {
    samples <- SummarizedExperiment::colData(bsseq)$sample
    me_names <- paste0(samples, "_Me")
    un_names <- paste0(samples, "_Un")
    edger_col_names <- character(2 * length(samples))
    for (i in 0:(length(samples) - 1)) {
        edger_col_names[2 * i + 1] <- me_names[i + 1]
        edger_col_names[2 * i + 2] <- un_names[i + 1]
    }

    edger_col_names
}

bsseq_to_edger.get_edger_row_names <- function(bsseq) {
    gr <- bsseq::getBSseq(bsseq, type = "gr")
    seq <- as.character(SummarizedExperiment::seqnames(gr))
    pos <- as.integer(SummarizedExperiment::start(gr))
    edger_row_names <- paste(seq, pos, sep = "-")

    edger_row_names
}

test_that("Conversion functions work", {
    # setup
    nmr <- load_example_nanomethresult()

    # tests
    bsseq <- expect_no_warning(methy_to_bsseq(nmr))
    expect_no_warning(methy_to_bsseq(methy(nmr)))
    regions <- expect_no_warning(exons_to_genes(nmr))
    expect_no_warning(exons_to_genes(NanoMethViz::exons(nmr)))

    # convert via bsseq
    edger_site <- expect_no_warning(bsseq_to_edger(bsseq))
    expect_ncol(edger_site, 2*ncol(bsseq))
    expect_nrow(edger_site, nrow(bsseq))

    # convert directly from nmr
    edger_direct_site <- expect_no_warning(methy_to_edger(nmr))
    expect_ncol(edger_direct_site, 2*ncol(bsseq))
    expect_nrow(edger_direct_site, nrow(bsseq))

    # should be identical
    expect_identical(edger_site, edger_direct_site)

    # convert to edgeR
    edger_region <- expect_no_warning(bsseq_to_edger(bsseq, regions))
    expect_ncol(edger_region, 2*ncol(bsseq))
    expect_nrow(edger_region, nrow(regions))

    # convert to log-methylation-ratio
    lmr_regions <- expect_no_warning(bsseq_to_log_methy_ratio(bsseq, regions))
    expect_ncol(lmr_regions, ncol(bsseq))
    expect_nrow(lmr_regions, nrow(regions))

    expect_warning(
        bsseq_to_log_methy_ratio(bsseq, regions, prior_count = 0.5),
        "prior_count of 1 or higher is recommended"
   )

    expect_error(bsseq_to_log_methy_ratio(bsseq, regions, prior_count = -1))
})

 # heatmap
theme_methy_heatmap <- ggplot2::theme_bw() +
        ggplot2::theme(
            axis.ticks.y = ggplot2::element_blank(),
            axis.title.y = ggplot2::element_blank(),
            axis.text.y = ggplot2::element_blank(),
            panel.grid.major = ggplot2::element_blank(),
            panel.grid.minor = ggplot2::element_blank()
        )

heatmap_fill_scale <- ggplot2::scale_fill_gradient2(low = "#CCCC3D", mid = "#53857E", high = "#1933B2", midpoint = 0.5)
heatmap_col_scale <- ggplot2::scale_color_gradient2(low = "#CCCC3D", mid = "#53857E", high = "#1933B2", midpoint = 0.5)

# Old scheme
# heatmap_fill_scale <- scico::scale_fill_scico(palette = "imola", direction = -1)
# heatmap_col_scale <- scico::scale_colour_scico(palette = "imola", direction = -1)

# Future release candidate
# heatmap_fill_scale <- scico::scale_fill_scico(palette = "roma", direction = -1)
# heatmap_col_scale <- scico::scale_colour_scico(palette = "roma", direction = -1)

#' @rdname plot_region_heatmap
#'
#' @param window_prop the size of flanking region to plot. Can be a vector of two
#'   values for left and right window size. Values indicate proportion of region
#'   length.
#' @param pos_style the style for plotting the base positions along the x-axis.
#'   Defaults to "to_scale", plotting (potentially) overlapping squares
#'   along the genomic position to scale. The "compact" options plots only the
#'   positions with measured modification.
#' @param subsample the number of read of packed read rows to subsample to.
#'
#' @return a ggplot plot containing the heatmap.
#'
#' @details
#' This function creates a heatmap visualization of methylation data for a specified
#' genomic region. Each row represents one or more packed reads, with colored segments
#' indicating methylation probability at each position. Reads are packed vertically
#' to minimize plot height while avoiding overlaps.
#'
#' @examples
#' nmr <- load_example_nanomethresult()
#' plot_region_heatmap(nmr, "chr7", 6703892, 6730431)
#'
#' @export
setMethod(
    "plot_region_heatmap",
    signature(
        x = "NanoMethResult",
        chr = "character",
        start = "numeric",
        end = "numeric"),

    function(
        x,
        chr,
        start,
        end,
        pos_style = c("to_scale", "compact"),
        window_prop = 0,
        subsample = 50

    ) {
        pos_style <- match.arg(pos_style)

        plot_region_heatmap_impl(
            x = x,
            chr = chr,
            start = start,
            end = end,
            pos_style = pos_style,
            window_prop = window_prop,
            subsample = subsample
        )
    }
)

#' @rdname plot_region_heatmap
#'
#' @export
setMethod(
    "plot_region_heatmap",
    signature(
        x = "ModBamResult",
        chr = "character",
        start = "numeric",
        end = "numeric"),

    function(
        x,
        chr,
        start,
        end,
        pos_style = c("to_scale", "compact"),
        window_prop = 0,
        subsample = 50

    ) {
        pos_style <- match.arg(pos_style)

        plot_region_heatmap_impl(
            x = x,
            chr = chr,
            start = start,
            end = end,
            pos_style = pos_style,
            window_prop = window_prop,
            subsample = subsample
        )
    }
)

#' @rdname plot_region_heatmap
#'
#' @export
setMethod("plot_region_heatmap",
    signature(
        x = "NanoMethResult",
        chr = "factor",
        start = "numeric",
        end = "numeric"),

    function(
        x,
        chr,
        start,
        end,
        pos_style = c("to_scale", "compact"),
        window_prop = 0,
        subsample = 50
    ) {
        chr <- as.character(chr)
        plot_region_heatmap_impl(
            x = x,
            chr = chr,
            start = start,
            end = end,
            pos_style = pos_style,
            window_prop = window_prop,
            subsample = subsample
        )
    }
)

#' @rdname plot_region_heatmap
#'
#' @export
setMethod("plot_region_heatmap",
    signature(
        x = "ModBamResult",
        chr = "factor",
        start = "numeric",
        end = "numeric"),

    function(
        x,
        chr,
        start,
        end,
        pos_style = c("to_scale", "compact"),
        window_prop = 0,
        subsample = 50
    ) {
        chr <- as.character(chr)
        plot_region_heatmap_impl(
            x = x,
            chr = chr,
            start = start,
            end = end,
            pos_style = pos_style,
            window_prop = window_prop,
            subsample = subsample
        )
    }
)

plot_region_heatmap_impl <- function(
    x,
    chr,
    start,
    end,
    window_prop,
    pos_style,
    subsample
) {
    if (length(window_prop) == 1) {
        window_prop <- c(window_prop, window_prop)
    }

    window_left <- (end - start) * window_prop[1]
    window_right <- (end - start) * window_prop[2]

    plot_left <- round(start - window_left)
    plot_right <- round(end + window_right)

    methy_data <- query_methy(x, chr, plot_left, plot_right)

    # add sample information
    methy_data <- dplyr::inner_join(
        NanoMethViz::samples(x),
        methy_data,
        by = "sample",
        multiple = "all"
    )

    read_data <- methy_data %>%
        dplyr::group_by(.data$read_name) %>%
        dplyr::summarise(start = min(.data$pos), end = max(.data$pos))

    group_data <- methy_data %>%
            dplyr::group_by(.data$read_name) %>%
            dplyr::summarise(group = unique(.data$group))

    # get read starts and ends
    read_data <- dplyr::inner_join(read_data, group_data, by = "read_name", multiple = "all")

    # get grouping indices to pack reads
    append_read_group <- function(x, k) {
        x$read_group <- paste0(k, stacked_interval_inds(x))
        x
    }
    read_data <- read_data %>%
        dplyr::group_by(.data$group) %>%
        dplyr::group_modify(append_read_group) %>%
        dplyr::ungroup()

    methy_data$mod_prob <- sigmoid(methy_data$statistic)

    methy_data <- dplyr::inner_join(
        methy_data,
        dplyr::select(read_data, "read_name", "read_group"),
        by = "read_name",
        multiple = "all"
    )

    plot_methy_data_heatmap(
        methy_data = methy_data,
        pos_style = pos_style,
        subsample = subsample
    )
}

# plot methylation data queried using NanoMethViz
plot_methylation_data <- function(
    methy_data,
    sample_anno,
    chr,
    start,
    end,
    title,
    read_anno = NULL,
    group_col = "group",
    palette_col = ggplot2::scale_colour_brewer(palette = "Set1"),
    anno_regions = NULL,
    binary_threshold = NULL,
    avg_method = c("mean", "median"),
    spaghetti = FALSE,
    points = FALSE,
    smoothing_window = 2000,
    highlight_col = getOption("NanoMethViz.highlight_col", "grey50"),
    line_size = 1,
    mod_scale = c(0, 1),
    span = NULL
) {
    if (!missing("span")) {
        warning("the 'span' argument has been deprecated, please use 'smoothing_window' instead")
    }
    # assign averaging method
    avg_method <- match.arg(avg_method)
    avg_func <- switch(
        avg_method,
        mean = confidence_weighted_mean,
        median = median
    )

    # if there are regions to annotate
    if (!is.null(anno_regions)) {
        # filter annotation regions to be within plotting region
        anno_regions <- anno_regions %>%
            dplyr::filter(
                .data$chr == unique(methy_data$chr),
                .data$end >= min(methy_data$pos),
                .data$start <= max(methy_data$pos)
            )
    }

    plot_data <- methy_data %>%
        binarise_statistics(binary_threshold = binary_threshold) %>%
        join_samples_anno(sample_anno) %>%
        add_read_anno(read_anno)

    # set up plot
    # rug first so it appears on the bottom layer
    p <- ggplot(plot_data, aes(x = .data$pos, col = .data[[group_col]])) +
        ggplot2::geom_rug(aes(col = NULL), sides = "b")

    # add annotated regions
    if (!is.null(anno_regions)) {
        for (i in seq_len(nrow(anno_regions))) {
            region <- anno_regions[i, ]
            p <- p +
                ggplot2::annotate(
                    "rect",
                    xmin = region$start,
                    xmax = region$end,
                    ymin = -Inf,
                    ymax = Inf,
                    alpha = 0.2,
                    fill = highlight_col
                )
        }
    }

    # add points
    if (points) {
        p <- p +
            ggplot2::geom_point(
                aes(y = .data$mod_prob),
                alpha = 0.75
            )
    }

    # add spaghetti
    if (spaghetti) {
        p <- p +
            stat_lm(
                aes(y = .data$mod_prob, group = .data$read_name),
                alpha = 0.25,
                na.rm = TRUE
            )
    }

    # add smoothed line
    plot_data_smooth <- plot_data %>%
        dplyr::group_by(.data[[group_col]], .data$pos) %>%
        dplyr::summarise(
            mod_prob = avg_func(.data$mod_prob)
        ) %>%
        dplyr::mutate(
            mod_prob = rolling_average(
                .data$mod_prob,
                .data$pos,
                smoothing_window = smoothing_window
            )
        )

    p <- p +
        ggplot2::geom_line(
            aes(y = .data$mod_prob),
            data = plot_data_smooth,
            na.rm = TRUE,
            linewidth = line_size
        )

    # add auxiliary elements and style
    if (is.numeric(mod_scale) && length(mod_scale) == 2) {
        p <- p + ggplot2::ylim(mod_scale[1], mod_scale[2])
    }
    p +
        ggplot2::ggtitle(title) +
        ggplot2::xlab(chr) +
        palette_col +
        ggplot2::theme_bw() +
        ggplot2::theme(panel.grid.minor = ggplot2::element_blank()) +
        ggplot2::scale_x_continuous(labels = scales::label_number(scale_cut = scales::cut_si("b")))
}

binarise_statistics <- function(plot_data, binary_threshold) {
    if (!is.null(binary_threshold)) {
        # if binary threshold is provided, convert probabilities to binary values
        plot_data$mod_prob <- as.numeric(
            sigmoid(plot_data$statistic) > binary_threshold
        )
    }
    plot_data
}

join_samples_anno <- function(plot_data, sample_anno) {
    plot_data %>%
        dplyr::inner_join(sample_anno, by = "sample", multiple = "all")
}

add_read_anno <- function(plot_data, read_anno) {
    # incorporate read annotation if available
    if (!is.null(read_anno)) {
        # remove any duplicate columns and use plot_data as reference
        plot_data_names <- setdiff(names(plot_data), "read_name")
        read_anno <- dplyr::select(read_anno, -dplyr::any_of(plot_data_names))
        plot_data <- dplyr::inner_join(plot_data, read_anno, by = "read_name") %>%
            tidyr::drop_na()
    }

    plot_data
}

confidence_weighted_mean <- function(x, threshold = 0.5) {
    weights <- abs(x - threshold)^2
    weights <- weights / sum(weights)
    sum(x * weights)
}

rolling_average <- function(y, x, smoothing_window = 2000, neighbours = smoothing_window / 10) {
    tricube_kern <- function(x) {
        ifelse(abs(x) < 1, (1 - abs(x))^3, 0)
    }

    n <- length(y)
    y_smooth <- rep(NA, n)
    for (i in 1:n) {
        start <- max(1, i - neighbours)
        end <- min(n, i + neighbours)
        dists <- abs(x[start:end] - x[i])
        weights <- tricube_kern(dists / smoothing_window)
        weights <- weights / sum(weights)

        y_smooth[i] <- sum(y[start:end] * weights)
    }

    y_smooth
}

# Generated by using Rcpp::compileAttributes() -> do not edit by hand
# Generator token: 10BE3573-1514-4C36-9D1C-5A225CD40393

convert_methy_to_dss_cpp <- function(input, output_dir) {
    .Call('_NanoMethViz_convert_methy_to_dss_cpp', PACKAGE = 'NanoMethViz', input, output_dir)
}

count_cg_cpp <- function(str) {
    .Call('_NanoMethViz_count_cg_cpp', PACKAGE = 'NanoMethViz', str)
}

get_char_pos_cpp <- function(x, c) {
    .Call('_NanoMethViz_get_char_pos_cpp', PACKAGE = 'NanoMethViz', x, c)
}

cigar_tokeniser_cpp <- function(x) {
    .Call('_NanoMethViz_cigar_tokeniser_cpp', PACKAGE = 'NanoMethViz', x)
}

get_coord_map_cpp <- function(cigar) {
    .Call('_NanoMethViz_get_coord_map_cpp', PACKAGE = 'NanoMethViz', cigar)
}

mod_tokeniser_cpp <- function(string, scores) {
    .Call('_NanoMethViz_mod_tokeniser_cpp', PACKAGE = 'NanoMethViz', string, scores)
}

parse_bam_cpp <- function(seq, cigar, mm_string, ml_string, map_pos, strand, mod_code) {
    .Call('_NanoMethViz_parse_bam_cpp', PACKAGE = 'NanoMethViz', seq, cigar, mm_string, ml_string, map_pos, strand, mod_code)
}

parse_bam_list_cpp <- function(seq, cigar, mm_string, ml_string, map_pos, strand, mod_code) {
    .Call('_NanoMethViz_parse_bam_list_cpp', PACKAGE = 'NanoMethViz', seq, cigar, mm_string, ml_string, map_pos, strand, mod_code)
}

#' Calculate region methylation statistics
#'
#' Calculate the average methylation probability and prevalence based on
#' specified probability threshold.
#'
#' @param nmr the NanoMethResult object.
#' @param regions the table of regions to query statistics for.
#' @param threshold the threshold to use for determining methylation calls
#'   for the calculation of prevalence.
#'
#' @return table of regions with additional columns of methylation summary
#'   statistics.
#' @export
#'
#' @examples
#' nmr <- load_example_nanomethresult()
#' gene_anno <- exons_to_genes(NanoMethViz::exons(nmr))
#' region_methy_stats(nmr, gene_anno)
region_methy_stats <- function(nmr, regions, threshold = 0.5) {

    # calculate methylation statistics for each region using
    methy_stats <- purrr::map_df(
        seq_len(nrow(regions)),
        get_region_methy_stats,
        nmr = nmr,
        regions = regions,
        threshold = threshold
    )

    dplyr::bind_cols(regions, methy_stats)
}

get_region_methy_stats <- function(index, nmr, regions, threshold) {

    # query methylation data for the region
    chr <- regions$chr[index]
    start <- regions$start[index]
    end <- regions$end[index]
    methy <- query_methy(nmr, chr, start, end, force = TRUE)

    # calculate methylation statistics if methylation data is available
    if (nrow(methy) != 0) {
        methy %>%
            summarise(
                mean_methy_prob = mean(sigmoid(.data$statistic)),
                prevalence = mean(sigmoid(.data$statistic) > threshold)
            )
    } else {
        tibble(mean_methy_prob = NA_real_, prevalence = NA_real_)
    }
}

#' @title Convert BAM with modifications to tabix format
#'
#' @description The `modbam_to_tabix` function takes a ModBamResult object and
#'   converts it into a tabix file format, which is efficient for indexing and
#'   querying large datasets.
#'
#' @param x the `ModBamResult` object.
#' @param out_file the path of the output tabix.
#' @param mod_code the modification code to use, defaults to 'm' for 5mC
#'  methylation.
#'
#' @details
#' The possible tags for mod_code can be found at
#'  \url{https://samtools.github.io/hts-specs/SAMtags.pdf} under the
#'  'Base modifications' section.
#'
#' @return invisibly returns the name of the created tabix file.
#'
#' @examples
#' out_file <- paste0(tempfile(), ".tsv.bgz")
#' mbr <- ModBamResult(
#'     methy = ModBamFiles(
#'         samples = "sample1",
#'         paths = system.file("peg3.bam", package = "NanoMethViz",
#'         mustWork = FALSE)
#'     ),
#'     samples = data.frame(
#'         sample = "sample1",
#'         group = "group1"
#'     )
#' )
#'
#' modbam_to_tabix(mbr, out_file)
#'
#' @export
modbam_to_tabix <- function(x, out_file, mod_code = NanoMethViz::mod_code(x)) {
    assert_that(is(x, "ModBamResult"))

    assert_that(
        tools::file_ext(out_file) == "bgz",
        msg = "output_file must end with .bgz extension."
    )

    assert_that(
        !(fs::file_exists(out_file) && !fs::is_file(out_file)),
        msg = "output_file exists and is not a file"
    )

    if (fs::file_exists(out_file)) {
        cli::cli_alert_info(paste0("Output file exists, will overwrite ", out_file))
    }

    tmp_tsv_path <- base::tempfile(fileext = ".tsv")
    cli::cli_progress_step(paste0("Writing data to temporary file: ", tmp_tsv_path))

    cli::cli_progress_step("Converting data to TSV")
    tsv_file <- run_modbam_to_tsv_converter(x, tmp_tsv_path, mod_code)

    cli::cli_progress_step("Sorting data")
    sorted <- sort_methy_file(tsv_file)

    cli::cli_progress_step("Compressing data")
    tmp_out <- tabix_compress(sorted)

    cli::cli_alert_info(paste0("Moving data to final location: ", out_file))

    output_dir = fs::path_dir(out_file)
    if (output_dir != "" && fs::file_exists(output_dir)) {
        fs::dir_create(fs::path_dir(out_file))
    }

    fs::file_move(tmp_out, out_file)

    cli::cli_progress_step(paste0("Tabix file created: ", out_file))
    invisible(out_file)
}

run_modbam_to_tsv_converter <- function(x, out_file, mod_code) {
    # if .bgz at end of output name then trim it so final output
    # doesn't end with .bgz.bgz
    if (stringr::str_detect(out_file, ".bgz$")) {
        out_file <- out_file %>%
            stringr::str_remove(".bgz$")
    }

    bam_info <- dplyr::inner_join(samples(x), methy(x), by = dplyr::join_by(sample))

    n_files <- nrow(bam_info)
    for (i in seq_len(n_files)) {
        path <- bam_info$path[i]
        sample <- bam_info$sample[i]
        total <- get_bam_total_reads(path)
        fname <- fs::path_file(path)

        prog_bar_id <- cli::cli_progress_bar(
            glue::glue("Converting file {i}/{n_files}: {fname}"),
            total = total,
            format_done = paste0("{.alert-success Data converted: ", fname, " {.timestamp {cli::pb_elapsed}}}"),
            format_failed = paste0("{.alert-danger Data conversion failed: ", fname, " {.timestamp {cli::pb_elapsed}}}"),
            clear = FALSE
        )
        modbam_file_to_tsv(path, out_file, sample, mod_code, prog_bar_id)
    }

    out_file
}

modbam_file_to_tsv <- function(path, out_file, sample, mod_code, prog_bar_id = NULL) {
    parse_read_chunk <- function(x) {
        parse_modbam(x[[1]], sample, mod_code = mod_code) %>%
            select("sample", "chr", "pos", "strand", "statistic", "read_name")
    }

    bam_file <- Rsamtools::BamFile(path, yieldSize = 2000)
    open(bam_file)
    while (Rsamtools::isIncomplete(bam_file)) {
        reads <- read_bam(bam_file)

        if (!is.null(reads[[1]]) && length(reads[[1]]$qname) > 0) {
            # parse if valid data exists
            data <- parse_read_chunk(reads)
            readr::write_tsv(data, out_file, append = TRUE, progress = FALSE)
        }

        if (!is.null(prog_bar_id)) {
            n_reads <- length(reads[[1]][[1]])
            cli::cli_progress_update(n_reads, id = prog_bar_id)
        }
    }
    close(bam_file)
}

#' Get methylation data
#' @keywords internal
#' @param object the object.
#' @return the path to the methylation data.
#' @examples
#' showMethods("methy")
#'
#' @export
setGeneric("methy", valueClass = "ANY", function(object) {
    standardGeneric("methy")
})

#' Set methylation data
#' @keywords internal
#' @export
setGeneric("methy<-", function(object, value) {
    standardGeneric("methy<-")
})

#' Get sample annotation
#' @keywords internal
#' @export
setGeneric("samples", valueClass = "data.frame", function(object) {
    standardGeneric("samples")
})

#' Set sample annotation
#' @keywords internal
#' @export
setGeneric("samples<-", function(object, value) {
    standardGeneric("samples<-")
})

#' Get exon annotation
#' @keywords internal
#' @export
setGeneric("exons", valueClass = "data.frame", function(object) {
    standardGeneric("exons")
})

#' Set exon annotation
#' @keywords internal
#' @export
setGeneric("exons<-", function(object, value) {
    standardGeneric("exons<-")
})

#' Get mod code
#' @keywords internal
#' @export
setGeneric("mod_code", valueClass = "character", function(object) {
    standardGeneric("mod_code")
})

#' Set mod code
#' @keywords internal
#' @export
setGeneric("mod_code<-", function(object, value) {
    standardGeneric("mod_code<-")
})

#' Plot gene methylation
#'
#' Plot the methylation of a gene symbol specified within the exon(x) slot.
#'
#' @param x the NanoMethResult or ModBamResult object.
#' @param gene the gene symbol for the gene to plot.
#' @param ... additional arguments.
#'
#' @return a patchwork plot containing the methylation profile in the specified
#'   region.
#'
#' @importFrom ggrastr rasterise
#' @export
setGeneric("plot_gene", function(x, gene, ...) {
    standardGeneric("plot_gene")
})

#' Plot gene methylation heatmap
#'
#' Plot the methylation heatmap of a gene symbol specified within the exon(x) slot.
#'
#' @param x the NanoMethResult or ModBamResult object.
#' @param gene the gene symbol for the gene to plot.
#' @param ... additional arguments.
#'
#' @return a ggplot object of the heatmap
#'
#' @export
setGeneric("plot_gene_heatmap", function(x, gene, ...) {
    standardGeneric("plot_gene_heatmap")
})

#' Plot region methylation
#'
#' Plot the methylation of a genomic region.
#'
#' @param x the NanoMethResult or ModBamResult object.
#' @param chr the chromosome to plot.
#' @param start the start of the plotting region.
#' @param end the end of the plotting region.
#' @param ... additional arguments.
#'
#' @return a patchwork plot containing the methylation profile in the specified
#'   region.
#'
#' @importFrom ggrastr rasterise
#' @export
setGeneric("plot_region", function(x, chr, start, end, ...) {
    standardGeneric("plot_region")
})

#' Plot region methylation heatmap
#'
#' Plot the methylation heatmap of a genomic region.
#'
#' @param x the NanoMethResult or ModBamResult object.
#' @param chr the chromosome to plot.
#' @param start the start of the plotting region.
#' @param end the end of the plotting region.
#' @param ... additional arguments.
#'
#' @return a ggplot object of the heatmap.
#'
#' @export
setGeneric("plot_region_heatmap", function(x, chr, start, end, ...) {
    standardGeneric("plot_region_heatmap")
})

stacked_intervals <- function(reads) {
    # record original order of reads
    reads$old_order <- seq_len(nrow(reads))

    # sort by start and assign index number
    reads <- dplyr::arrange(reads, .data$start)
    reads$index <- seq_len(nrow(reads))

    # set initial groups as one per read
    reads$group <- seq_len(nrow(reads))

    # initialize 'merged' as an empty numeric vector
    merged <- numeric()

    # loop over rows
    for (i in seq_len(nrow(reads))) {
        # skip if row is already merged
        if (i %in% merged) {
            next
        }

        # set group end to 'end' of current row
        curr_end <- reads$end[i]

        # find rows for merging
        candidates <- reads[!reads$index %in% merged & reads$start > curr_end, ]

        # while candidates exist, merge left-most candidate into group
        while (nrow(candidates) > 0) {
            # merge first candidate into group
            merged <- c(merged, candidates$index[1])
            reads$group[candidates$index[1]] <- reads$group[i]

            # update end position of current group
            curr_end <- candidates$end[1]

            # find new candidates
            candidates <- reads[!reads$index %in% merged & reads$start > curr_end, ]
        }

        # add current read into list of merged reads
        merged <- c(merged, i)
    }

    reads$group <- factor(reads$group)
    reads
}

stacked_interval_inds <-  function(reads) {
    stacked_intervals(reads) %>%
        dplyr::arrange(.data$old_order) %>%
        dplyr::pull(.data$group)
}

test_that("Stacking works", {
    # setup
    nmr <- load_example_nanomethresult()
    methy <- query_methy_gene(nmr, gene = "Peg3")
    reads <- methy %>%
        dplyr::group_by(read_name) %>%
        dplyr::summarise(start = min(pos), end = max(pos))

    stacked_intervals <- stacked_intervals(reads)
    stacked_interval_inds <- stacked_interval_inds(reads)

    # test
    expect_equal(nrow(stacked_intervals), nrow(reads))
    expect_lte(dplyr::n_distinct(stacked_intervals$group), nrow(reads))
    expect_s3_class(stacked_interval_inds, "factor")

    # assertion 1: the number of unique groups should be less than or equal to the number of rows
    expect_true(length(unique(stacked_intervals$group)) <= nrow(reads),
        info = "the number of unique groups should be less than or equal to the number of rows")

    # assertion 2: reads within group should not overlap
    for (i in unique(stacked_intervals$group)) {
        group_reads <- stacked_intervals %>%
            dplyr::filter(group == i) %>%
            dplyr::arrange(start)

        if (nrow(group_reads) == 1) {
            next
        }

        for (j in 1:(nrow(group_reads) - 1)) {
            expect_true(group_reads$end[j] < group_reads$start[j + 1],
                info = glue::glue("no reads overlap within each group: {group_reads$end[j]} > {group_reads$start[j + 1]}"))
        }
    }
})

temp_path <- tempfile()
download.file("https://s3-us-west-2.amazonaws.com/human-pangenomics/T2T/CHM13/assemblies/annotation/chm13v2.0_RefSeq_Liftoff_v5.2.gff3.gz", temp_path)

gff_col_names <- c(
    "chr",
    "source",
    "type",
    "start",
    "end",
    "score",
    "strand",
    "phase",
    "attributes"
)

anno_t2t <- read_tsv(
    temp_path,
    col_names = gff_col_names,
    comment = "#"
)

anno_t2t <- read_tsv(
    temp_path,
    col_names = gff_col_names,
    comment = "#"
)

exon_anno_t2t <- anno_t2t %>%
    filter(type == "exon") %>%
    mutate(
        transcript_id = str_extract(attributes, "Parent=(\\w+)", group = 1),
        gene_id = str_extract(attributes, "GeneID:(\\w+)", group = 1),
        symbol = str_extract(attributes, "gene=(\\w+)", group = 1)
    )

exon_anno_t2t_formatted <- exon_anno_t2t %>%
    select(
        gene_id,
        chr,
        strand,
        start,
        end,
        transcript_id,
        symbol
    ) %>%
    mutate(
        gene_id = as.character(gene_id),
        chr = as.character(chr),
        strand = as.character(strand),
        start = as.integer(start),
        end = as.integer(end),
        transcript_id = as.character(transcript_id),
        symbol = as.character(symbol)
    ) %>%
    drop_na()

anno_name <- paste0("inst/exons_t2t.rds")
saveRDS(exon_anno_t2t_formatted, anno_name, compress = "xz")

fs::file_delete(temp_path)

#' @importFrom Biostrings width
# nocov start
get_read_entropy <- function(bam_path, sample = fs::path_file(bam_path)) {
    # helper functions ----
    count_crossings <- function(numbers) {
        crossings <- sum(((numbers[-1] - 127.5) * (numbers[-length(numbers)] - 127.5)) < 0)
        return(crossings)
    }

    parse_chunk <- function(reads, sample) {
        reads <- reads[!is.null(reads$seq)]
        tibble::tibble(
            sample = factor(sample),
            read_name = reads$qname,
            chr = factor(reads$rname),
            pos = reads$pos,
            width = Biostrings::width(reads$seq),
            center = as.integer(.data$pos + .data$width / 2),
            cg_count = purrr::map_int(as.character(reads$seq), count_cg_cpp),
            crossings = purrr::map_int(reads$tag$ML, count_crossings),
            entropy = .data$crossings / .data$cg_count
        )
    }

    # function body ----
    fname <- fs::path_file(bam_path)
        cli::cli_progress_bar(
            glue::glue("Parsing file: {fname}"),
            total = get_bam_total_reads(bam_path),
            format_done = paste0(
                "{.alert-success Data parsed: ", fname, " {.timestamp {cli::pb_elapsed}}}"),
            format_failed = paste0(
                "{.alert-danger Data parsing failed: ", fname, " {.timestamp {cli::pb_elapsed}}}"),
            clear = FALSE
        )

    bam_file <- Rsamtools::BamFile(bam_path, yieldSize = 15000)

    i <- 1
    df_list <- list()
    open(bam_file)
    while (Rsamtools::isIncomplete(bam_file)) {
        reads <- read_bam(bam_file)[[1]]
        df_list[[i]] <- parse_chunk(reads, sample)
        i <- i + 1
        cli::cli_progress_update(length(reads[[1]]))
    }
    close(bam_file)

    bind_rows(df_list)
}
# nocov end

#' Create a tabix file using methylation calls
#'
#' @param input_files the files to convert
#' @param output_file the output file to write results to (must end in .bgz)
#' @param samples the names of samples corresponding to each file
#' @param verbose TRUE if progress messages are to be printed
#'
#' @return invisibly returns the output file path, creates a tabix file (.bgz)
#'   and its index (.bgz.tbi)
#'
#' @export
#' @examples
#' methy_calls <- system.file(package = "NanoMethViz",
#'     c("sample1_nanopolish.tsv.gz", "sample2_nanopolish.tsv.gz"), mustWork = FALSE)
#' temp_file <- paste0(tempfile(), ".tsv.bgz")
#'
#' create_tabix_file(methy_calls, temp_file)
create_tabix_file <- function(
    input_files,
    output_file,
    samples = extract_file_names(input_files),
    verbose = TRUE
) {
    assert_that(
        tools::file_ext(output_file) == "bgz",
        msg = "output_file must end with .bgz extension."
    )
    assert_that(
        assertthat::not_empty(input_files),
        is.character(input_files),
        is.string(output_file),
        is.character(samples),
        length(input_files) == length(samples)
    )

    if (.Platform$OS.type == "windows") {
        timed_log("WARNING: creating tabix file on windows requires at least twice as much memory as total size of methylation data")
    }

    if (packageVersion("cpp11") < package_version("0.2.5")) {
        warning("cpp11 versions < 0.2.5 may crash when reading in large tables")
    }

    temp_file <- tempfile()
    if (verbose) {
        timed_log("creating methylation table")
    }
    convert_methy_format(input_files, temp_file, samples, verbose = verbose)

    if (verbose) {
        timed_log("sorting methylation table")
    }
    sort_methy_file(temp_file)

    if (verbose) {
        timed_log("compressing methylation table to tabix with index")
    }
    raw_methy_to_tabix(temp_file)

    fs::file_move(paste0(temp_file, ".bgz"), output_file)
    fs::file_move(paste0(temp_file, ".bgz.tbi"), paste0(output_file, ".tbi"))
    fs::file_delete(temp_file)

    invisible(output_file)
}

plot_region_impl <- function(
        x,
        chr,
        start,
        end,
        anno_regions = NULL,
        binary_threshold = NULL,
        avg_method = c("mean", "median"),
        spaghetti = FALSE,
        heatmap = TRUE,
        heatmap_subsample = 50,
        smoothing_window = 2000,
        gene_anno = TRUE,
        window_prop = 0,
        palette = ggplot2::scale_colour_brewer(palette = "Set1"),
        line_size = 1,
        mod_scale = c(0, 1),
        span = NULL
) {
    # validate inputs
    chr <- as.character(chr)
    start <- as.numeric(start)
    end <- as.numeric(end)

    # validate genomic coordinates
    assert_valid_genomic_coords(chr, start, end)

    sample_anno <- samples(x)

    if (length(window_prop) == 1) {
        window_prop <- c(window_prop, window_prop)
    }

    feature_width <- end - start
    window_left <- feature_width * window_prop[1]
    window_right <- feature_width * window_prop[2]

    # query data
    methy_data <- query_methy(
        x,
        chr,
        floor(start - window_left * 1.1),
        ceiling(end + window_right * 1.1),
        simplify = TRUE
    )

    if (nrow(methy_data) == 0) {
        warning(glue::glue(
            "No methylation data found in region {chr}:{start}-{end}.\n",
            "This could be due to:\n",
            "  - No data in this genomic region\n",
            "  - Incorrect chromosome naming (try 'chr{chr}' if using '{chr}')\n",
            "  - Region outside of data coverage\n",
            "Returning empty plot."
        ))
        return(ggplot() + theme_void())
    }

    methy_data <- methy_data %>%
        dplyr::select(-"strand") %>%
        tibble::as_tibble()


    # setup base plot
    title <- glue::glue("{chr}:{start}-{end}")
    xlim <- round(c(start - window_left, end + window_right))
    p1 <- plot_methylation_data(
            methy_data = methy_data,
            start = start,
            end = end,
            chr = chr,
            title = title,
            anno_regions = anno_regions,
            binary_threshold = binary_threshold,
            avg_method = avg_method,
            spaghetti = spaghetti,
            sample_anno = sample_anno,
            smoothing_window = smoothing_window,
            palette_col = palette,
            line_size = line_size,
            mod_scale = mod_scale
        ) +
        ggplot2::coord_cartesian(xlim = xlim, expand = FALSE) +
        ggplot2::labs(x = "Position", y = "Mean Modification Probability")

    p_out <- p1

    # if exon anno exists, append it to plot
    if (gene_anno && nrow(exons(x)) != 0) {
        exons_anno <- query_exons_region(x, chr = chr, start = start, end = end)

        p2 <- plot_gene_annotation(exons_anno, xlim[1], xlim[2]) +
            ggplot2::coord_cartesian(xlim = xlim, expand = FALSE) +
            ggplot2::scale_x_continuous(labels = scales::label_number(scale_cut = scales::cut_si("b")))

        anno_height <- attr(p2, "plot_height")
        heights <- c(1, 0.075 * anno_height)

        p_out <- p1 / p2 + patchwork::plot_layout(heights = heights)
    }

    # if heatmap requested, append it to plot
    if (heatmap) {
        p_heatmap <- plot_region_heatmap(x, chr, start, end, window_prop = window_prop, subsample = heatmap_subsample) +
            ggplot2::coord_cartesian(
                xlim = xlim
            )

        p_out <- stack_plots(p_out, ggrastr::rasterise(p_heatmap, dpi = 300))
    }

    p_out
}

#' @rdname plot_region
#'
#' @param anno_regions the data.frame of regions to be annotated.
#' @param binary_threshold the modification probability such that calls with
#'   modification probability above the threshold are set to 1 and probabilities
#'   equal to or below the threshold are set to 0.
#' @param avg_method the average method for pre-smoothing at each genomic position.
#'   Data is pre-smoothed at each genomic position before the smoothed aggregate line
#'   is generated for performance reasons. The default is "mean" which corresponds
#'   to the average methylation fraction. The alternative "median" option is
#'   closer to an average within the more common methylation state.
#' @param spaghetti whether or not individual reads should be shown.
#' @param heatmap whether or not read-methylation heatmap should be shown.
#' @param heatmap_subsample how many packed rows of reads to subsample to.
#' @param smoothing_window the window size for smoothing the trend line.
#' @param gene_anno whether to show gene annotation.
#' @param window_prop the size of flanking region to plot. Can be a vector of two
#'   values for left and right window size. Values indicate proportion of gene
#'   length.
#' @param palette the ggplot colour palette used for groups.
#' @param line_size the size of the lines.
#' @param mod_scale the scale range for modification probabilities. Default c(0, 1), set to "auto" for automatic
#'   limits.
#' @param span DEPRECATED, use smoothing_window instead. Will be removed in next version.
#'
#' @details
#' This function plots the methylation data for a given region. The main trendline plot shows the average methylation
#' probability across the region. The heatmap plot shows the methylation probability for each read across the region.
#' The gene annotation plot shows the exons of the region. In the heatmap, each row represents one or more
#' non-overlapping reads where the coloured segments represent the methylation probability at each position. Data along
#' a read is connected by a grey line. The gene annotation plot shows the isoforms and exons of genes within the region,
#' with arrows indicating the direction of transcription.
#'
#' Since V3.0.0 NanoMethViz has changed the smoothing strategy from a loess smoothing to a weighted moving average. This
#' is because the loess smoothing was too computationally expensive for large datasets and had a span parameter that was
#' difficult to tune. The new smoothing strategy is controlled by the smoothing_window argument.
#'
#' @examples
#' nmr <- load_example_nanomethresult()
#' plot_region(nmr, "chr7", 6703892, 6730431)
#'
#' @export
setMethod("plot_region",
    signature(x = "NanoMethResult", chr = "character", start = "numeric", end = "numeric"),
    plot_region_impl
)

#' @rdname plot_region
#' @export
setMethod("plot_region",
    signature(x = "ModBamResult", chr = "character", start = "numeric", end = "numeric"),
    plot_region_impl
)

#' @rdname plot_region
#' @export
setMethod("plot_region",
    signature(x = "NanoMethResult", chr = "factor", start = "numeric", end = "numeric"),
    plot_region_impl
)

#' @rdname plot_region
#' @export
setMethod("plot_region",
    signature(x = "ModBamResult", chr = "factor", start = "numeric", end = "numeric"),
    plot_region_impl
)

test_that("CGI getters work", {
    cgi_getters <- list(
        get_cgi_hg19,
        get_cgi_hg38,
        get_cgi_mm10,
        get_cgi_grcm39
    )

    for (cgi_fn in cgi_getters) {
        cgi_anno <- expect_no_warning(cgi_fn())
        expect_s3_class(cgi_anno, "data.frame")
        expect_gt(nrow(cgi_anno), 0)
        expect_gt(ncol(cgi_anno), 0)
        expect_contains(colnames(cgi_anno), c("gene_id", "chr", "strand", "start", "end", "transcript_id", "symbol"))
    }
})

---
title: "User's Guide"
author: "Shian Su"
output:
    BiocStyle::html_document:
        toc: true
vignette: >
  %\VignetteIndexEntry{User's Guide}
  %\VignetteEngine{knitr::rmarkdown}
  \usepackage[utf8]{inputenc}
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(echo = TRUE)
knitr::opts_chunk$set(fig.width = 7)
knitr::opts_chunk$set(fig.height = 5)
knitr::opts_chunk$set(dpi = 72)

# preload to avoid loading messages
library(NanoMethViz)
library(dplyr)
exon_tibble <- get_exons_mm10()
```

# Introduction

The `NanoMethViz` package is designed to visualise methylation data from long-read sequencing. It provides a simple interface to plot methylation data, and to perform dimensionality reduction on the data. The package is designed to work with methylation data on a per-read basis.

We provide functions for converting to various formats for downstream analysis, and visualising methylation data at various levels of resolution. Features are all developed to work with 5mC CpG methylation data, but it will work with other modifications if the data is in the correct format, as the type of modification is not generally specified in the data.

This package is designed to work with methylation data from the Oxford Nanopore Technologies (ONT) platform. Currently it supports data from the following software:

- dorado (modBAM)
- modkit (tsv generated by `modkit extract full`)
- Megalodon (tsv)
- Nanopolish (tsv)
- f5c (tsv)

It is possible to use data from other software as long as it matches the format of the data used in this package. If you would like any further other formats supported please create an issue at https://github.com/Shians/NanoMethViz/issues.

# Installation

To install this package, run

```{r, eval = FALSE}
if (!requireNamespace("BiocManager", quietly = TRUE))
    install.packages("BiocManager")

BiocManager::install("NanoMethViz")
```

```{r}
library(NanoMethViz)
library(dplyr)
```

# Importing data

In order to use this package, your data must be converted from the output of methylation calling software to a tabix-indexed bgzipped format. The data needs to be sorted by genomic position to respect the requirements of the samtools [tabix](http://www.htslib.org/doc/tabix.html) indexing tool. On Linux and macOS systems this is done using the bash `sort` utility, which is memory efficient, but on Windows this is done by loading the entire table and sorting within R.

## modBAM data

In the latest versions of `dorado` and other ONT software, the modifications are stored in the BAM file. This can be used directly with `NanoMethViz` using the `ModBamResult` class as a substitute for `NanoMethResult`.

This requires a `ModBamFiles` object, which is a list of paths to the BAM files and the sample names. The sample names must match the sample names in the `sample_anno` data frame.

Because the BAM files contain significantly more data than the tabix files, it may be useful to convert the BAM files to tabix files using the  `modbam_to_tabix()` function for sharing the data.

In modBAM files the modification calls are made along the read, allowing sequencing errors to produce modifications calls where there is none in the genome. This may lead to noise impacting the averaging results. The `NanoMethViz.site_filter` option can be used to remove sites with low coverage, by default it is set to 3 to remove any sites with coverage less than 3. See the "Site filtering" section for more information.

```{r}
# construct with a ModBamFiles object as the methylation data
mbr <- ModBamResult(
    methy = ModBamFiles(
        samples = "sample1",
        system.file(package = "NanoMethViz", "peg3.bam")
    ),
    samples = data.frame(
        sample = "sample1",
        group = 1
    ),
    exons = exon_tibble
)

# use in the same way as you would a NanoMethResult object
plot_gene(mbr, "Peg3", heatmap = TRUE)
```


## Tabular data (Modkit, Megalodon, Nanopolish, f5c)

Tabular data is the most common format for methylation data, and is the format used by most of the software. The columns in the tabular data varies according to the software used. `NanoMethViz` will attempt to automatically detect the format of the data.

### Modkit

Modkit can be used to extract site-level modification data modBAM files. The `modkit extract full` command will output a tab-separated file with the required columns that can be converted useing the `create_tabix_file()` function.

NOTE: `modkit` BED files are not supported as they do not contain read-level information, only position-aggregated data.

## Megalodon

To import data from Megalodon's modification calls, the [per-read modified bases](https://nanoporetech.github.io/megalodon/file_formats.html#per-read-modified-bases) file must be generated. This can be done by either adding `--write-mods-text` argument to Megalodon run or using the `megalodon_extras per_read_text modified_bases` utility.

Later version of Megalodon will also produce `modBAM` files, which can be used directly with `NanoMethViz` using the `ModBamResult` class as a substitute for `NanoMethResult`.

## Importing to tabix format

Once you have produce the correct output from the software listed above, the conversion can be done using the `create_tabix_file()` function. We provide example data of nanopolish output within the package.

```{r}
methy_calls <- system.file(package = "NanoMethViz",
    c("sample1_nanopolish.tsv.gz", "sample2_nanopolish.tsv.gz"))
```

We then create a temporary path to store a converted file, this will be deleted once you exit your R session. Once `create_tabix_file()` is run, it will create a .bgz file along with its tabix index. Because we have a small amount of data, we can read in a small portion of it for inspection, do not do this with large datasets as it decompresses all the data and will take very long to run.

## Importing data

```{r, message=F}
# create a temporary file to store the converted data
methy_tabix <- file.path(tempdir(), "methy_data.bgz")
samples <- c("sample1", "sample2")

# you should see messages when running this yourself
create_tabix_file(methy_calls, methy_tabix, samples)

# don't do this with full datasets, it will take a long time
# we have to use gzfile to tell R that we have a gzip compressed file
methy_data <- read.table(
    gzfile(methy_tabix), col.names = methy_col_names(), nrows = 6)

methy_data
```

Now `methy_tabix` will be the path to a tabix object that is ready for use with NanoMethViz.

# Quick start

## Constructing the NanoMethResult object

To generate a methylation plot we need 3 components:

* methylation data in tabix format
* annotation of exons
* annotation of samples

The methylation information has been modified from the output of nanopolish/f5c. It has then been compressed and indexed using `bgzip()` and `indexTabix()` from the `Rsamtools` package. Here `system_file()` is used to retrieve data internal to the package, for your down data you simply need to provide a path without the `system_file()` function.


```{r}
# methylation data stored in tabix file
methy <- system.file(package = "NanoMethViz", "methy_subset.tsv.bgz")

# tabix is just a special gzipped tab-separated-values file
read.table(gzfile(methy), col.names = methy_col_names(), nrows = 6)
```

The exon annotation was obtained from the `Mus.musculus` package, and joined into a single table. It is important that the chromosomes share the same convention as that found in the methylation data.

```{r}
# helper function extracts exons from TxDb package
exon_tibble <- get_exons_mm10()

head(exon_tibble)
```

We will defined the sample annotation ourselves. It is important that the sample names match those found in the methylation data.

```{r}
sample <- c(
  "B6Cast_Prom_1_bl6", "B6Cast_Prom_1_cast",
  "B6Cast_Prom_2_bl6", "B6Cast_Prom_2_cast",
  "B6Cast_Prom_3_bl6", "B6Cast_Prom_3_cast"
)

group <- c("bl6", "cast", "bl6", "cast", "bl6", "cast")

sample_anno <- data.frame(sample, group, stringsAsFactors = FALSE)

sample_anno
```

For convenience we assemble these three pieces of data into a single object.

```{r}
nmr <- NanoMethResult(methy, sample_anno, exon_tibble)
```

The genes we have available are

* Peg3
* Meg3
* Impact
* Xist
* Brca1
* Brca2

## Basic Plotting

For demonstrative purposes we will plot Peg3. The plot on top shows the average methylation probabilities across all samples for each experimental condition. The heatmap below it shows the methylation probabilities along individual reads, with disjoint reads stacked onto the same row. The annotation down the bottom shows the isoforms of the Peg3 gene as well as its directionality.

```{r}
plot_gene(nmr, "Peg3")
```

We can also load in some DMR results to highlight DMR regions.

```{r}
# loading saved results from previous bsseq analysis
bsseq_dmr <- read.table(
    system.file(package = "NanoMethViz", "dmr_subset.tsv.gz"),
    sep = "\t",
    header = TRUE,
    stringsAsFactors = FALSE
)
```

```{r}
plot_gene(nmr, "Peg3", anno_regions = bsseq_dmr)
```

See other plotting options in the [package documentation](https://shians.github.io/NanoMethViz/reference/index.html).

# Exporting data

The methylation data can be exported into formats appropriate for bsseq, DSS, or edgeR.

## bsseq and DSS

Both bsseq and DSS make use of the BSSeq object, and these can be obtained from the NanoMethResult objects using the `methy_to_bsseq()` function.

```{r, message = FALSE}
nmr <- load_example_nanomethresult()
bss <- methy_to_bsseq(nmr)

bss
```

## edgeR

edgeR can also be used to perform differential methylation analysis: https://f1000research.com/articles/6-2055. BSSeq objects can be converted into the appropriate format using the `bsseq_to_edger()` function. This can be used to count reads on a per-site basis or over regions.

```{r}
gene_regions <- exons_to_genes(NanoMethViz::exons(nmr))
edger_mat <- bsseq_to_edger(bss, gene_regions)

edger_mat
```

# Importing Annotations

This package comes with helper functions that import exon annotations from the Bioconductor packages `Homo.sapiens` and `Mus.musculus`. The functions `get_exons_hg19()` and `get_exons_mm10()` simply take data from the respective packages, and reorganise the columns into the following columns:

-   gene_id
-   chr
-   strand
-   start
-   end
-   transcript_id
-   symbol

This is used to provide gene annotations for the gene or region plots.

For other annotations, they will most likely be able to be imported using `rtracklayer::import()` and manipulated into the desired format. As an example, we can use a small sample of the C. Elegans gene annotation provided by ENSEMBL. `rtracklayer` will import the annotation as a `GRanges` object, this can be coerced into a data.frame and manipulated using `dplyr`.

```{r}
anno <- rtracklayer::import(system.file(package = "NanoMethViz", "c_elegans.gtf.gz"))

head(anno)
```

```{r}
anno <- anno %>%
    as.data.frame() %>%
    dplyr::rename(
        chr = "seqnames",
        symbol = "gene_name"
    ) %>%
    dplyr::select("gene_id", "chr", "strand", "start", "end", "transcript_id", "symbol")

head(anno)
```

## Importing Annotations

Annotations can be simplified if full exon and isoform information is not required. For example, gene-body annotation can be represented as single exon genes. For example we can take the example dataset and transform the isoform annotations of Peg3 into a single gene-body block. The helper function `exons_to_genes()` can help with this common conversion.

```{r, message = FALSE}
nmr <- load_example_nanomethresult()

plot_gene(nmr, "Peg3")
```

```{r}
new_exons <- NanoMethViz::exons(nmr) %>%
    exons_to_genes() %>%
    mutate(transcript_id = gene_id)

NanoMethViz::exons(nmr) <- new_exons

plot_gene(nmr, "Peg3")
```

# Dimensionality reduction

Dimensionality reduction is used to represent high dimensional data in a more tractable form. It is commonly used in RNA-seq analysis, where each sample is characterised by tens of thousands of gene expression values, to visualise samples in a 2D plane with distances between points representing similarity and dissimilarity. For RNA-seq the data used is generally gene counts, for methylation there are generally two relevant count matrices, the count of methylated bases, and the count of methylated bases. The information from these two matrices can be combined by taking log-methylation ratios as done in Chen et al. 2018.

### Preparing data for dimensionality reduction

It is assumed that users of this package have imported the data into the gzipped tabix format. From there, further processing is required to create the log-methylation-ratio matrix used in dimensionality reduction. Namely we go through the BSseq format as it is easily coerced into the desired matrix and is itself useful for various other analyses.

```{r, message = FALSE}
# convert to bsseq
bss <- methy_to_bsseq(nmr)
bss
```

We can generate the log-methylation-ratio based on individual methylation sites or computed over genes, or other feature types. Aggregating over features will generally provide more stable and robust results, here we will use genes.

```{r}
# create gene annotation from exon annotation
gene_anno <- exons_to_genes(NanoMethViz::exons(nmr))

# create log-methylation-ratio matrix
lmr <- bsseq_to_log_methy_ratio(bss, regions = gene_anno)
```

NanoMethViz currently provides two options, a MDS plot based on the limma implementation of MDS, and a PCA plot using BiocSingular.

```{r}
plot_mds(lmr) +
    ggtitle("MDS Plot")

plot_pca(lmr) +
    ggtitle("PCA Plot")
```

Additional coloring and labeling options can be provided via arguments to either function. Further customisations can be done using typical ggplot2 commands.

```{r}
new_labels <- gsub("B6Cast_Prom_", "", colnames(lmr))
new_labels <- gsub("(\\d)_(.*)", "\\2 \\1", new_labels)
groups <- gsub(" \\d", "", new_labels)

plot_mds(lmr, labels = new_labels, groups = groups) +
    ggtitle("MDS Plot") +
    scale_colour_brewer(palette = "Set1")
```

Points can also be plotted without labels by setting `labels = NULL`.

```{r}
plot_mds(lmr, labels = NULL, groups = groups) +
    ggtitle("MDS Plot") +
    scale_colour_brewer(palette = "Set1")
```

# Package options

## Site filtering

NanoMethViz offers a site filtering option to remove sites with low coverage. This is particularly useful for modBAM files as the modifications calls are made along the reads, allowing methylation to be called at miscalled sites. The site filter is set to 3 by default, to filter any sites with coverage less than 3. This can be changed using the option `NanoMethViz.site_filter`. The following will remove any sites with coverage less than 5 from queries and plots.

```{r, eval = FALSE}
options("NanoMethViz.site_filter" = 5)
```

## Region annotation colours

By default the `anno_regions` argument in plotting functions will create bands coloured grey, specifically `grey50`. This can be changed globally across the package using the option `NanoMethViz.highlight_col`. For example the following will change the colour to red.

```{r, eval = FALSE}
options("NanoMethViz.highlight_col" = "red")
```

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/AllGenerics.R, R/plot_region_heatmap.R
\name{plot_region_heatmap}
\alias{plot_region_heatmap}
\alias{plot_region_heatmap,NanoMethResult,character,numeric,numeric-method}
\alias{plot_region_heatmap,ModBamResult,character,numeric,numeric-method}
\alias{plot_region_heatmap,NanoMethResult,factor,numeric,numeric-method}
\alias{plot_region_heatmap,ModBamResult,factor,numeric,numeric-method}
\title{Plot region methylation heatmap}
\usage{
plot_region_heatmap(x, chr, start, end, ...)

\S4method{plot_region_heatmap}{NanoMethResult,character,numeric,numeric}(
  x,
  chr,
  start,
  end,
  pos_style = c("to_scale", "compact"),
  window_prop = 0,
  subsample = 50
)

\S4method{plot_region_heatmap}{ModBamResult,character,numeric,numeric}(
  x,
  chr,
  start,
  end,
  pos_style = c("to_scale", "compact"),
  window_prop = 0,
  subsample = 50
)

\S4method{plot_region_heatmap}{NanoMethResult,factor,numeric,numeric}(
  x,
  chr,
  start,
  end,
  pos_style = c("to_scale", "compact"),
  window_prop = 0,
  subsample = 50
)

\S4method{plot_region_heatmap}{ModBamResult,factor,numeric,numeric}(
  x,
  chr,
  start,
  end,
  pos_style = c("to_scale", "compact"),
  window_prop = 0,
  subsample = 50
)
}
\arguments{
\item{x}{the NanoMethResult or ModBamResult object.}

\item{chr}{the chromosome to plot.}

\item{start}{the start of the plotting region.}

\item{end}{the end of the plotting region.}

\item{...}{additional arguments.}

\item{pos_style}{the style for plotting the base positions along the x-axis.
Defaults to "to_scale", plotting (potentially) overlapping squares
along the genomic position to scale. The "compact" options plots only the
positions with measured modification.}

\item{window_prop}{the size of flanking region to plot. Can be a vector of two
values for left and right window size. Values indicate proportion of region
length.}

\item{subsample}{the number of read of packed read rows to subsample to.}
}
\value{
a ggplot object of the heatmap.

a ggplot plot containing the heatmap.
}
\description{
Plot the methylation heatmap of a genomic region.
}
\details{
This function creates a heatmap visualization of methylation data for a specified
genomic region. Each row represents one or more packed reads, with colored segments
indicating methylation probability at each position. Reads are packed vertically
to minimize plot height while avoiding overlaps.
}
\examples{
nmr <- load_example_nanomethresult()
plot_region_heatmap(nmr, "chr7", 6703892, 6730431)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/modbam_to_tabix.R
\name{modbam_to_tabix}
\alias{modbam_to_tabix}
\title{Convert BAM with modifications to tabix format}
\usage{
modbam_to_tabix(x, out_file, mod_code = NanoMethViz::mod_code(x))
}
\arguments{
\item{x}{the \code{ModBamResult} object.}

\item{out_file}{the path of the output tabix.}

\item{mod_code}{the modification code to use, defaults to 'm' for 5mC
methylation.}
}
\value{
invisibly returns the name of the created tabix file.
}
\description{
The \code{modbam_to_tabix} function takes a ModBamResult object and
converts it into a tabix file format, which is efficient for indexing and
querying large datasets.
}
\details{
The possible tags for mod_code can be found at
\url{https://samtools.github.io/hts-specs/SAMtags.pdf} under the
'Base modifications' section.
}
\examples{
out_file <- paste0(tempfile(), ".tsv.bgz")
mbr <- ModBamResult(
    methy = ModBamFiles(
        samples = "sample1",
        paths = system.file("peg3.bam", package = "NanoMethViz",
        mustWork = FALSE)
    ),
    samples = data.frame(
        sample = "sample1",
        group = "group1"
    )
)

modbam_to_tabix(mbr, out_file)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bsseq_to_edger.R
\name{methy_to_edger}
\alias{methy_to_edger}
\title{Convert NanoMethResult object to edgeR methylation matrix}
\usage{
methy_to_edger(methy, regions = NULL, out_folder = tempdir(), verbose = TRUE)
}
\arguments{
\item{methy}{the NanoMethResult object or path to the methylation tabix file.}

\item{regions}{the regions to calculate log-methylation ratios over. If left
NULL, ratios will be calculated per site.}

\item{out_folder}{the folder to store intermediate files. One file is created
for each sample and contains columns "chr", "pos", "total" and
"methylated".}

\item{verbose}{TRUE if progress messages are to be printed}
}
\value{
a matrix compatible with the edgeR differential methylation pipeline
}
\description{
Convert NanoMethResult object to edgeR methylation matrix
}
\examples{
nmr <- load_example_nanomethresult()
edger_mat <- methy_to_edger(nmr)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/AllGenerics.R
\name{samples}
\alias{samples}
\title{Get sample annotation}
\usage{
samples(object)
}
\description{
Get sample annotation
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/query_methy.R
\name{query_methy}
\alias{query_methy}
\title{Query methylation data}
\usage{
query_methy(
  x,
  chr,
  start,
  end,
  simplify = TRUE,
  force = FALSE,
  truncate = TRUE,
  site_filter = getOption("NanoMethViz.site_filter", 3L)
)
}
\arguments{
\item{x}{the NanoMethResults object or a path to the methylation data
(tabix-bgzipped).}

\item{chr}{the vector of chromosomes}

\item{start}{the vector of start positions}

\item{end}{the vector of end positions}

\item{simplify}{whether returned results should be row-concatenated}

\item{force}{whether to force empty output when query region 'chr' does not
appear in data. Without 'force', an empty result indicates that the
requested 'chr' appears in the data but no data overlaps with requested
region, and an invalid 'chr' will cause an error.}

\item{truncate}{when querying from ModBamFiles, whether or not to truncate
returned results to only those within the specified region. Otherwise
methylation data for entire reads overlapping the region will be returned.}

\item{site_filter}{the minimum amount of coverage to report a site. This
filters the queried data such that any site with less than the filter is
not returned. The default is 1, which means that all sites are returned.
This option can be set globally using the \code{options(site_filter = ...)}
which will affect all plotting functions in NanoMethviz.}
}
\value{
A table containing the data within the queried regions. If simplify
is TRUE (default) then returns all data in a single table, otherwise returns
a list of tables where each table is the data from one region.
}
\description{
Query methylation data
}
\details{
The argument \code{site_filter} can be set globally using the \code{options(site_filter = ...)} command. The same data entry may appear multiple times in the output
if it overlaps multiple regions.
}
\examples{
nmr <- load_example_nanomethresult()
query_methy(methy(nmr), "chr7", 6703892, 6730431)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot_mds.R
\name{plot_mds}
\alias{plot_mds}
\title{Plot MDS}
\usage{
plot_mds(
  x,
  top = 500,
  plot_dims = c(1, 2),
  labels = colnames(x),
  groups = NULL,
  legend_name = "group"
)
}
\arguments{
\item{x}{a numeric matrix of log-methylation-ratio values where rows are genomic regions/features and columns
are samples.}

\item{top}{the number of top variable features (genes/regions) used to calculate
pairwise distances between samples.}

\item{plot_dims}{the numeric vector of the two dimensions to be plotted
(default c(1, 2) for first two principal coordinates).}

\item{labels}{the character vector of labels for data points. By default uses
column names of x, set to NULL to plot unlabeled points.}

\item{groups}{the character vector or numeric vector of groups that data points
will be coloured by. For character vectors, discrete colours are used. For
numeric vectors, continuous colour scale is used. By default, groups is NULL
and points are not coloured.}

\item{legend_name}{the name for the colour legend.}
}
\value{
ggplot object of the MDS plot.
}
\description{
Plot multi-dimensional scaling plot using algorithm of limma::plotMDS(). It is recommended this be done with the
log-methylation-ratio matrix generated by bsseq_to_log_methy_ratio().
}
\details{
Multi-dimensional scaling (MDS) is a dimensionality reduction technique that
represents high-dimensional distances between samples in a lower-dimensional
space (typically 2D). This function uses limma's plotMDS algorithm, which
selects the most variable features and calculates leading log-fold-change
distances between samples. The resulting plot shows samples as points where
proximity indicates similarity in methylation profiles. When variance explained
is available, it is shown as percentages on the axis labels.
}
\examples{
nmr <- load_example_nanomethresult()
bss <- methy_to_bsseq(nmr)
lmr <- bsseq_to_log_methy_ratio(bss)
plot_mds(lmr)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/AllGenerics.R
\name{mod_code<-}
\alias{mod_code<-}
\title{Set mod code}
\usage{
mod_code(object) <- value
}
\description{
Set mod code
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot_violin.R
\name{plot_violin}
\alias{plot_violin}
\title{Plot violin for regions}
\usage{
plot_violin(
  x,
  regions,
  binary_threshold = 0.5,
  group_col = "group",
  palette = ggplot2::scale_colour_brewer(palette = "Set1")
)
}
\arguments{
\item{x}{the NanoMethResult object.}

\item{regions}{a table of regions containing at least columns chr, strand,
start and end. Any additional columns can be used for grouping.}

\item{binary_threshold}{the modification probability such that calls with
modification probability above the threshold are considered methylated, and
those with probability equal or below are considered unmethylated.}

\item{group_col}{the column to group aggregated trends by. This column can
be in from the regions table or samples(x).}

\item{palette}{the ggplot colour palette used for groups.}
}
\value{
a ggplot object containing the methylation violin plot.
}
\description{
This function plots a violin plot of the methylation proportion for each region in the regions table. The methylation proportion is calculated as the mean of the modification probability wihtin each region and the violin represents the . The regions are then grouped and coloured by the group_col column in the regions table or samples(x).
}
\examples{
nmr <- load_example_nanomethresult()
gene_anno <- exons_to_genes(NanoMethViz::exons(nmr))
plot_violin(nmr, gene_anno)
plot_violin(nmr, gene_anno, group_col = "sample")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot_agg_genes.R
\name{plot_agg_genes}
\alias{plot_agg_genes}
\title{Plot gene aggregate plot}
\usage{
plot_agg_genes(
  x,
  genes = NULL,
  binary_threshold = 0.5,
  group_col = NULL,
  flank = 2000,
  stranded = TRUE,
  span = 0.05,
  palette = ggplot2::scale_colour_brewer(palette = "Set1")
)
}
\arguments{
\item{x}{the NanoMethResult or ModBamResult object.}

\item{genes}{a character vector of gene symbols to include in aggregate plot.
If NULL (default), all genes in exons(x) are used.}

\item{binary_threshold}{the modification probability such that calls with
modification probability above the threshold are considered methylated, and
those with probability equal or below are considered unmethylated.}

\item{group_col}{the column name to group aggregated trends by. This column can
be found in either the regions table or samples(x). When NULL (default), all
data is aggregated together. Common values include "sample" to show individual
samples or "group" to show sample groups.}

\item{flank}{the number of flanking bases to add to each side of each region.}

\item{stranded}{if TRUE, negative strand features will have their coordinates
flipped to reflect biological features like transcription start sites
(e.g., for genes, coordinates run from TSS to TES regardless of strand).}

\item{span}{the span parameter for loess smoothing of the trend lines.}

\item{palette}{the ggplot colour palette used for groups.}
}
\value{
a ggplot object containing the aggregate methylation trend of genes.
}
\description{
Plot gene aggregate plot
}
\details{
This function creates an aggregate methylation profile across multiple genes by
scaling all genes to the same relative coordinates (0 to 1) and averaging
methylation levels at each relative position. Genes are optionally extended by
flanking regions specified by the flank parameter. The resulting plot shows
smoothed trends of average methylation probability from gene start to gene end,
with optional flanking regions.
}
\examples{
nmr <- load_example_nanomethresult()
plot_agg_genes(nmr)

# Plot specific genes only
plot_agg_genes(nmr, genes = c("Peg3", "Impact"))

# Group by sample
plot_agg_genes(nmr, group_col = "sample")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/NanoMethResults.R
\docType{class}
\name{NanoMethResult-class}
\alias{NanoMethResult-class}
\alias{NanoMethResult}
\alias{methy,NanoMethResult-method}
\alias{methy<-,NanoMethResult,ANY-method}
\alias{samples,NanoMethResult-method}
\alias{samples<-,NanoMethResult,data.frame-method}
\alias{exons,NanoMethResult-method}
\alias{exons<-,NanoMethResult,data.frame-method}
\title{Nanopore Methylation Result}
\usage{
NanoMethResult(methy, samples, exons = NULL)

\S4method{methy}{NanoMethResult}(object)

\S4method{methy}{NanoMethResult,ANY}(object) <- value

\S4method{samples}{NanoMethResult}(object)

\S4method{samples}{NanoMethResult,data.frame}(object) <- value

\S4method{exons}{NanoMethResult}(object)

\S4method{exons}{NanoMethResult,data.frame}(object) <- value
}
\arguments{
\item{methy}{the path to the methylation tabix file.}

\item{samples}{the data.frame of sample annotation containing at least
columns sample and group.}

\item{exons}{(optional) the data.frame of exon information containing at
least columns gene_id, chr, strand, start, end, transcript_id and symbol.}

\item{object}{the NanoMethResult object.}

\item{value}{the exon annotation.}
}
\value{
a NanoMethResult object to be used with plotting functions

the path to the methylation data.

the sample annotation.

the exon annotation.
}
\description{
A NanoMethResult object stores data used for NanoMethViz visualisation. It
contains stores a path to the methylation data, sample information and
optional exon information. The object is constructed using the
NanoMethResult() constructor function described in "Usage".
}
\section{Functions}{
\itemize{
\item \code{NanoMethResult()}: Constructor

\item \code{methy(NanoMethResult)}: methylation data path getter.

\item \code{methy(object = NanoMethResult) <- value}: methylation data path setter.

\item \code{samples(NanoMethResult)}: sample annotation getter.

\item \code{samples(object = NanoMethResult) <- value}: sample annotation setter.

\item \code{exons(NanoMethResult)}: exon annotation getter.

\item \code{exons(object = NanoMethResult) <- value}: exon annotation setter.

}}
\section{Slots}{

\describe{
\item{\code{methy}}{the path to the methylation tabix file.}

\item{\code{samples}}{the data.frame of sample annotation containing at least columns
sample and group.}

\item{\code{exons}}{the data.frame of exon information containing at least columns
gene_id, chr, strand, start, end, transcript_id and symbol.}
}}

\examples{
methy <- system.file(package = "NanoMethViz", "methy_subset.tsv.bgz", mustWork = FALSE)
sample <- c(
    "B6Cast_Prom_1_bl6",
    "B6Cast_Prom_1_cast",
    "B6Cast_Prom_2_bl6",
    "B6Cast_Prom_2_cast",
    "B6Cast_Prom_3_bl6",
    "B6Cast_Prom_3_cast"
)
group <- c(
    "bl6",
    "cast",
    "bl6",
    "cast",
    "bl6",
    "cast"
)
sample_anno <- data.frame(sample, group, stringsAsFactors = FALSE)
exon_tibble <- get_example_exons_mus_musculus()
NanoMethResult(methy, sample_anno, exon_tibble)

x <- load_example_nanomethresult()
methy(x)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cluster_regions.R
\name{cluster_regions}
\alias{cluster_regions}
\title{Cluster regions by K-means}
\usage{
cluster_regions(x, regions, centers = 2, grid_method = c("density", "uniform"))
}
\arguments{
\item{x}{the NanoMethResult object.}

\item{regions}{a table of regions containing at least columns chr, strand,
start and end.}

\item{centers}{number of centers for k-means, identical to the number of
output clusters.}

\item{grid_method}{the method for generating the sampling grid. The default
option "density" attempts to create a grid with similar density as the
data, "uniform" creates a grid of uniform density.}
}
\value{
the table of regions given by the 'regions' argument with the column
'cluster' added.
}
\description{
Cluster regions by k-means based on their methylation profiles. In order to
cluster using k-means the methylation profile of each region is interpolated
and sampled at fixed points. The first 10 principal components are used for
the k-means clustering. The clustering is best behaved in regions of similar
width and CpG density.
}
\examples{
nmr <- load_example_nanomethresult()
gene_anno <- exons_to_genes(NanoMethViz::exons(nmr))
# uniform grid due to low number of input features
gene_anno_clustered <- cluster_regions(nmr, gene_anno, centers = 2, grid_method = "uniform")
plot_agg_regions(nmr, gene_anno_clustered, group_col = "cluster")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot_agg_regions.R
\name{plot_agg_regions}
\alias{plot_agg_regions}
\title{Plot aggregate regions}
\usage{
plot_agg_regions(
  x,
  regions,
  binary_threshold = 0.5,
  group_col = NULL,
  flank = 2000,
  stranded = TRUE,
  span = 0.05,
  palette = ggplot2::scale_colour_brewer(palette = "Set1")
)
}
\arguments{
\item{x}{the NanoMethResult or ModBamResult object.}

\item{regions}{a table of regions containing at least columns chr, strand,
start and end. Any additional columns can be used for grouping.}

\item{binary_threshold}{the modification probability such that calls with
modification probability above the threshold are considered methylated, and
those with probability equal or below are considered unmethylated.}

\item{group_col}{the column name to group aggregated trends by. This column can
be found in either the regions table or samples(x). When NULL (default), all
data is aggregated together. Common values include "sample" to show individual
samples or "group" to show sample groups.}

\item{flank}{the number of flanking bases to add to each side of each region.}

\item{stranded}{if TRUE, negative strand features will have their coordinates
flipped to reflect biological features like transcription start sites
(e.g., for genes, coordinates run from TSS to TES regardless of strand).}

\item{span}{the span parameter for loess smoothing of the trend lines.}

\item{palette}{the ggplot colour palette used for groups.}
}
\value{
a ggplot object containing the aggregate methylation trend.
}
\description{
Plot aggregate regions
}
\examples{
nmr <- load_example_nanomethresult()
gene_anno <- exons_to_genes(NanoMethViz::exons(nmr))
plot_agg_regions(nmr, gene_anno)
plot_agg_regions(nmr, gene_anno, group_col = "sample")
plot_agg_regions(nmr, gene_anno, group_col = "group")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot_pca.R
\name{plot_pca}
\alias{plot_pca}
\title{Plot PCA}
\usage{
plot_pca(
  x,
  plot_dims = c(1, 2),
  labels = colnames(x),
  groups = NULL,
  legend_name = "group"
)
}
\arguments{
\item{x}{a numeric matrix where rows are genomic regions/features and columns
are samples. Should be a log-methylation-ratio matrix for best results.}

\item{plot_dims}{the numeric vector of the two dimensions to be plotted
(default c(1, 2) for PC1 vs PC2).}

\item{labels}{the character vector of labels for data points. By default uses
column names of x, set to NULL to plot unlabeled points.}

\item{groups}{the character vector or numeric vector of groups that data points
will be coloured by. For character vectors, discrete colours are used. For
numeric vectors, continuous colour scale is used. By default, groups is NULL
and points are not coloured.}

\item{legend_name}{the name for the colour legend.}
}
\value{
ggplot object of the PCA plot.
}
\description{
Plot principal component analysis plot using BiocSingular::runPCA(). It is recommended this be done with
the log-methylation-ratio matrix generated by bsseq_to_log_methy_ratio().
}
\details{
Principal Component Analysis (PCA) is a dimensionality reduction technique that
identifies the directions of maximum variance in high-dimensional data. Unlike
MDS, PCA axis labels show the principal component numbers rather than variance
explained percentages. The first principal component (PC1) captures the most
variance in the data, PC2 captures the second most, and so on. Samples that
cluster together have similar methylation profiles across the genomic regions
included in the analysis.
}
\examples{
nmr <- load_example_nanomethresult()
bss <- methy_to_bsseq(nmr)
lmr <- bsseq_to_log_methy_ratio(bss)
plot_pca(lmr)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tabix_utils.R
\name{sort_methy_file}
\alias{sort_methy_file}
\title{Sort methylation file}
\usage{
sort_methy_file(x)
}
\arguments{
\item{x}{the path to the methylation file to sort}
}
\value{
invisibly returns path of sorted file
}
\description{
Sort methylation file
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{load_example_nanomethresult}
\alias{load_example_nanomethresult}
\title{Load an example NanoMethResult object}
\usage{
load_example_nanomethresult()
}
\value{
a NanoMethResults object
}
\description{
Load an example NanoMethResult object for demonstration of plotting
functions. Run \code{load_example_nanomethresults} without the function call to
see how the object is constructed.
}
\examples{
nmr <- load_example_nanomethresult()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/AllGenerics.R
\name{methy<-}
\alias{methy<-}
\title{Set methylation data}
\usage{
methy(object) <- value
}
\description{
Set methylation data
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/AllGenerics.R, R/plot_gene_heatmap.R
\name{plot_gene_heatmap}
\alias{plot_gene_heatmap}
\alias{plot_gene_heatmap,NanoMethResult,character-method}
\alias{plot_gene_heatmap,ModBamResult,character-method}
\title{Plot gene methylation heatmap}
\usage{
plot_gene_heatmap(x, gene, ...)

\S4method{plot_gene_heatmap}{NanoMethResult,character}(
  x,
  gene,
  window_prop = 0.3,
  pos_style = c("to_scale", "compact"),
  subsample = 50
)

\S4method{plot_gene_heatmap}{ModBamResult,character}(
  x,
  gene,
  window_prop = 0.3,
  pos_style = c("to_scale", "compact"),
  subsample = 50
)
}
\arguments{
\item{x}{the NanoMethResult or ModBamResult object.}

\item{gene}{the gene symbol for the gene to plot.}

\item{...}{additional arguments.}

\item{window_prop}{the size of flanking region to plot. Can be a vector of two
values for left and right window size. Values indicate proportion of gene
length.}

\item{pos_style}{the style for plotting the base positions along the x-axis.
Defaults to "to_scale", plotting (potentially) overlapping squares
along the genomic position to scale. The "compact" options plots only the
positions with measured modification.}

\item{subsample}{the number of read of packed read rows to subsample to.}
}
\value{
a ggplot object of the heatmap

a ggplot plot containing the heatmap.
}
\description{
Plot the methylation heatmap of a gene symbol specified within the exon(x) slot.
}
\details{
This function creates a heatmap visualization of methylation data for a specific gene.
Each row in the heatmap represents one or more packed reads, where colored segments
indicate methylation probability at each genomic position.
}
\examples{
nmr <- load_example_nanomethresult()
plot_gene_heatmap(nmr, "Peg3")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tabix_utils.R
\name{raw_methy_to_tabix}
\alias{raw_methy_to_tabix}
\title{Convert methylation file to tabix format}
\usage{
raw_methy_to_tabix(x)
}
\arguments{
\item{x}{the path to the sorted methylation file}
}
\value{
invisibly returns the path to the tabix file
}
\description{
Convert methylation file to tabix format
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/filter_methy.R
\name{filter_methy}
\alias{filter_methy}
\title{Create filtered methylation file}
\usage{
filter_methy(x, output_file, ...)
}
\arguments{
\item{x}{the path to the methylation file or a NanoMethResult object.}

\item{output_file}{the output file to write results to (must end in .bgz).}

\item{...}{filtering criteria given in dplyr syntax. Use methy_col_names()
to get available column names.}
}
\value{
invisibly returns 'output_file' if x is a file path, otherwise returns
NanoMethResult object with methy(x) replaced with filtered value.
}
\description{
Create a filtered methylation file from an existing one.
}
\examples{
nmr <- load_example_nanomethresult()
output_file <- paste0(tempfile(), ".tsv.bgz")
filter_methy(nmr, output_file = output_file, chr == "chrX")
filter_methy(methy(nmr), output_file = output_file, chr == "chrX")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_exons.R
\name{get_example_exons_mus_musculus}
\alias{get_example_exons_mus_musculus}
\title{Get example exon annotations for mus musculus (mm10)}
\usage{
get_example_exons_mus_musculus()
}
\value{
data.frame containing exons
}
\description{
This is a small subset of the exons returned by
\code{get_exons_mus_musculus()} for demonstrative purposes. It contains
the exons for the genes Brca1, Brca2, Impact, Meg3, Peg3 and Xist.
}
\examples{
example_exons <- get_example_exons_mus_musculus()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_cgi.R, R/get_exons.R
\name{get_cgi_mm10}
\alias{get_cgi_mm10}
\alias{get_cgi_grcm39}
\alias{get_cgi_t2t}
\alias{get_cgi_hg19}
\alias{get_cgi_hg38}
\alias{get_exons}
\alias{get_exons_mm10}
\alias{get_exons_grcm39}
\alias{get_exons_hg19}
\alias{get_exons_hg38}
\alias{get_exons_t2t}
\title{Get exon annotations}
\usage{
get_cgi_mm10()

get_cgi_grcm39()

get_cgi_t2t()

get_cgi_hg19()

get_cgi_hg38()

get_exons_mm10()

get_exons_grcm39()

get_exons_hg19()

get_exons_hg38()

get_exons_t2t()
}
\value{
tibble (data.frame) object containing exon annotation.
}
\description{
Helper functions are provided for obtaining exon annotations from relevant
TxDb packages on Bioconductor for the construction of NanoMethResults
objects.
}
\examples{
cgi_mm10 <- get_cgi_mm10()

cgi_GRCm39 <- get_cgi_grcm39()

cgi_t2t <- get_cgi_t2t()

cgi_hg19 <- get_cgi_hg19()

cgi_hg38 <- get_cgi_hg38()

mm10_exons <- get_exons_mm10()

grcm39_exons <- get_exons_grcm39()

hg19_exons <- get_exons_hg19()

hg38_exons <- get_exons_hg38()

t2t_exons <- get_exons_t2t()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ModBamResults.R
\docType{class}
\name{ModBamFiles-class}
\alias{ModBamFiles-class}
\title{ModBamFiles class}
\description{
This is a class for holding information about modbam files. It is a
data.frame containing information about samples and paths to modbam files.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot_grange.R
\name{plot_grange}
\alias{plot_grange}
\title{Plot GRanges}
\usage{
plot_grange(
  x,
  grange,
  anno_regions = NULL,
  binary_threshold = NULL,
  avg_method = c("mean", "median"),
  spaghetti = FALSE,
  heatmap = TRUE,
  heatmap_subsample = 50,
  gene_anno = TRUE,
  smoothing_window = 2000,
  window_prop = 0,
  palette = ggplot2::scale_colour_brewer(palette = "Set1"),
  line_size = 1,
  span = NULL
)
}
\arguments{
\item{x}{the NanoMethResult object.}

\item{grange}{the GRanges object with one entry.}

\item{anno_regions}{the data.frame of regions to be annotated.}

\item{binary_threshold}{the modification probability such that calls with
modification probability above the threshold are set to 1 and probabilities
equal to or below the threshold are set to 0.}

\item{avg_method}{the average method for pre-smoothing at each genomic position.
Data is pre-smoothed at each genomic position before the smoothed aggregate line
is generated for performance reasons. The default is "mean" which corresponds
to the average methylation fraction. The alternative "median" option is
closer to an average within the more common methylation state.}

\item{spaghetti}{whether or not individual reads should be shown.}

\item{heatmap}{whether or not read-methylation heatmap should be shown.}

\item{heatmap_subsample}{how many packed rows of reads to subsample to.}

\item{gene_anno}{whether to show gene annotation.}

\item{smoothing_window}{the window size for smoothing the trend line.}

\item{window_prop}{the size of flanking region to plot. Can be a vector of two
values for left and right window size. Values indicate proportion of gene
length.}

\item{palette}{the ggplot colour palette used for groups.}

\item{line_size}{the size of the lines.}

\item{span}{DEPRECATED, use smoothing_window instead. Will be removed in next version.}
}
\value{
a patchwork plot containing the methylation profile in the specified
region.
}
\description{
Plot GRanges
}
\examples{
nmr <- load_example_nanomethresult()
plot_grange(nmr, GenomicRanges::GRanges("chr7:6703892-6730431"))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/convert_methy_format.R
\name{convert_methy_format}
\alias{convert_methy_format}
\title{Convert methylation calls to NanoMethViz format}
\usage{
convert_methy_format(
  input_files,
  output_file,
  samples = fs::path_ext_remove(fs::path_file(input_files)),
  verbose = TRUE
)
}
\arguments{
\item{input_files}{the files to convert}

\item{output_file}{the output file to write results to (must end in .bgz)}

\item{samples}{the names of samples corresponding to each file}

\item{verbose}{TRUE if progress messages are to be printed}
}
\value{
invisibly returns the output file path, creates a tabix file (.bgz)
and its index (.bgz.tbi)
}
\description{
Convert methylation calls to NanoMethViz format
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_exons.R
\name{get_exons_homo_sapiens}
\alias{get_exons_homo_sapiens}
\title{Get exon annotations for Homo sapiens (hg19)}
\usage{
get_exons_homo_sapiens()
}
\value{
data.frame containing exons
}
\description{
Get exon annotations for Homo sapiens (hg19)
}
\examples{
h_sapiens_exons <- get_exons_homo_sapiens()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cluster_reads.R
\name{cluster_reads}
\alias{cluster_reads}
\title{Cluster reads based on methylation}
\usage{
cluster_reads(x, chr, start, end, min_pts = 5)
}
\arguments{
\item{x}{a ModBamResult object.}

\item{chr}{the chromosome name where to find the region.}

\item{start}{the start position of the region.}

\item{end}{the end position of the region.}

\item{min_pts}{the minimum number of points needed to form a cluster (default = 10).}
}
\value{
A tibble with information about each read's cluster assignment and read statistics.
}
\description{
Cluster reads based on methylation
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ModBamResults.R
\docType{class}
\name{ModBamResult-class}
\alias{ModBamResult-class}
\alias{methy,ModBamResult-method}
\alias{methy<-,ModBamResult,ModBamFiles-method}
\alias{samples,ModBamResult-method}
\alias{samples<-,ModBamResult,data.frame-method}
\alias{exons,ModBamResult-method}
\alias{exons<-,ModBamResult,data.frame-method}
\alias{mod_code,ModBamResult-method}
\alias{mod_code<-,ModBamResult,character-method}
\alias{ModBamResult}
\title{Modbam methylation results}
\usage{
\S4method{methy}{ModBamResult}(object)

\S4method{methy}{ModBamResult,ModBamFiles}(object) <- value

\S4method{samples}{ModBamResult}(object)

\S4method{samples}{ModBamResult,data.frame}(object) <- value

\S4method{exons}{ModBamResult}(object)

\S4method{exons}{ModBamResult,data.frame}(object) <- value

\S4method{mod_code}{ModBamResult}(object)

\S4method{mod_code}{ModBamResult,character}(object) <- value

ModBamResult(methy, samples, exons = NULL, mod_code = "m")
}
\arguments{
\item{object}{the ModBamResult object.}

\item{value}{the mod code.}

\item{methy}{a ModBamFiles object.}

\item{samples}{the data.frame of sample annotation containing at least
columns sample and group.}

\item{exons}{(optional) the data.frame of exon information containing at
least columns gene_id, chr, strand, start, end, transcript_id and symbol.}

\item{mod_code}{a character with the mod code of interest. Defaults to "m"
for 5mC. See details for other options.}
}
\value{
a NanoMethResult object to be used with plotting functions

a ModBamFiles data.frame.

the sample annotation.

the exon annotation.

the mod code.
}
\description{
A ModBamResult object stores modbam data used for NanoMethViz
visualisation. It contains stores a ModBamFiles object, sample information
and optional exon information. The object is constructed using the
ModBamResult() constructor function described in "Usage".
}
\details{
The possible tags for mod_code can be found at
\url{https://samtools.github.io/hts-specs/SAMtags.pdf} under the
'Base modifications' section.
}
\section{Functions}{
\itemize{
\item \code{methy(ModBamResult)}: modbam information getter.

\item \code{methy(object = ModBamResult) <- value}: modbam information setter.

\item \code{samples(ModBamResult)}: sample annotation getter.

\item \code{samples(object = ModBamResult) <- value}: sample annotation setter.

\item \code{exons(ModBamResult)}: exon annotation getter.

\item \code{exons(object = ModBamResult) <- value}: exon annotation setter.

\item \code{mod_code(ModBamResult)}: mod code getter.

\item \code{mod_code(object = ModBamResult) <- value}: mod code setter.

\item \code{ModBamResult()}: Constructor

}}
\section{Slots}{

\describe{
\item{\code{methy}}{a ModBamFiles data.frame specifying the samples and paths to bam
files.}

\item{\code{samples}}{the data.frame of sample annotation containing at least columns
sample and group.}

\item{\code{exons}}{the data.frame of exon information containing at least columns
gene_id, chr, strand, start, end, transcript_id and symbol.}

\item{\code{mod_code}}{the modification code of interest.}
}}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/AllGenerics.R
\name{exons}
\alias{exons}
\title{Get exon annotation}
\usage{
exons(object)
}
\description{
Get exon annotation
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_cgi.R
\name{get_cgi}
\alias{get_cgi}
\title{Get CpG islands annotations}
\usage{
get_cgi(genome)
}
\value{
tibble (data.frame) object containing CpG islands annotation.
}
\description{
Helper functions are provided for obtaining CpG islands annotations from UCSC table browser.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_exons.R
\name{get_exons_mus_musculus}
\alias{get_exons_mus_musculus}
\title{Get exon annotations for Mus musculus (mm10)}
\usage{
get_exons_mus_musculus()
}
\value{
data.frame containing exons
}
\description{
Get exon annotations for Mus musculus (mm10)
}
\examples{
m_musculus_exons <- get_exons_mus_musculus()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/methy_to_bsseq.R
\name{methy_to_bsseq}
\alias{methy_to_bsseq}
\title{Create BSSeq object from methylation tabix file}
\usage{
methy_to_bsseq(methy, out_folder = tempdir(), verbose = TRUE)
}
\arguments{
\item{methy}{the NanoMethResult object or path to the methylation tabix file.}

\item{out_folder}{the folder to store intermediate files. One file is created
for each sample and contains columns "chr", "pos", "total" and
"methylated".}

\item{verbose}{TRUE if progress messages are to be printed}
}
\value{
a BSSeq object.
}
\description{
Create BSSeq object from methylation tabix file
}
\examples{
nmr <- load_example_nanomethresult()
bsseq <- methy_to_bsseq(nmr)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/AllGenerics.R
\name{samples<-}
\alias{samples<-}
\title{Set sample annotation}
\usage{
samples(object) <- value
}
\description{
Set sample annotation
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bsseq_to_edger.R
\name{bsseq_to_log_methy_ratio}
\alias{bsseq_to_log_methy_ratio}
\title{Convert BSseq object to log-methylation-ratio matrix}
\usage{
bsseq_to_log_methy_ratio(
  bsseq,
  regions = NULL,
  prior_count = 2,
  drop_na = TRUE
)
}
\arguments{
\item{bsseq}{the BSseq object.}

\item{regions}{the regions to calculate log-methylation ratios over. If left NULL, ratios will be calculated per
site.}

\item{prior_count}{the prior count added to avoid taking log of 0.}

\item{drop_na}{whether to drop rows with all NA values.}
}
\value{
a matrix containing log-methylation-ratios.
}
\description{
Creates a log-methylation-ratio matrix from a BSseq object that is useful for
dimensionality reduction plots.
}
\examples{
nmr <- load_example_nanomethresult()
bsseq <- methy_to_bsseq(nmr)
regions <- exons_to_genes(NanoMethViz::exons(nmr))
log_m_ratio <- bsseq_to_log_methy_ratio(bsseq, regions)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/col_specs.R
\name{methy_col_names}
\alias{methy_col_names}
\title{Column names for methylation data}
\usage{
methy_col_names()
}
\value{
column names for methylation data
}
\description{
Column names for methylation data
}
\examples{
methy_col_names()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/AllGenerics.R
\name{methy}
\alias{methy}
\title{Get methylation data}
\usage{
methy(object)
}
\arguments{
\item{object}{the object.}
}
\value{
the path to the methylation data.
}
\description{
Get methylation data
}
\examples{
showMethods("methy")

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/exons_to_genes.R
\name{exons_to_genes}
\alias{exons_to_genes}
\title{Convert exon annotation to genes}
\usage{
exons_to_genes(x)
}
\arguments{
\item{x}{the exon level annotation containing columns "gene_id", "chr",
"strand" and "symbol".}
}
\value{
the gene level annotation where each gene is taken to span the
earliest start position and latest end position of its exons.
}
\description{
Convert exon annotation to genes
}
\examples{
nmr <- load_example_nanomethresult()
exons_to_genes(NanoMethViz::exons(nmr))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/query_exons.R
\name{query_exons}
\alias{query_exons}
\alias{query_exons_region}
\alias{query_exons_gene_id}
\alias{query_exons_symbol}
\title{Query exons}
\usage{
query_exons_region(x, chr, start, end)

query_exons_gene_id(x, gene_id)

query_exons_symbol(x, symbol)
}
\arguments{
\item{x}{the object to query.}

\item{chr}{the chromosome to query.}

\item{start}{the start of the query region.}

\item{end}{the end of the query region.}

\item{gene_id}{the gene_id to query.}

\item{symbol}{the gene_id to query.}
}
\value{
data.frame of queried exons.
}
\description{
Query a data.frame, NanoMethResult or ModBamResult for exon annotation.
}
\section{Functions}{
\itemize{
\item \code{query_exons_region()}: Query region.

\item \code{query_exons_gene_id()}: Query gene ID.

\item \code{query_exons_symbol()}: Query gene symbol.

}}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot_grange.R
\name{plot_grange_heatmap}
\alias{plot_grange_heatmap}
\title{Plot GRanges heatmap}
\usage{
plot_grange_heatmap(
  x,
  grange,
  pos_style = c("to_scale", "compact"),
  window_prop = 0,
  subsample = 50
)
}
\arguments{
\item{x}{the NanoMethResult object.}

\item{grange}{the GRanges object with one entry.}

\item{pos_style}{the style for plotting the base positions along the x-axis.
Defaults to "to_scale", plotting (potentially) overlapping squares
along the genomic position to scale. The "compact" options plots only the
positions with measured modification.}

\item{window_prop}{the size of flanking region to plot. Can be a vector of two
values for left and right window size. Values indicate proportion of region
length.}

\item{subsample}{the number of read of packed read rows to subsample to.}
}
\value{
a ggplot plot containing the heatmap.
}
\description{
Plot GRanges heatmap
}
\examples{
nmr <- load_example_nanomethresult()
plot_grange_heatmap(nmr, GenomicRanges::GRanges("chr7:6703892-6730431"))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/NanoMethViz-package.R
\docType{package}
\name{NanoMethViz-package}
\alias{NanoMethViz}
\alias{NanoMethViz-package}
\title{NanoMethViz: Visualise methylation data from Oxford Nanopore sequencing}
\description{
NanoMethViz is a toolkit for visualising methylation data from Oxford Nanopore sequencing. It can be used to explore methylation patterns from reads derived from Oxford Nanopore direct DNA sequencing with methylation called by callers including nanopolish, f5c and megalodon. The plots in this package allow the visualisation of methylation profiles aggregated over experimental groups and across classes of genomic features.
}
\details{
The main plotting functions in this package are \code{\link[=plot_gene]{plot_gene()}} and \code{\link[=plot_region]{plot_region()}}.
\itemize{
\item See \code{vignette("UserGuide", package = "NanoMethViz")} for documentation of how to use this package.
}
}
\seealso{
Useful links:
\itemize{
  \item \url{https://github.com/shians/NanoMethViz}
  \item \url{https://shians.github.io/NanoMethViz/}
  \item Report bugs at \url{https://github.com/Shians/NanoMethViz/issues}
}

}
\author{
\strong{Maintainer}: Shian Su \email{su.s@wehi.edu.au}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/region_methy_stats.R
\name{region_methy_stats}
\alias{region_methy_stats}
\title{Calculate region methylation statistics}
\usage{
region_methy_stats(nmr, regions, threshold = 0.5)
}
\arguments{
\item{nmr}{the NanoMethResult object.}

\item{regions}{the table of regions to query statistics for.}

\item{threshold}{the threshold to use for determining methylation calls
for the calculation of prevalence.}
}
\value{
table of regions with additional columns of methylation summary
statistics.
}
\description{
Calculate the average methylation probability and prevalence based on
specified probability threshold.
}
\examples{
nmr <- load_example_nanomethresult()
gene_anno <- exons_to_genes(NanoMethViz::exons(nmr))
region_methy_stats(nmr, gene_anno)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{load_example_modbamresult}
\alias{load_example_modbamresult}
\title{Load an example ModBamResult object}
\usage{
load_example_modbamresult()
}
\value{
a ModBamResult object
}
\description{
Load an example ModBamResult object for demonstration of plotting
functions. Run \code{load_example_modbamresult} without the function call to
see how the object is constructed.
}
\examples{
mbr <- load_example_modbamresult()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/AllGenerics.R, R/plot_gene.R
\name{plot_gene}
\alias{plot_gene}
\alias{plot_gene,NanoMethResult,character-method}
\alias{plot_gene,ModBamResult,character-method}
\title{Plot gene methylation}
\usage{
plot_gene(x, gene, ...)

\S4method{plot_gene}{NanoMethResult,character}(
  x,
  gene,
  window_prop = 0.3,
  anno_regions = NULL,
  binary_threshold = NULL,
  avg_method = c("mean", "median"),
  spaghetti = FALSE,
  heatmap = TRUE,
  heatmap_subsample = 50,
  smoothing_window = 2000,
  gene_anno = TRUE,
  palette = ggplot2::scale_colour_brewer(palette = "Set1"),
  line_size = 1,
  mod_scale = c(0, 1),
  span = NULL
)

\S4method{plot_gene}{ModBamResult,character}(
  x,
  gene,
  window_prop = 0.3,
  anno_regions = NULL,
  binary_threshold = NULL,
  avg_method = c("mean", "median"),
  spaghetti = FALSE,
  heatmap = TRUE,
  heatmap_subsample = 50,
  smoothing_window = 2000,
  gene_anno = TRUE,
  palette = ggplot2::scale_colour_brewer(palette = "Set1"),
  line_size = 1,
  mod_scale = c(0, 1),
  span = NULL
)
}
\arguments{
\item{x}{the NanoMethResult or ModBamResult object.}

\item{gene}{the gene symbol for the gene to plot.}

\item{...}{additional arguments.}

\item{window_prop}{the size of flanking region to plot. Can be a vector of two
values for left and right window size. Values indicate proportion of gene
length.}

\item{anno_regions}{the data.frame of regions to be annotated.}

\item{binary_threshold}{the modification probability such that calls with
modification probability above the threshold are set to 1 and probabilities
equal to or below the threshold are set to 0.}

\item{avg_method}{the average method for pre-smoothing at each genomic position.
Data is pre-smoothed at each genomic position before the smoothed aggregate line
is generated for performance reasons. The default is "mean" which corresponds
to the average methylation fraction. The alternative "median" option is
closer to an average within the more common methylation state.}

\item{spaghetti}{whether or not individual reads should be shown.}

\item{heatmap}{whether or not read-methylation heatmap should be shown.}

\item{heatmap_subsample}{how many packed rows of reads to subsample to.}

\item{smoothing_window}{the window size for smoothing the trend line.}

\item{gene_anno}{whether to show gene annotation.}

\item{palette}{the ggplot colour palette used for groups.}

\item{line_size}{the size of the lines.}

\item{mod_scale}{the scale range for modification probabilities. Default c(0, 1), set to "auto" for automatic
limits.}

\item{span}{DEPRECATED, use smoothing_window instead. Will be removed in next version.}
}
\value{
a patchwork plot containing the methylation profile in the specified
region.
}
\description{
Plot the methylation of a gene symbol specified within the exon(x) slot.
}
\details{
This function plots the methylation data for a given gene. The main trendline plot shows the average methylation
probability across the gene. The heatmap plot shows the methylation probability for each read across the gene. The
gene annotation plot shows the exons of the gene. In the heatmap, each row represents one or more non-overlapping reads
where the coloured segments represent the methylation probability at each position. Data along a read is connected by
a grey line. The gene annotation plot shows the isoforms and exons of the gene, with arrows indicating the direction
of transcription.

Since V3.0.0 NanoMethViz has changed the smoothing strategy from a loess smoothing to a weighted moving average. This
is because the loess smoothing was too computationally expensive for large datasets and had a span parameter that was
difficult to tune. The new smoothing strategy is controlled by the smoothing_window argument.
}
\section{Functions}{
\itemize{
\item \code{plot_gene(x = ModBamResult, gene = character)}: S4 method for ModBamResult

}}
\examples{
nmr <- load_example_nanomethresult()
plot_gene(nmr, "Peg3")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ModBamResults.R
\docType{methods}
\name{ModBamFiles}
\alias{ModBamFiles}
\alias{show,ModBamFiles-method}
\title{Constructor for a ModBamFiles object}
\usage{
ModBamFiles(samples, paths)

\S4method{show}{ModBamFiles}(object)
}
\arguments{
\item{samples}{a character vector with the names of the samples.}

\item{paths}{a character vector with the file paths for the BAM files.}

\item{object}{a ModBamFiles object.}
}
\value{
A ModBamFiles object with the sample and path information.
}
\description{
This function creates a ModBamFiles object containing information about the
samples and file paths. This constructor checks that the files are readable
and have an index.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bsseq_to_edger.R
\name{bsseq_to_edger}
\alias{bsseq_to_edger}
\title{Convert BSseq object to edgeR methylation matrix}
\usage{
bsseq_to_edger(bsseq, regions = NULL)
}
\arguments{
\item{bsseq}{the BSseq object.}

\item{regions}{the regions to calculate log-methylation ratios over. If left NULL, ratios will be calculated per
site.}
}
\value{
a matrix compatible with the edgeR differential methylation pipeline
}
\description{
Convert BSseq object to edgeR methylation matrix
}
\examples{
methy <- system.file("methy_subset.tsv.bgz", package = "NanoMethViz", mustWork = FALSE)
bsseq <- methy_to_bsseq(methy)
edger_mat <- bsseq_to_edger(bsseq)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/AllGenerics.R
\name{exons<-}
\alias{exons<-}
\title{Set exon annotation}
\usage{
exons(object) <- value
}
\description{
Set exon annotation
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/create_tabix_file.R
\name{create_tabix_file}
\alias{create_tabix_file}
\title{Create a tabix file using methylation calls}
\usage{
create_tabix_file(
  input_files,
  output_file,
  samples = extract_file_names(input_files),
  verbose = TRUE
)
}
\arguments{
\item{input_files}{the files to convert}

\item{output_file}{the output file to write results to (must end in .bgz)}

\item{samples}{the names of samples corresponding to each file}

\item{verbose}{TRUE if progress messages are to be printed}
}
\value{
invisibly returns the output file path, creates a tabix file (.bgz)
and its index (.bgz.tbi)
}
\description{
Create a tabix file using methylation calls
}
\examples{
methy_calls <- system.file(package = "NanoMethViz",
    c("sample1_nanopolish.tsv.gz", "sample2_nanopolish.tsv.gz"), mustWork = FALSE)
temp_file <- paste0(tempfile(), ".tsv.bgz")

create_tabix_file(methy_calls, temp_file)
}
